/**********************************************************

This software is part of J.-S. Caux's ABACUS library.

Copyright (c).

-----------------------------------------------------------

File:  LiebLin_State_Ensemble.cc

Purpose:  State ensembles for Lieb-Liniger


***********************************************************/

#include "JSC.h"

using namespace std;
using namespace JSC;

namespace JSC {

  // Constructors:

  LiebLin_Diagonal_State_Ensemble::LiebLin_Diagonal_State_Ensemble ()
    : nstates(0), state(Vect<LiebLin_Bethe_State>(1)), weight(0.0, 1)
  {
  }

  LiebLin_Diagonal_State_Ensemble::LiebLin_Diagonal_State_Ensemble (const LiebLin_Bethe_State& RefState, int nstates_req)
    : nstates(nstates_req), state(Vect<LiebLin_Bethe_State>(RefState, nstates_req)), weight(1.0/nstates_req, nstates_req)
  {
  }
  /*
  LiebLin_Diagonal_State_Ensemble::LiebLin_Diagonal_State_Ensemble (const LiebLin_Bethe_State& RefState, int nstates_req, const Vect<DP>& weight_ref)
    : nstates(nstates_req), state(Vect<LiebLin_Bethe_State>(RefState, nstates_req)), weight(weight_ref)
  {
    if (weight_ref.size() != nstates_req) JSCerror("Incompatible vector size in LiebLin_Diagonal_State_Ensemble constructor.");
  }
  */

  // Recursively go through all arbitrary-order type 2 descendents
  void Generate_type_2_descendents (LiebLin_Bethe_State& ActualState, int* ndesc_ptr, const LiebLin_Bethe_State& OriginState) 
  {
    int type_required = 3;
    Vect<string> desc_label = Descendents (ActualState, OriginState, type_required);

    if (desc_label.size() > 0) { // follow down recursively
      LiebLin_Bethe_State ActualState_here = ActualState;
      for (int idesc = 0; idesc < desc_label.size(); ++idesc) {
	ActualState_here.Set_to_Label (desc_label[idesc], OriginState.Ix2);
	// output data:
	cout << *ndesc_ptr << "\t" << desc_label[idesc] << endl;
	cout << "origin: " << OriginState.Ix2 << endl;
	cout << "shiftd: " << ActualState_here.Ix2 << endl;
	ActualState_here.Compute_All(false);
	cout << "\t\t\t\t\t\t" << ActualState.E - OriginState.E << endl;
	*ndesc_ptr += 1;
	Generate_type_2_descendents (ActualState_here, ndesc_ptr, OriginState);

	// do them mod st:
	idesc += 6;
      }
    }
    return;
  }
  
  
  

  //LiebLin_Diagonal_State_Ensemble::LiebLin_Diagonal_State_Ensemble (DP c_int, DP L, int N, const Root_Density& rho, int nstates_req)
  //: nstates(nstates_req)
  LiebLin_Diagonal_State_Ensemble::LiebLin_Diagonal_State_Ensemble (DP c_int, DP L, int N, const Root_Density& rho)
  {
    //version with 4 states in ensemble
    
    Root_Density x(rho.Npts, rho.lambdamax);

    for(int ix=0; ix<x.Npts; ++ix) {
      x.value[ix] = x.lambda[ix];
      for( int ip=0; ip<rho.Npts; ++ip) {
	x.value[ix] += 2.0 * rho.value[ip] * rho.dlambda[ip] * atan ((x.lambda[ix] - rho.lambda[ip])/c_int);
      }
      x.value[ix] /= 2.0*PI; //normalization

      //cout << x.lambda[ix] << "\t" << x.value[ix] << "\t" << rho.lambda[ix] "\t" << rho.value[ix] <<endl;
    }
    
    // Now carry on as per Discretized_LiebLin_Bethe_State:
    // Each time N \int_{-\infty}^\lambda d\lambda' \rho(\lambda') crosses a half integer, add a particle:
    DP integral = 0.0;
    DP integral_prev = 0.0;
    int Nfound = 0;
    Vect<DP> Ix2_found(0.0, N);
    int Ix2_left, Ix2_right;
    Vect<int> Ix2(N);
    
    
    LiebLin_Bethe_State rhostate(c_int, N, L);
    
    nstates = 4;
    state = Vect<LiebLin_Bethe_State>(rhostate, nstates);
    weight = Vect<DP>(nstates);
    weight[0] = 1.0/nstates;
    weight[1] = 1.0/nstates;
    weight[2] = 1.0/nstates;
    weight[3] = 1.0/nstates;
    
    int n_moves = 0;
    bool change = true;
    
    for (int i = 0; i < rho.Npts; ++i) {
      integral_prev = integral;
      integral += L * rho.value[i] * rho.dlambda[i];
      //cout << x.value[i] << "\t" << integral << endl;
      if (integral > Nfound + 0.5) {
	// Subtle error: if the rho is too discontinuous, i.e. if more than one rapidity is found, must correct for this.
		if (integral > Nfound + 1.5 && integral < Nfound + 2.5) { // found two rapidities
		  JSCerror("The distribution of particles is too discontinous for discretisation");
		}
		
		else {
		  // few variables to clarify the computations, they do not need to be vectors, not used outside this loop
		  //Ix2_found[Nfound] = 2.0 * L *  (x.value[i] * (integral - Nfound - 0.5) + x.value[i-1] * (Nfound + 0.5 - integral_prev))/(integral - integral_prev);
		  Ix2_found[Nfound] = 2.0 * L * x.Return_Value(rho.lambda[i]);
		  
		  Ix2_left = floor(Ix2_found[Nfound]);
		  Ix2_left -= (Ix2_left + N + 1)%2 ? 1 : 0;	//adjust parity
		  
		  Ix2_right = ceil(Ix2_found[Nfound]);
		  Ix2_right += (Ix2_right + N + 1)%2 ? 1 : 0;	//adjust parity
		  
		  int Ix2_in = (Ix2_found[Nfound] > 0 ? Ix2_left : Ix2_right);
		  int Ix2_out = (Ix2_found[Nfound] > 0 ? Ix2_right : Ix2_left);
		  cout << rho.lambda[i] << "\t" << x.Return_Value(rho.lambda[i]) << "\t" << Ix2_found[Nfound] << endl;
		  
		  //choose the saddle point state and remember the uncertain choices
		  if(Ix2_found[Nfound] - Ix2_left < 0.5) {
			state[0].Ix2[Nfound] = Ix2_left;
			state[1].Ix2[Nfound] = Ix2_left;
			state[2].Ix2[Nfound] = Ix2_left;
			state[3].Ix2[Nfound] = Ix2_left;
		  }
		  else if (Ix2_right - Ix2_found[Nfound] < 0.5) {
			state[0].Ix2[Nfound] = Ix2_right;
			state[1].Ix2[Nfound] = Ix2_right;
			state[2].Ix2[Nfound] = Ix2_right;
			state[3].Ix2[Nfound] = Ix2_right;
		  }
		  else { //it's a kind of magic: the zeroth goes in whle the first goes out, the second goes left if the third goes right
		  	state[0].Ix2[Nfound] = Ix2_in;
		  	state[1].Ix2[Nfound] = Ix2_out;
		  	state[2+change].Ix2[Nfound] = Ix2_left;
		  	state[2+!change].Ix2[Nfound] = Ix2_right;
		  	change = !change;   
		  	++n_moves;
		  }
		  Nfound++;
		}
      }
      //cout << "\ti = " << i << "\tintegral = " << integral << "\tNfound = " << Nfound << endl;
    }  
    //cout << endl;
  	
    //fix state[3] and state[4]
    for(int i=0; i<N-1; ++i) {
      if(state[2].Ix2[i] == state[2].Ix2[i+1]) {
	if(state[2].Ix2[i] != state[2].Ix2[i-1] + 2) state[2].Ix2[i] -= 2;
	else state[2].Ix2[i+1] += 2;
      }
      if(state[3].Ix2[i] == state[3].Ix2[i+1]) {
	if(state[3].Ix2[i] != state[3].Ix2[i-1] + 2) state[3].Ix2[i] -= 2;
	else state[3].Ix2[i+1] += 2;
      }
    }

    bool all_Diff = true;
    //check that all 'ns' states are different, we assume that Ix2 are ordered 
    for(int i=0; i<nstates; ++i) 
      for(int j=i+1; j<nstates; ++j) {
	if(state[i].Ix2 == state[j].Ix2) 
	  all_Diff = false;
      }

    if(!all_Diff) {
      //check if the first two states are different
      if(state[0].Ix2 == state[1].Ix2)
	JSCerror("Cannot create 4 (nor 2) different states in LiebLin_Diagonal_State_Ensemble. Consider increasing system size");
      else {
	nstates = 2;
	weight[0] = 0.5;
	weight[1] = 0.5;
      }
    }
    //set the weights accordingly to the distance between the states. 
//   	int n_moves_2 = 0, n_moves_3 = 0;
//   	for(int i=0; i < N; ++i) {
//   		if (!state[0].Ix2.is_in(state[2].Ix2[i])) ++n_moves_2;
//   		if (!state[0].Ix2.is_in(state[3].Ix2[i])) ++n_moves_3;
//   	}
//   	
//   	n_moves_2 = min(n_moves_2, n_moves - n_moves_2);
//   	n_moves_3 = min(n_moves_3, n_moves - n_moves_3);
//   	
//   	weight[2] *= 2.0*n_moves_2/n_moves;
//   	weight[3] *= 2.0*n_moves_3/n_moves;
//   	
//   	DP sum_weight = weight[0] + weight[1] + weight[2] + weight[3];
  	
    for(int i=0; i<nstates; ++i) {
      //weight[i] /= sum_weight;
      state[i].Set_Label_from_Ix2(state[0].Ix2);
      state[i].Compute_All(true);
    }
  	
	
    cout << weight[0] << "\t" << state[0].Ix2 << endl << weight[0] << "\t" << state[1].Ix2 << endl;// << state[2].Ix2 << endl << state[3].Ix2 << endl;
    
    return;
    

    //old working version but not enough to saturate +fsumrule
/*    
  	Root_Density x = rho;
  	for(int ix=0; ix<x.Npts; ++ix) {
  		x.value[ix] = x.lambda[ix];
  		for( int ip=0; ip<rho.Npts; ++ip) {
  			x.value[ix] += 2.0 * rho.value[ip] * rho.dlambda[ip] * atan ((x.lambda[ix] - rho.lambda[ip])/c_int);
  		}
  		x.value[ix] /= 2.0*PI; //normalization
  	}
  
    // Now carry on as per Discretized_LiebLin_Bethe_State:
    // Each time N \int_{-\infty}^\lambda d\lambda' \rho(\lambda') crosses a half integer, add a particle:
    DP integral = 0.0;
    DP integral_prev = 0.0;
    int Nfound = 0;
    Vect<DP> Ix2_found(0.0, N);
    int Ix2_left, Ix2_right;
    Vect<int> Ix2(N);
    Vect<int> Ix2_uncertain(0, N);
    Vect<int> index_uncertain(0, N);
    int n_uncertain = 0, n_states_raw = 1;
    
    for (int i = 0; i < rho.Npts; ++i) {
      integral_prev = integral;
      integral += L * rho.value[i] * rho.dlambda[i];
      //cout << x.value[i] << "\t" << integral << endl;
      if (integral > Nfound + 0.5) {
	// Subtle error: if the rho is too discontinuous, i.e. if more than one rapidity is found, must correct for this.
		if (integral > Nfound + 1.5 && integral < Nfound + 2.5) { // found two rapidities
		  JSCerror("The distribution of particles is too discontinous for discretisation");
		}
		
		else {
		  // few variables to clarify the computations, they do not need to be vectors, not used outside this loop
		  Ix2_found[Nfound] = 2.0 * L *  (x.value[i] * (integral - Nfound - 0.5) + x.value[i-1] * (Nfound + 0.5 - integral_prev))/(integral - integral_prev);
		  
		  Ix2_left = floor(Ix2_found[Nfound]);
		  Ix2_left -= (Ix2_left + N + 1)%2 ? 1 : 0;	//adjust parity
		  
		  Ix2_right = ceil(Ix2_found[Nfound]);
		  Ix2_right += (Ix2_right + N + 1)%2 ? 1 : 0;	//adjust parity
		  
		  //cout << Ix2_found[Nfound] << "\t";
		  
		  //choose the saddle point state and remember the uncertain choices
		  if(Ix2_found[Nfound] - Ix2_left < 1) {
			Ix2[Nfound] = Ix2_left;
			if(Ix2_found[Nfound] - Ix2_left > 0.75) {
				Ix2_uncertain[n_uncertain] = -1;
				index_uncertain[n_uncertain] = Nfound;
				++n_uncertain;
				n_states_raw *= 2;
			}
		  }
		  else if (Ix2_right - Ix2_found[Nfound] < 1) {
			Ix2[Nfound] = Ix2_right;
			if(Ix2_right - Ix2_found[Nfound] > 0.75) {
				Ix2_uncertain[n_uncertain] = 1;
				index_uncertain[n_uncertain] = Nfound;
				++n_uncertain;
				n_states_raw *= 2;
			}
		  }
		  else JSCerror("Cannot deduce a quantum number from x(lambda)");
		  Nfound++;
		}
      }
      //cout << "\ti = " << i << "\tintegral = " << integral << "\tNfound = " << Nfound << endl;
    }  
  	//cout << endl;
  	
  	//cout << Ix2 << endl;
  	
  	// create ensamble of states and compute its weights with respect to exact Ix2 saddle point
  	Vect<Vect<int> > Ix2states(Ix2, n_states_raw);
  	Vect<DP> Ix2weight(0.0, n_states_raw);
  	
// 	Ix2states[0] = Ix2;
//	Ix2weight[0] = ;
  	int n_states_proper = 0;
	for(int i=0; i < n_states_raw; ++i) {
		//only symmetric modifications
		for(int mod_index = 0; mod_index<n_uncertain; ++mod_index) {
			if((i & (1 << mod_index)) >> mod_index != 0) {
				Ix2states[n_states_proper][index_uncertain[mod_index]] += (-2)*Ix2_uncertain[mod_index];
				////Ix2states[state_index][N - 1 - hole_position[mod_index]] += (2)*holes[hole_position[mod_index]];
			}
		}
		//check if it's a proper set of quantum numbers
		bool OK = true;
		for(int j=0; j<N-1; ++j) if(Ix2states[n_states_proper][j] == Ix2states[n_states_proper][j+1]) OK = false;
		
		if(OK) {
			DP sum = 0;
			for(int j=0; j<N; ++j) sum += abs(Ix2_found[j] - Ix2states[n_states_proper][j]);
			Ix2weight[n_states_proper] = exp(-sum);
			n_states_proper++;
		}
		else Ix2states[n_states_proper] = Ix2;
	}
	
	
	
	//sort the states in increasing weight order;
	Vect<int> index(0, n_states_raw);
    for (int nrs = 0; nrs < n_states_raw; ++nrs) index[nrs] = nrs;
    Ix2weight.QuickSort(index);
    
    //cut the number of states if above 21 <- arbitrary number for now
    nstates = min(n_states_proper, 21);
    cout << n_states_raw << "\t" << n_states_proper << "\t" << nstates << endl;
    
    //create ensamble of states with normalised weights and ordered accordingly
	LiebLin_Bethe_State rhostate(c_int, N, L);
    rhostate.Ix2 = Ix2;
    rhostate.Compute_All(true);
     
	state = Vect<LiebLin_Bethe_State>(rhostate, nstates);
    weight = Vect<DP>(nstates);
    
    DP sum_weight = 0.0;
    for(int i=0; i<nstates; ++i) {
    	state[i].Ix2 = Ix2states[index[n_states_raw - i - 1]];
    	state[i].Set_Label_from_Ix2 (state[0].Ix2);
    	weight[i] = Ix2weight[n_states_raw - i - 1];
    	sum_weight += weight[i];
    }
    
    //renormalise
    for(int i=0; i<nstates; ++i) weight[i] /= sum_weight;
    
//    for(int i=0; i<nstates; ++i) {
//    	cout << weight[i] << "\t" << state[i].Ix2 << endl;
//    }    
    
    return;
*/
  	//cout << rho_t.value << endl;
  
  	// different attempts - to delete soon
//  	JSCerror("Stop here.");
/*
  
    // This function returns a state ensemble matching the continuous density rho.
    // The logic closely resembles the one used in Discretized_LiebLin_Bethe_State.

    // Now carry on as per Discretized_LiebLin_Bethe_State:
    // Each time N \int_{-\infty}^\lambda d\lambda' \rho(\lambda') crosses a half integer, add a particle:
    DP integral = 0.0;
    DP integral_prev = 0.0;
    int Nfound = 0;
    Vect<DP> lambda_found(0.0, 2*N);
    
    for (int i = 0; i < rho.Npts; ++i) {
      integral_prev = integral;
      integral += L * rho.value[i] * rho.dlambda[i];
      //cout << x.value[i] << "\t" << integral << endl;
      if (integral > Nfound + 0.5) {
      cout << L*x.value[i] << "\t" << integral << endl;
	// Subtle error: if the rho is too discontinuous, i.e. if more than one rapidity is found, must correct for this.
	if (integral > Nfound + 1.5 && integral < Nfound + 2.5) { // found two rapidities
	  lambda_found[Nfound++] = 0.25 * (3.0 * rho.lambda[i-1] + rho.lambda[i]);
	  lambda_found[Nfound++] = 0.25 * (rho.lambda[i-1] + 3.0 * rho.lambda[i]);
	}
	else {
	  //lambda_found[Nfound] = rho.lambda[i];
	  // Better: center the lambda_found between these points:
	  //lambda_found[Nfound] = rho.lambda[i-1] + (rho.lambda[i] - rho.lambda[i-1]) * ((Nfound + 1.0) - integral_prev)/(integral - integral_prev);
	  lambda_found[Nfound] = 0.5 * (rho.lambda[i-1] + rho.lambda[i]);
	  Nfound++;
	}
      }
      //cout << "\ti = " << i << "\tintegral = " << integral << "\tNfound = " << Nfound << endl;
    }
    //cout << "rho: " << rho.Npts << " points" << endl << rho.value << endl;
    //cout << "sym: " << rho.value[0] << " " << rho.value[rho.value.size() - 1] 
    // << "\t" << rho.value[rho.value.size()/2] << " " << rho.value[rho.value.size()/2 + 1] << endl;
    //cout << "Found " << Nfound << " particles." << endl;
    //cout << "lambda_found = " << lambda_found << endl;

    Vect<DP> lambda(N);
    // Fill up the found rapidities:
    for (int il = 0; il < JSC::min(N, Nfound); ++il) lambda[il] = lambda_found[il];
    // If there are missing ones, put them at the end; ideally, this should never be called
    for (int il = Nfound; il < N; ++il) lambda[il] = lambda_found[Nfound-1] + (il - Nfound + 1) * (lambda_found[Nfound-1] - lambda_found[Nfound-2]);

	//cout << lambda << endl;
*/	
	//try a different method
	//first determine bins, that is intervals in lambda which contain a single rapidity
/*	
	Vect<int> lambda_bin(0.0, N+1);
	integral = 0.0;
	Nfound = 0;
	
	lambda_bin[0] = 0;
	lambda_bin[N] = rho.Npts-1;
	for(int i=0; i<rho.Npts; ++i) {
		integral += L * rho.value[i] * rho.dlambda[i];
		if (integral > Nfound + 1.0) lambda_bin[++Nfound] = i - (rho.lambda[i]>0);
	}
	
	//put rapidity at the mean value of the distribution inside the bin
	for(int ib=0; ib < N; ++ib) {
		lambda_found[ib] = 0.0;
		for(int i=lambda_bin[ib]; i<lambda_bin[ib+1]; ++i) lambda_found[ib] += L * rho.value[i] * rho.lambda[i] * rho.dlambda[i];
	}
	
	//cout << lambda_found << endl;
	
	for(int i=0; i<N; ++i) cout << L*lambda[i]/PI << "\t" << L*lambda_found[i]/PI << endl;


	JSCerror("Stop here.");
*/
/*
    // Calculate quantum numbers: 2\pi * (L lambda + \sum_j 2 atan((lambda - lambda_j)/c)) = I_j
    // Use rounding.
    Vect<DP> Ix2_exact(N);
    for (int i = 0; i < N; ++i) {
      DP sum = 0.0;
      for (int j = 0; j < N; ++j) sum += 2.0 * atan((lambda[i] - lambda[j])/c_int);
      Ix2_exact[i] = (L * lambda[i] + sum)/PI;
	  Ix2_left[i] = floor(Ix2_exact[i]);
	  Ix2_left[i] -= (Ix2_left[i] + N + 1)%2 ? 1 : 0;
	  Ix2_right[i] = ceil(Ix2_exact[i]);
	  Ix2_right[i] += (Ix2_right[i] + N + 1)%2 ? 1 : 0;
	  
	  //cout << Ix2_left[i] << "\t" << Ix2_exact[i] << "\t" << Ix2_right[i] << endl;
      //Ix2[i] = 2.0* int((L * lambda[i] + sum)/twoPI) + (N % 2) - 1;
      // For N is even/odd, we want to round off to the nearest odd/even integer. 
      //Ix2[i] = 2.0 * floor((L* lambda[i] + sum)/twoPI + 0.5 * (N%2 ? 1 : 2)) + (N%2) - 1;
    }
    //cout << "Found quantum numbers " << endl << Ix2 << endl;
*/

/*
	//create the reference state and mark possible hole positions
	Vect<int> Ix2(N);
	Vect<float> holes(0.0, N);
	for(int i=0; i < N; ++i) {
		if(Ix2_found[i] - Ix2_left[i] < 1) { 
			Ix2[i] = Ix2_left[i];
				if(Ix2_found[i] - Ix2_left[i] > 0.75) holes[i] = -1;
		}
		else {
			Ix2[i] = Ix2_right[i];
			if(Ix2_right[i] - Ix2_found[i] > 0.75) holes[i] = 1;
		}
	}
*/		
//	cout << endl << Ix2 << endl;
//	cout << endl << holes << endl;
/*	
	//count number of possible states 
	int n = 0;
	for(int i=0; i < N; ++i) {
		if(holes[i] != 0) ++n;
	}
	int n_states_raw = 1;
*/	
/* only symmetric modifications
	for(int i=0; i<0.5*n; ++i) n_states_raw *= 2.0; //we count only symmetric modiications

	Vect<Vect<int> > Ix2states(Ix2, n_states_raw);
	
	Vect<int> hole_position(0, 0.5*n);
	int hole_index = 0;
	for(int i=N/2 + N%2; i<N; ++i) {
		if(holes[i] != 0) hole_position[hole_index++] = i;
	}
	
	
	//create modifications, we use bit representation of integers between 0 and n_states-1 treating zero as no change and 1 as mutltiplication by -1
	int state_index = 1;
	for(int i=1; i < n_states_raw; ++i) {
		//only symmetric modifications
		for(int mod_index = 0; mod_index<0.5*n; ++mod_index) {
			if((i & (1 << mod_index)) >> mod_index != 0) {
				Ix2states[state_index][hole_position[mod_index]] += (-2)*holes[hole_position[mod_index]];
				Ix2states[state_index][N - 1 - hole_position[mod_index]] += (2)*holes[hole_position[mod_index]];
			}
		}
		//check if it's a proper set of quantum numbers
		bool OK = true;
		for(int j=0; j<N-1; ++j) if(Ix2states[state_index][j] == Ix2states[state_index][j+1]) OK = false;
		
		if(OK) state_index++;
		else Ix2states[state_index] = Ix2;
	}
*/
/*
	for(int i=0; i<n; ++i) n_states_raw *= 2.0; //we count all posibilities

	Vect<Vect<int> > Ix2states(Ix2, n_states_raw);
	
	Vect<int> hole_position(0, n);
	int hole_index = 0;
	for(int i=0; i<N; ++i) {
		if(holes[i] != 0) hole_position[hole_index++] = i;
	}
	
	
	//create modifications, we use bit representation of integers between 0 and n_states-1 treating zero as no change and 1 as mutltiplication by -1
	int state_index = 1;
	for(int i=1; i < n_states_raw; ++i) {
		//only symmetric modifications
		for(int mod_index = 0; mod_index<n; ++mod_index) {
			if((i & (1 << mod_index)) >> mod_index != 0) {
				Ix2states[state_index][hole_position[mod_index]] += (-2)*holes[hole_position[mod_index]];
				//Ix2states[state_index][N - 1 - hole_position[mod_index]] += (2)*holes[hole_position[mod_index]];
			}
		}
		//check if it's a proper set of quantum numbers
		bool OK = true;
		for(int j=0; j<N-1; ++j) if(Ix2states[state_index][j] == Ix2states[state_index][j+1]) OK = false;
		
		if(OK) state_index++;
		else Ix2states[state_index] = Ix2;
	}
	
*/	
/*	
	n_states_raw = state_index;
//	for(int i=0; i<n_states; ++i) cout << Ix2states[i] << endl;
	
	LiebLin_Bethe_State rhostate(c_int, N, L);
    rhostate.Ix2 = Ix2;
    rhostate.Compute_All(true);

	Vect<LiebLin_Bethe_State> state_raw = Vect<LiebLin_Bethe_State>(rhostate, n_states_raw);
    Vect<DP> weight_raw(0.0, n_states_raw);
*/

/*    
    DP energy_rho = 0;
    for (int i = 0; i < rho.Npts; ++i) {
		energy_rho += L * rho.value[i] * rho.dlambda[i] * rho.lambda[i] * rho.lambda[i];
    }
    
    cout << energy_rho << endl;
    
    DP energy_discrete = 0;
    for (int i=0; i < N; ++i) {
    	energy_discrete += lambda[i] * lambda[i];
    }
    energy_discrete /= N;
    cout << energy_discrete;
*/
/*
	DP sum_weight_raw = 0.0;
    for(int i=0; i<n_states_raw; ++i) {
    	//state_raw[i].Ix2 = Ix2states[i];
    	//state_raw[i].Set_Label_from_Ix2 (state_raw[0].Ix2);
    	//state_raw[i].Compute_All(true);
    	DP sum = 0;
    	for(int j=0; j<N; ++j) sum += fabs(Ix2states[i][j] - Ix2_found[j]);
    	weight_raw[i] = exp(sum)/L;
    	
    	sum_weight_raw += weight_raw[i];
    }
    
    for(int i=0; i<n_states_raw; ++i) weight_raw[i] /= sum_weight_raw;
    
    // Order the weights in decreasing value:
    Vect<int> index(n_states_raw);
    for (int nrs = 0; nrs < n_states_raw; ++nrs) index[nrs] = nrs;
    weight_raw.QuickSort(index);
    
    //nstates = 0;
    //for(int i=0; i<n_states_raw; ++i) if (weight_raw[i] > 0.01) ++nstates;
    
    nstates = JSC::min(n_states_raw, 21);
    
    cout << nstates << endl;
    
	state = Vect<LiebLin_Bethe_State>(rhostate, nstates);
    weight = Vect<DP>(nstates);
    
    DP sum_weight = 0.0;
    for(int i=0; i<nstates; ++i) {
    	state[i].Ix2 = Ix2states[index[i]];
    	state[i].Set_Label_from_Ix2 (state[0].Ix2);
    	weight[i] = weight_raw[index[i]];
    	sum_weight += weight[i];
    }
    
    //renormalise
    for(int i=0; i<nstates; ++i) weight[i] /= sum_weight;
    
    for(int i=0; i<nstates; ++i) {
    	cout << weight[i] << "\t" << state[i].Ix2 << endl;
    }
*/
	//JSCerror("Stop here.");
	
	// end of different attempts
/*	
    // Check that the quantum numbers are all distinct:
    bool allOK = false;
    while (!allOK) {
      for (int i = 0; i < N-1; ++i) if (Ix2[i] == Ix2[i+1] && Ix2[i] < 0) Ix2[i] -= 2;
      for (int i = 1; i < N; ++i) if (Ix2[i] == Ix2[i-1] && Ix2[i] > 0) Ix2[i] += 2;
      allOK = true;
      for (int i = 0; i < N-1; ++i) if (Ix2[i] == Ix2[i+1]) allOK = false;
    }
    //cout << "Found modified quantum numbers " << endl << Ix2 << endl;

    LiebLin_Bethe_State rhostate(c_int, N, L);
    rhostate.Ix2 = Ix2;
    rhostate.Compute_All(true);
    //cout << "rapidities of state found: " << rhostate.lambda << endl;

    // Until here, the code is identical to that in Discretized_LiebLin_Bethe_State.
    // Construct states in the vicinity by going through type 2 descendents recursively:
    int ndesc_type2 = 0;
    int* ndesc_type2_ptr = &ndesc_type2;
    Generate_type_2_descendents (rhostate, ndesc_type2_ptr, rhostate);


    cout << "Found " << ndesc_type2 << " descendents for state " << rhostate << endl;
    //JSCerror("Stop here...");

    // Now try to construct states in the vicinity.
    int nrstates1mod = 1;
    for (int i = 0; i < N; ++i) {
      if (i == 0 || Ix2[i-1] < Ix2[i] - 2) nrstates1mod++;
      if (i == N-1 || Ix2[i+1] > Ix2[i] + 2) nrstates1mod++;
    }
    //if (nrstates1mod < nstates_req) JSCerror("nrstates1mod < nstates_req in LiebLin_Diagonal_State_Ensemble.");
    nstates = nrstates1mod;

    Vect<Vect<int> > Ix2states1mod(nrstates1mod);
    for (int nrs = 0; nrs < nrstates1mod; ++nrs) Ix2states1mod[nrs] = Vect<int> (N);

    int nfound = 0;
    Ix2states1mod[nfound++] = Ix2; // this is the sp state
    for (int i = 0; i < N; ++i) {
      if (i == 0 || Ix2[i-1] < Ix2[i] - 2) {
	Ix2states1mod[nfound] = Ix2;
	Ix2states1mod[nfound++][i] -= 2;
      }
      if (i == N-1 || Ix2[i+1] > Ix2[i] + 2) {
	Ix2states1mod[nfound] = Ix2;
	Ix2states1mod[nfound++][i] += 2;
      }
    }

    // Evaluate the weight of all found states:
    Vect<DP> rawweight(nrstates1mod);

    for (int nrs = 0; nrs < nrstates1mod; ++nrs) {
      DP sumweight = 0.0;
      for (int i = 0; i < N; ++i) sumweight += fabs(Ix2states1mod[nrs][i] - Ix2_exact[i]);
      rawweight[nrs] = exp(-sumweight/log(L));
    }

    // Order the weights in decreasing value:
    Vect<int> index(nrstates1mod);
    for (int nrs = 0; nrs < nrstates1mod; ++nrs) index[nrs] = nrs;
    rawweight.QuickSort(index);

    // Calculate weight normalization:
    DP weightnorm = 0.0;
    //for (int ns = 0; ns < nstates_req; ++ns) weightnorm += rawweight[nrstates1mod - 1 - ns];
    for (int ns = 0; ns < nstates; ++ns) weightnorm += rawweight[nrstates1mod - 1 - ns];

    state = Vect<LiebLin_Bethe_State>(rhostate, nstates);
    weight = Vect<DP> (nstates);

    for (int ns = 0; ns < nstates; ++ns) {
      weight[ns] = rawweight[nrstates1mod - 1 - ns]/weightnorm;
      state[ns].Ix2 = Ix2states1mod[index[nrstates1mod - 1 - ns] ];
      state[ns].Set_Label_from_Ix2 (rhostate.Ix2);
      state[ns].Compute_All(true);
    }
*/   


  }
  
  /*
  LiebLin_Diagonal_State_Ensemble::LiebLin_Diagonal_State_Ensemble (DP c_int, DP L, int N, const Root_Density& rho, int nstates_req)
    : nstates(nstates_req)
  {
    // This function returns a state ensemble matching the continuous density rho.
    // The logic closely resembles the one used in Discretized_LiebLin_Bethe_State.

    // Each `exact' (from rho) quantum number is matched to its two possible values,
    // assigning a probability to each according to the proximity of the quantum number values.

    // The set is then ordered in energy, and split into nstates_req boxes from which one
    // representative is selected, and given the probability = sum of probabilities in the box.


    // Begin as per Discretized_LiebLin_Bethe_State, to find the saddle-point state.
    // Each time N \int_{-\infty}^\lambda d\lambda' \rho(\lambda') crosses a half integer, add a particle:
    DP integral = 0.0;
    DP integral_prev = 0.0;
    int Nfound = 0;
    Vect<DP> lambda_found(0.0, 2*N);
    
    for (int i = 0; i < rho.Npts; ++i) {
      integral_prev = integral;
      integral += L * rho.value[i] * rho.dlambda[i];
      if (integral > Nfound + 0.5) {
	// Subtle error: if the rho is too discontinuous, i.e. if more than one rapidity is found, must correct for this.
	if (integral > Nfound + 1.5 && integral < Nfound + 2.5) { // found two rapidities
	  lambda_found[Nfound++] = 0.25 * (3.0 * rho.lambda[i-1] + rho.lambda[i]);
	  lambda_found[Nfound++] = 0.25 * (rho.lambda[i-1] + 3.0 * rho.lambda[i]);
	}
	else {
	  //lambda_found[Nfound] = rho.lambda[i];
	  // Better: center the lambda_found between these points:
	  //lambda_found[Nfound] = rho.lambda[i-1] + (rho.lambda[i] - rho.lambda[i-1]) * ((Nfound + 1.0) - integral_prev)/(integral - integral_prev);
	  lambda_found[Nfound] = 0.5 * (rho.lambda[i-1] + rho.lambda[i]);
	  Nfound++;
	}
      }
      //cout << "\ti = " << i << "\tintegral = " << integral << "\tNfound = " << Nfound << endl;
    }
    //cout << "rho: " << rho.Npts << " points" << endl << rho.value << endl;
    //cout << "sym: " << rho.value[0] << " " << rho.value[rho.value.size() - 1] 
    // << "\t" << rho.value[rho.value.size()/2] << " " << rho.value[rho.value.size()/2 + 1] << endl;
    //cout << "Found " << Nfound << " particles." << endl;
    //cout << "lambda_found = " << lambda_found << endl;

    Vect<DP> lambda(N);
    // Fill up the found rapidities:
    for (int il = 0; il < JSC::min(N, Nfound); ++il) lambda[il] = lambda_found[il];
    // If there are missing ones, put them at the end; ideally, this should never be called
    for (int il = Nfound; il < N; ++il) lambda[il] = lambda_found[Nfound-1] + (il - Nfound + 1) * (lambda_found[Nfound-1] - lambda_found[Nfound-2]);

    // Calculate quantum numbers: 2\pi * (L lambda + \sum_j 2 atan((lambda - lambda_j)/c)) = I_j
    // Use rounding.
    Vect<DP> Ix2_exact(N);
    Vect<int> Ix2(N);
    for (int i = 0; i < N; ++i) {
      DP sum = 0.0;
      for (int j = 0; j < N; ++j) sum += 2.0 * atan((lambda[i] - lambda[j])/c_int);
      Ix2_exact[i] = (L * lambda[i] + sum)/PI;
      //Ix2[i] = 2.0* int((L * lambda[i] + sum)/twoPI) + (N % 2) - 1;
      // For N is even/odd, we want to round off to the nearest odd/even integer. 
      Ix2[i] = 2.0 * floor((L* lambda[i] + sum)/twoPI + 0.5 * (N%2 ? 1 : 2)) + (N%2) - 1;
    }
    //cout << "Found quantum numbers " << endl << Ix2 << endl;

    // Check that the quantum numbers are all distinct:
    bool allOK = false;
    while (!allOK) {
      for (int i = 0; i < N-1; ++i) if (Ix2[i] == Ix2[i+1] && Ix2[i] < 0) Ix2[i] -= 2;
      for (int i = 1; i < N; ++i) if (Ix2[i] == Ix2[i-1] && Ix2[i] > 0) Ix2[i] += 2;
      allOK = true;
      for (int i = 0; i < N-1; ++i) if (Ix2[i] == Ix2[i+1]) allOK = false;
    }
    //cout << "Found modified quantum numbers " << endl << Ix2 << endl;

    LiebLin_Bethe_State rhostate(c_int, N, L);
    rhostate.Ix2 = Ix2;
    rhostate.Compute_All(true);
    //cout << "rapidities of state found: " << rhostate.lambda << endl;

    // Until here, the code is identical to that in Discretized_LiebLin_Bethe_State.

    // Now try to construct states in the vicinity.
    int nrstates1mod = 1;
    for (int i = 0; i < N; ++i) {
      if (i == 0 || Ix2[i-1] < Ix2[i] - 2) nrstates1mod++;
      if (i == N-1 || Ix2[i+1] > Ix2[i] + 2) nrstates1mod++;
    }
    if (nrstates1mod < nstates_req) JSCerror("nrstates1mod < nstates_req in LiebLin_Diagonal_State_Ensemble.");

    Vect<Vect<int> > Ix2states1mod(nrstates1mod);
    for (int nrs = 0; nrs < nrstates1mod; ++nrs) Ix2states1mod[nrs] = Vect<int> (N);

    int nfound = 0;
    Ix2states1mod[nfound++] = Ix2; // this is the sp state
    for (int i = 0; i < N; ++i) {
      if (i == 0 || Ix2[i-1] < Ix2[i] - 2) {
	Ix2states1mod[nfound] = Ix2;
	Ix2states1mod[nfound++][i] -= 2;
      }
      if (i == N-1 || Ix2[i+1] > Ix2[i] + 2) {
	Ix2states1mod[nfound] = Ix2;
	Ix2states1mod[nfound++][i] += 2;
      }
    }

    // Evaluate the weight of all found states:
    Vect<DP> rawweight(nrstates1mod);

    for (int nrs = 0; nrs < nrstates1mod; ++nrs) {
      DP sumweight = 0.0;
      for (int i = 0; i < N; ++i) sumweight += fabs(Ix2states1mod[nrs][i] - Ix2_exact[i]);
      rawweight[nrs] = exp(-sumweight/log(L));
    }

    // Order the weights in decreasing value:
    Vect<int> index(nrstates1mod);
    for (int nrs = 0; nrs < nrstates1mod; ++nrs) index[nrs] = nrs;
    rawweight.QuickSort(index);

    // Calculate weight normalization:
    DP weightnorm = 0.0;
    for (int ns = 0; ns < nstates_req; ++ns) weightnorm += rawweight[nrstates1mod - 1 - ns];

    state = Vect<LiebLin_Bethe_State>(rhostate, nstates_req);
    weight = Vect<DP> (nstates_req);

    for (int ns = 0; ns < nstates_req; ++ns) {
      weight[ns] = rawweight[nrstates1mod - 1 - ns]/weightnorm;
      state[ns].Ix2 = Ix2states1mod[index[nrstates1mod - 1 - ns] ];
      state[ns].Set_Label_from_Ix2 (rhostate.Ix2);
      state[ns].Compute_All(true);
    }
  }
  */

  LiebLin_Diagonal_State_Ensemble& LiebLin_Diagonal_State_Ensemble::operator= (const LiebLin_Diagonal_State_Ensemble& rhs)
  {
    if (this != &rhs) {
      nstates = rhs.nstates;
      state = rhs.state;
      weight = rhs.weight;
    }

    return(*this);
  }
  
  void LiebLin_Diagonal_State_Ensemble::Load (DP c_int, DP L, int N, const char* ensfile_Cstr)
  {
    ifstream infile(ensfile_Cstr);

    // Count nr of lines in file:
    string dummy;
    int nrlines = 0;
    while (getline(infile, dummy))  ++nrlines;
    infile.close();

    //cout << "Found " << nrlines << " lines in ens file." << endl;
    nstates = nrlines;
    LiebLin_Bethe_State examplestate (c_int, L, N);
    state = Vect<LiebLin_Bethe_State> (examplestate, nstates);
    weight = Vect<DP> (0.0, nstates);

    infile.open(ensfile_Cstr);

    string dummylabel;
    for (int ns = 0; ns < nstates; ++ns) {
      infile >> weight[ns] >> dummylabel;
      for (int i = 0; i < state[ns].N; ++i) infile >> state[ns].Ix2[i];
    }      
    infile.close();

    for (int ns = 0; ns < nstates; ++ns) {
      state[ns].Set_Label_from_Ix2 (state[0].Ix2); // labels are ref to the saddle-point state
      state[ns].Compute_All(true);
    }
  }

  void LiebLin_Diagonal_State_Ensemble::Save (const char* ensfile_Cstr)
  {
    ofstream outfile;

    outfile.open(ensfile_Cstr);

    for (int ns = 0; ns < nstates; ++ns) { 
      if (ns > 0) outfile << endl;
      outfile << setprecision(16) << weight[ns] << "\t" << state[ns].label << "\t";
      for (int i = 0; i < state[ns].N; ++i) outfile << " " << state[ns].Ix2[i];
    }
  }
  

  //LiebLin_Diagonal_State_Ensemble LiebLin_Thermal_Saddle_Point_Ensemble (DP c_int, DP L, int N, DP kBT, int nstates_req)
  LiebLin_Diagonal_State_Ensemble LiebLin_Thermal_Saddle_Point_Ensemble (DP c_int, DP L, int N, DP kBT)
  {
    // This function constructs a manifold of states around the thermal saddle-point.

    // Start as per Canonical_Saddle_Point_State:

    // This function returns the discretized state minimizing the canonical free energy
    // F = E - T S.

    // This is obtained by rediscretizing the solution coming from TBA.


    // Otherwise, return the discretized TBA saddle-point state:
    LiebLin_TBA_Solution TBAsol = LiebLin_TBA_Solution_fixed_nbar (c_int, N/L, kBT, 1.0e-4, 100);

    LiebLin_Diagonal_State_Ensemble ensemble(c_int, L, N, TBAsol.rho);
    
    cout << "nbar: " << TBAsol.nbar << endl;
    cout << "ebar: " << TBAsol.ebar << endl;

    cout << ensemble.state[0].E << endl;
    //return(LiebLin_Diagonal_State_Ensemble (c_int, L, N, TBAsol.rho, nstates_req));
    return(ensemble);
  }

} // namespace JSC






