/**********************************************************

This software is part of J.-S. Caux's ABACUS library.

Copyright (c).

-----------------------------------------------------------

File:  LiebLin_ln_Overlap.cc

Purpose:  Calculates Nikita Slavnov's determinant, case RHS is Bethe
          and LHS is arbitrary

***********************************************************/

#include "JSC.h"

using namespace std;
using namespace JSC;

namespace JSC {
  /*
  complex<DP> Fn_V (int j, int sign, Vect<complex<DP> >& lstate_lambdaoc, LiebLin_Bethe_State& rstate)
  {
    complex<DP> result_num = 1.0;
    complex<DP> result_den = 1.0;

    complex<DP> signcx = complex<DP>(sign);

    for (int m = 0; m < rstate.N; ++m) {
      result_num *= (lstate_lambdaoc[m] - rstate.lambdaoc[j] + signcx * II);
      result_den *= (rstate.lambdaoc[m] - rstate.lambdaoc[j] + signcx * II);
    }

    return(result_num/result_den);
  }
  */

  complex<DP> LiebLin_ln_Overlap (Vect<DP> lstate_lambdaoc, DP lstate_lnnorm, LiebLin_Bethe_State& rstate)
  {
    Vect<complex<DP> > lstate_lambdaoc_CX(rstate.N);
    for (int i = 0; i < rstate.N; ++i) lstate_lambdaoc_CX[i] = complex<DP>(lstate_lambdaoc[i]);
    return(LiebLin_ln_Overlap(lstate_lambdaoc_CX, lstate_lnnorm, rstate));
  }

  complex<DP> LiebLin_ln_Overlap (Vect<complex<DP> > lstate_lambdaoc, DP lstate_lnnorm, LiebLin_Bethe_State& rstate)
  {
    // Computes the log of the overlap between the left state and the Bethe rstate

    // IMPORTANT ASSUMPTIONS:
    // The length over which the overlap is calculated is that in rstate

    if (lstate_lambdaoc.size() != rstate.N) {
      cout << "Caution:  calling overlap of states with difference charges.  Are you sure that's what you want ?" << endl;
      return(0.0);
    }

    // \mu are rstate rapidities, \lambda are from lstate (not BE)

    Vect_CX ln_prod_plus(0.0, rstate.N);  // contains \prod_a (\mu_a - lambdaoc_k + II)
    Vect_CX ln_prod_minus(0.0, rstate.N);  // contains \prod_a (\mu_a - \lambdaoc_k - II)
    Vect_CX ln_prod_ll_plus (0.0, rstate.N); // contains \prod_{a \neq k} (\lambdaoc_a - \lambdaoc_k + II)
    Vect_CX ln_prod_ll_minus (0.0, rstate.N); // contains \prod_{a \neq k} (\lambdaoc_a - \lambdaoc_k - II)

    for (int j = 0; j < rstate.N; ++j) 
      for (int k = 0; k < rstate.N; ++k) {
	ln_prod_plus[k] += log(rstate.lambdaoc[j] - lstate_lambdaoc[k] + II);
	ln_prod_minus[k] += log(rstate.lambdaoc[j] - lstate_lambdaoc[k] - II);
	if (j != k) {
	  ln_prod_ll_plus[k] += log(lstate_lambdaoc[j] - lstate_lambdaoc[k] + II);
	  ln_prod_ll_minus[k] += log(lstate_lambdaoc[j] - lstate_lambdaoc[k] - II);
	}
      }

    // Build the matrix:
    SQMat_CX Omega (0.0, rstate.N);

    for (int j = 0; j < rstate.N; ++j) 
      for (int k = 0; k < rstate.N; ++k)
	//Omega[j][k] = exp(-II * rstate.cxL * lstate_lambdaoc[k] + ln_prod_plus[k] - ln_prod_ll_plus[k]) 
	//* (II/((rstate.lambdaoc[j] - lstate_lambdaoc[k])*(rstate.lambdaoc[j] - lstate_lambdaoc[k] + II)))
	//- (II/((rstate.lambdaoc[j] - lstate_lambdaoc[k])*(rstate.lambdaoc[j] - lstate_lambdaoc[k] - II)))
	//* exp(ln_prod_minus[k] - ln_prod_ll_plus[k]);
	Omega[j][k] = exp(-II * rstate.cxL * lstate_lambdaoc[k] - ln_prod_plus[k] + ln_prod_minus[k])
	  * (II/((rstate.lambdaoc[j] - lstate_lambdaoc[k])*(rstate.lambdaoc[j] - lstate_lambdaoc[k] - II)))
	  - (II/((rstate.lambdaoc[j] - lstate_lambdaoc[k])*(rstate.lambdaoc[j] - lstate_lambdaoc[k] + II)));


    //for (int j = 0; j < rstate.N; ++j) {
    //for (int k = 0; k < rstate.N; ++k) 
    //cout << Omega[j][k] << "\t";
    //cout << endl;
    //}


    complex<DP> lndetOmega = lndet_LU_CX_dstry(Omega);

    //cout << "lndetOmega = " << lndetOmega << endl;

    // Prefactors:
    complex<DP> ln_prod_d_mu = II * 0.5 * rstate.cxL * rstate.lambdaoc.sum();
    complex<DP> ln_prod_d_lambdaoc = II * 0.5 * rstate.cxL * lstate_lambdaoc.sum();

    complex<DP> ln_prod_mu = 0.0;
    complex<DP> ln_prod_lambdaoc = 0.0;
    complex<DP> ln_prod_plusminus = 0.0;
    for (int j = 0; j < rstate.N - 1; ++j)
      for (int k = j+1; k < rstate.N; ++k) {
	ln_prod_mu += log(rstate.lambdaoc[k] - rstate.lambdaoc[j]);
	//ln_prod_lambdaoc += log((lstate_lambdaoc[j] - lstate_lambdaoc[k] + II) * (lstate_lambdaoc[j] - lstate_lambdaoc[k] - II)/(lstate_lambdaoc[k] - lstate_lambdaoc[j]));
	ln_prod_lambdaoc += log(lstate_lambdaoc[k] - lstate_lambdaoc[j]);
      }

    for (int j = 0; j < rstate.N; ++j)
      for (int k = 0; k < rstate.N; ++k)
	ln_prod_plusminus += log((rstate.lambdaoc[j] - lstate_lambdaoc[k] + II));

    //cout << "ln_prod_mu " << ln_prod_mu << "\tln_prod_lambdaoc " << ln_prod_lambdaoc << "\tln_prod_plusminus " << ln_prod_plusminus 
    // << "\texp1 " << exp(-ln_prod_mu - ln_prod_lambdaoc) << "\texp2 " << exp(ln_prod_plusminus) << endl;

    //if (real(ln_prod_d_mu + ln_prod_d_lambdaoc - ln_prod_mu + ln_prod_lambdaoc + lndetOmega - 0.5 * (lstate_lnnorm + rstate.lnnorm)) > 10.0) {
    //cout << ln_prod_d_mu << "\t" << ln_prod_d_lambdaoc << "\t" << -ln_prod_mu << "\t" << ln_prod_lambdaoc << "\t" << lndetOmega << "\t" << -0.5 * lstate_lnnorm << "\t" << -0.5 * rstate.lnnorm << endl;      cout << ln_prod_d_mu + ln_prod_d_lambdaoc - ln_prod_mu + ln_prod_lambdaoc + lndetOmega - 0.5 * (lstate_lnnorm + rstate.lnnorm) << endl;
    //JSCerror("Overlap exceeds 1.");
    //}

    //return(ln_prod_d_mu + ln_prod_d_lambdaoc - ln_prod_mu + ln_prod_lambdaoc + lndetOmega - 0.5 * (lstate_lnnorm + rstate.lnnorm));
    return(ln_prod_d_mu + ln_prod_d_lambdaoc - ln_prod_mu - ln_prod_lambdaoc + ln_prod_plusminus + lndetOmega - 0.5 * (lstate_lnnorm + rstate.lnnorm));
  }

  /*
  // Incorrect version
  complex<DP> LiebLin_ln_Overlap (Vect<complex<DP> > lstate_lambdaoc, DP lstate_lnnorm, LiebLin_Bethe_State& rstate)
  {
    // Computes the log of the overlap between the left state and the Bethe rstate

    // If momentum difference is zero but states are different, then form factor is zero:

    SQMat_CX one_plus_U (0.0, rstate.N);

    Vect_CX Vplus (0.0, rstate.N);  // contains V^+_j
    Vect_CX Vminus (0.0, rstate.N);  // contains V^-_j
    Vect_CX Fn_Prod (0.0, rstate.N);  // product_{m\neq j} (\mu_m - \lambdaoc_j)/(\lambdaoc_m - \lambdaoc_j)
    //Vect_DP rKern (0.0, rstate.N);   // K(lambdaoc_j - lambdaoc_p)

    //int p = 0;

    complex<DP> Kout = lstate_lambdaoc.sum() - rstate.K;

    for (int a = 0; a < rstate.N; ++a) {
      Vplus[a] = Fn_V (a, 1, lstate_lambdaoc, rstate);
      Vminus[a] = Fn_V (a, -1, lstate_lambdaoc, rstate);
      Fn_Prod[a] = 1.0;
      for (int m = 0; m < rstate.N; ++m) 
	if (m != a) Fn_Prod[a] *= (lstate_lambdaoc[m] - rstate.lambdaoc[a])/(rstate.lambdaoc[m] - rstate.lambdaoc[a]);
      //rKern[a] = rstate.Kernel (a, p);
    }

    for (int a = 0; a < rstate.N; ++a) 
      for (int b = 0; b < rstate.N; ++b) 
	one_plus_U[a][b] = (a == b ? 1.0 : 0.0) + II * ((lstate_lambdaoc[a] - rstate.lambdaoc[a])/(Vplus[a] - Vminus[a]))
	  * Fn_Prod[a] * rstate.Kernel(a,b);

    complex<DP> ln_ddalpha_sigma = lndet_LU_CX_dstry(one_plus_U);

    complex<DP> ln_prod_V = 0.0;
    for (int a = 0; a < rstate.N; ++a) ln_prod_V += log(Vplus[a] - Vminus[a]);
    
    complex<DP> ln_prod_2 = 0.0;
    for (int a = 0; a < rstate.N; ++a) 
      for (int b = 0; b < rstate.N; ++b) 
	ln_prod_2 += log((rstate.lambdaoc[a] - rstate.lambdaoc[b] + II)/(lstate_lambdaoc[a] - rstate.lambdaoc[b]));

    //cout << endl << ln_ddalpha_sigma << "\t" << ln_prod_V << "\t" << ln_prod_2 << "\t" << - log(2.0 * II * imag(Vplus[p])) << endl;
    cout << endl << ln_ddalpha_sigma << "\t" << ln_prod_V << "\t" << ln_prod_2 << "\t" << 0.5 * lstate_lnnorm << "\t" << 0.5 * rstate.lnnorm << endl;//- log(Vplus[p] - Vminus[p]) << endl;
      
    ln_ddalpha_sigma += ln_prod_V + ln_prod_2;// - log(Vplus[p] - Vminus[p]);

    //cout << "shift = " << (complex<DP>(rstate.N) * (lstate_lambdaoc[0] - rstate.lambdaoc[0])/twoPI) << "\tKout = " << Kout << "\texp(-II*Kout) = " << exp(-II * Kout)
    // << "\tlog(exp(-II * Kout) - 1.0) = " << log(exp(-II * Kout) - 1.0) << endl;
    //cout << ln_ddalpha_sigma << "\t" << ln_prod_V << "\t" << ln_prod_2 << "\t" << - log(Vplus[p] - Vminus[p]) << "\t" << lstate_lnnorm << endl << endl;

    return (ln_ddalpha_sigma - 0.5 * (lstate_lnnorm + rstate.lnnorm));
  }
  */

} // namespace JSC
