/**********************************************************

This software is part of J.-S. Caux's ABACUS++ library.

Copyright (c) 

-----------------------------------------------------------

File:  src/ODSLF/ODSLF_Sumrules.cc

Purpose:  defines sumrule factors for spinless fermions related to Heisenberg


***********************************************************/

#include "JSC.h"

using namespace JSC;
using namespace std;

namespace JSC {


  DP ODSLF_X_avg (char xyorz, DP Delta, int N, int M)
  {
    // Calculates \sum_j < S_j^a S_{j+1}^a >, a = x, y or z.

    DP eps_Delta = 0.00000001;
    
    // Define the chain:  J, Delta, h, Nsites
    Heis_Chain chain(1.0, Delta, 0.0, N);
    
    // Define the base:  chain, Mdown
    ODSLF_Base gbase(chain, M);

    // Define the chain:  J, Delta, h, Nsites
    Heis_Chain chain2(1.0, Delta + eps_Delta, 0.0, N);
    
    // Define the base:  chain, Mdown
    ODSLF_Base gbase2(chain2, M);
    
    DP E0_Delta = 0.0;
    DP E0_Delta_eps = 0.0; 

    if (Delta > 0.0 && Delta < 1.0) {

      // Define the ground state
      ODSLF_XXZ_Bethe_State gstate(chain, gbase);
    
      // Compute everything about the ground state
      gstate.Compute_All(true);

      E0_Delta = gstate.E;

      // Define the ground state
      ODSLF_XXZ_Bethe_State gstate2(chain2, gbase2);
      
      // Compute everything about the ground state
      gstate2.Compute_All(true);

      E0_Delta_eps = gstate2.E;
    }
    /*
    else if (Delta == 1.0) {
      // Define the ground state
      XXX_Bethe_State gstate(chain, gbase);
    
      // Compute everything about the ground state
      gstate.Compute_All(true);

      E0_Delta = gstate.E;

      // Define the ground state
      XXZ_gpd_Bethe_State gstate2(chain2, gbase2);  // need XXZ_gpd here
    
      // Compute everything about the ground state
      gstate2.Compute_All(true);

      E0_Delta_eps = gstate2.E;
    }

    else if (Delta > 1.0) {
      // Define the ground state
      XXZ_gpd_Bethe_State gstate(chain, gbase);
    
      // Compute everything about the ground state
      gstate.Compute_All(true);

      E0_Delta = gstate.E;

      // Define the ground state
      XXZ_gpd_Bethe_State gstate2(chain2, gbase2);
    
      // Compute everything about the ground state
      gstate2.Compute_All(true);

      E0_Delta_eps = gstate2.E;
    }
    */
    else JSCerror("Wrong anisotropy in ODSLF_S1_sumrule_factor.");

    DP answer = 0.0;
    if (xyorz == 'x' || xyorz == 'y') answer = 0.5 * (E0_Delta - Delta * (E0_Delta_eps - E0_Delta)/eps_Delta);

    // Careful for z !  Hamiltonian defined as S^z S^z - 1/4, so add back N/4:
    else if (xyorz == 'z') answer = (E0_Delta_eps - E0_Delta)/eps_Delta + 0.25 * N;

    else JSCerror("option not implemented in ODSLF_X_avg.");

    return(answer);
  }

  DP ODSLF_S1_sumrule_factor (char mporz, DP Delta, int N, int M, int iK)
  {

    DP X_x = ODSLF_X_avg ('x', Delta, N, M);
    DP X_z = ODSLF_X_avg ('z', Delta, N, M);

    DP sumrule = 0.0;

    if (mporz == 'm' || mporz == 'p') sumrule = - 2.0 * ((1.0 - Delta * cos((twoPI * iK)/N)) * X_x + (Delta - cos((twoPI * iK)/N)) * X_z)/N;
    //if (mporz == 'm' || mporz == 'p') sumrule = - 1.0 * ((1.0 - Delta * cos((twoPI * iK)/N)) * 2.0 * X_x + (Delta - cos((twoPI * iK)/N)) * (X_x + X_z))/N;

    else if (mporz == 'z') sumrule = iK == 0 ? 1.0 : -2.0 * X_x * (1.0 - cos((twoPI * iK)/N))/N;

    else if (mporz == 'a') sumrule = 1.0;
    else if (mporz == 'b') sumrule = 1.0;

    else JSCerror("option not implemented in ODSLF_S1_sumrule_factor.");

    //return(1.0/sumrule);
    return(1.0/(sumrule + 1.0e-16)); // sumrule is 0 for iK == 0 or N
  }

  DP ODSLF_S1_sumrule_factor (char mporz, DP Delta, int N, DP X_x, DP X_z, int iK)
  {

    //DP X_x = X_avg ('x', Delta, N, M);
    //DP X_z = X_avg ('z', Delta, N, M);

    DP sumrule = 0.0;

    if (mporz == 'm' || mporz == 'p') sumrule = - 2.0 * ((1.0 - Delta * cos((twoPI * iK)/N)) * X_x + (Delta - cos((twoPI * iK)/N)) * X_z)/N;
    //if (mporz == 'm' || mporz == 'p') sumrule = - 1.0 * ((1.0 - Delta * cos((twoPI * iK)/N)) * 2.0 * X_x + (Delta - cos((twoPI * iK)/N)) * (X_x + X_z))/N;

    else if (mporz == 'z') sumrule = -2.0 * X_x * (1.0 - cos((twoPI * iK)/N))/N;

    else if (mporz == 'a') sumrule = 1.0;
    else if (mporz == 'b') sumrule = 1.0;

    else JSCerror("option not implemented in ODSLF_S1_sumrule_factor.");

    //return(1.0/sumrule);
    return(1.0/(sumrule + 1.0e-16)); // sumrule is 0 for iK == 0 or N
  }

  //DP Sumrule_Factor (char whichDSF, Heis_Bethe_State& RefState, DP Chem_Pot, bool fixed_iK, int iKneeded)
  DP Sumrule_Factor (char whichDSF, ODSLF_Bethe_State& RefState, DP Chem_Pot, int iKmin, int iKmax)
  {
    DP sumrule_factor = 1.0;
    //if (!fixed_iK) {
    if (iKmin != iKmax) {
      if (whichDSF == 'Z') sumrule_factor = 1.0;
      else if (whichDSF == 'm') 
	sumrule_factor = 1.0/RefState.base.Mdown;
      else if (whichDSF == 'z') sumrule_factor = 1.0/(0.25 * RefState.chain.Nsites);
      else if (whichDSF == 'p') sumrule_factor = 1.0/(RefState.chain.Nsites - RefState.base.Mdown);
      else if (whichDSF == 'a') sumrule_factor = 1.0;
      else if (whichDSF == 'b') sumrule_factor = 1.0;
      else if (whichDSF == 'q') sumrule_factor = 1.0;

      else JSCerror("whichDSF option not consistent in Sumrule_Factor");
    }
    //else if (fixed_iK) {
    else if (iKmin == iKmax) {
      if (whichDSF == 'Z') sumrule_factor = 1.0;
      else if (whichDSF == 'm' || whichDSF == 'z' || whichDSF == 'p') 
	//sumrule_factor = S1_sumrule_factor (whichDSF, RefState.chain.Delta, RefState.chain.Nsites, RefState.base.Mdown, iKneeded);
	sumrule_factor = ODSLF_S1_sumrule_factor (whichDSF, RefState.chain.Delta, RefState.chain.Nsites, RefState.base.Mdown, iKmax);
      else if (whichDSF == 'a') sumrule_factor = 1.0;
      else if (whichDSF == 'b') sumrule_factor = 1.0;
      else if (whichDSF == 'q') sumrule_factor = 1.0;
      
      else JSCerror("whichDSF option not consistent in Sumrule_Factor");
    }

    

    return(sumrule_factor);
  }
  
  //void Evaluate_F_Sumrule (char whichDSF, const ODSLF_Bethe_State& RefState, DP Chem_Pot, int iKmin, int iKmax, const char* FFsq_Cstr, const char* FSR_Cstr)
  void Evaluate_F_Sumrule (string prefix, char whichDSF, const ODSLF_Bethe_State& RefState, DP Chem_Pot, int iKmin, int iKmax)
  {

    stringstream RAW_stringstream;    string RAW_string;
    RAW_stringstream << prefix << ".raw";    
    RAW_string = RAW_stringstream.str();    const char* RAW_Cstr = RAW_string.c_str();

    stringstream FSR_stringstream;    string FSR_string;
    FSR_stringstream << prefix << ".fsr";
    FSR_string = FSR_stringstream.str();    const char* FSR_Cstr = FSR_string.c_str();

    ifstream infile;
    infile.open(RAW_Cstr);
    if(infile.fail()) JSCerror("Could not open input file in Evaluate_F_Sumrule(ODSLF...).");

    // We run through the data file to chech the f sumrule at each positive momenta:
    //Vect<DP> Sum_omega_FFsq(0.0, RefState.chain.Nsites/2 + 1);  // 
    Vect<DP> Sum_omega_FFsq(0.0, iKmax - iKmin + 1);  // 

    DP omega, FF;
    int iK, conv;
    long long int base_id, type_id, id;

    while (infile.peek() != EOF) {
      infile >> omega >> iK >> FF >> conv >> base_id >> type_id >> id;
      //if (iK > 0 && iK <= RefState.chain.Nsites/2) Sum_omega_FFsq[iK] += omega * FFsq;
      if (iK >= iKmin && iK <= iKmax) Sum_omega_FFsq[iK - iKmin] += omega * FF * FF;
    }

    infile.close();

    ofstream outfile;
    outfile.open(FSR_Cstr);
    outfile.precision(16);

    DP X_x = X_avg ('x', RefState.chain.Delta, RefState.chain.Nsites, RefState.base.Mdown);
    DP X_z = X_avg ('z', RefState.chain.Delta, RefState.chain.Nsites, RefState.base.Mdown);

    /*
    outfile << 0 << "\t" << Sum_omega_FFsq[0] * S1_sumrule_factor (whichDSF, RefState.chain.Delta, RefState.chain.Nsites, X_x, X_z, 0);  
    for (int i = 1; i <= RefState.chain.Nsites/2; ++i)
      outfile << endl << i << "\t" << Sum_omega_FFsq[i] * S1_sumrule_factor (whichDSF, RefState.chain.Delta, RefState.chain.Nsites, X_x, X_z, i);
    */

    for (int i = iKmin; i <= iKmax; ++i) {
      if (i > iKmin) outfile << endl;
      outfile << i << "\t" << Sum_omega_FFsq[i] * ODSLF_S1_sumrule_factor (whichDSF, RefState.chain.Delta, RefState.chain.Nsites, X_x, X_z, i);
    }

    outfile.close();
  }

} // namespace JSC
