/**********************************************************

This software is part of J.-S. Caux's ABACUS library.

Copyright (c).

-----------------------------------------------------------

File:  Descendents.cc

Purpose:  New implementation of state scanning:  
          functions to descend down hierarchy of states.
          
***********************************************************/

#include "JSC.h"

using namespace std;
using namespace JSC;

namespace JSC {


  Vect<string> Descendent_States_with_iK_Stepped_Up (string ScanIx2_label, const Vect<Vect<int> >& OriginIx2, const Vect<Vect<int> >& BaseScanIx2, const Vect<int>& Ix2_min, const Vect<int>& Ix2_max, bool disperse_only_current_exc, bool preserve_nexc)
  {
    //cout << "Up start" << endl;
    // Given an OriginIx2 and a BaseScanIx2 (by which we mean a set of Ix2 providing a basis for scanning),
    // this function returns a vector of the labels of the states obtained by each allowable 
    // one-step increase of the quantum numbers.

    // Rules for moving quantum numbers:
    // - quantum numbers preserve their ordering (an Umklapp is thus made of many one-step movements)
    // - if a quantum number has moved as compared to BaseScanIx2, then it can only move further in the same direction

    // The ordering rule for allowable increases is that we start from the topmost, rightmost quantum number.
    // The increases can thus be applied to quantum numbers below the bottom-most, left-most right-displaced quantum nr.

    // The preserve_nexc boolean flags whether we maintain or increase the nr of ph pairs.
    // This is only checked if disperse_only_current_exc == false (otherwise it's irrelevant).

    Vect<Vect<int> > ScanIx2 = Return_Ix2_from_Label (ScanIx2_label, OriginIx2);

    // Determine the level and index of the bottom-most left-most right-moving quantum number sits:
    int exclevel = -1;
    int excindex = 0;
    bool excfound = false;
    do {
      exclevel++;
      if (exclevel == ScanIx2.size()) { // there isn't a single right-moving quantum number in ScanIx2
	break; 
      }
      for (int alpha = 0; alpha < ScanIx2[exclevel].size(); ++alpha) 
	if (ScanIx2[exclevel][alpha] > BaseScanIx2[exclevel][alpha]) {
	  excindex = alpha;
	  excfound = true;
	  break;
	}
    } while (!excfound);
    // If we haven't found an excitation, then exclevel == ScanIx2.size() and excindex = 0;

    //cout << "exclevel = " << exclevel << endl;
    //cout << "excindex = " << excindex << endl;

    // The quantum numbers which we can move right are thus those 
    // with (j < exclevel) and (j == exclelev and alpha <= excindex)

    int ndesc_possible = 1;
    if (!disperse_only_current_exc) {
      ndesc_possible = 0;
      for (int j = 0; j <= JSC::min(exclevel, ScanIx2.size() -1); ++j) ndesc_possible += ScanIx2[j].size();
    }

    // Now construct the actual descendents:
    Vect<string> desclabelfound (ndesc_possible);
    int ndesc_found = 0;

    if (!disperse_only_current_exc && !preserve_nexc) 
      // List descendents with a new excitation at a lower level:
      // this can only create an additional ph pair
      for (int j = 0; j < exclevel; ++j)
	for (int alpha = 0; alpha < ScanIx2[j].size(); ++alpha) {
	  if (ScanIx2[j][alpha] >= BaseScanIx2[j][alpha] // candidate either undisplaced or right-moving
	      && !ScanIx2[j].includes(ScanIx2[j][alpha] + 2) // there is a hole to its right
	      && ScanIx2[j][alpha] + 2 <= Ix2_max[j]) { // this new quantum number fits within the limits
	    // We've found a descendent
	    ScanIx2[j][alpha] += 2;
	    desclabelfound[ndesc_found] = Return_State_Label (ScanIx2, OriginIx2);
	    ScanIx2[j][alpha] -= 2;
	    ndesc_found++;
	  }
	}
    
    // List descendents with a new excitation at exclevel:
    if (exclevel < ScanIx2.size()) { // excfound == true, excindex is now guaranteed < ScanIx2[exclevel].size()
      int alphamin = (disperse_only_current_exc ? excindex : 0);
      int alphamax = (disperse_only_current_exc ? excindex : excindex - 1);
      //for (int alpha = 0; alpha <= excindex; ++alpha) {
      for (int alpha = alphamin; alpha <= alphamax; ++alpha) {
	if (ScanIx2[exclevel][alpha] >= BaseScanIx2[exclevel][alpha]
	    && !ScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] + 2)
	    && ScanIx2[exclevel][alpha] + 2 <= Ix2_max[exclevel]) {
	  // We've found a descendent
	  ScanIx2[exclevel][alpha] += 2;
	  desclabelfound[ndesc_found] = Return_State_Label (ScanIx2, OriginIx2);
	  ScanIx2[exclevel][alpha] -= 2;
	  // If we're dispersing a subleading Ix2, check whether we match the preserve_nexc condition:
	  // The descendent is acceptable if disperse_only_current_exc == true,
	  // or if preserve_nexc == true and nexc labels match,
	  // or if preserve_nexc == false and nexc labels don't match:
	  //if (!disperse_only_current_exc && (BaseScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] + 2) != (Extract_nexc_Label(desclabelfound[ndesc_found]).compare(Extract_nexc_Label(ScanIx2_label)) == 0))) JSCerror("Inconsistency in check in Descendents");
	  if (disperse_only_current_exc 
	      //|| (preserve_nexc == BaseScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] + 2)))
	      || (preserve_nexc == (Extract_nexc_Label(desclabelfound[ndesc_found]).compare(Extract_nexc_Label(ScanIx2_label)) == 0)))
	    ndesc_found++;
	}
      }
    } //if (exclevel < ScanIx2.size())

    // Resize desc:
    Vect<string> desclabelfound_resized(ndesc_found);
    for (int idesc = 0; idesc < ndesc_found; ++idesc)
      desclabelfound_resized[idesc] = desclabelfound[idesc];

    //cout << "Up done" << endl;

    //return(descIx2found);
    //return(desclabelfound);
    return(desclabelfound_resized);
  }


  Vect<string> Descendent_States_with_iK_Stepped_Down (string ScanIx2_label, const Vect<Vect<int> >& OriginIx2, const Vect<Vect<int> >& BaseScanIx2, const Vect<int>& Ix2_min, const Vect<int>& Ix2_max, bool disperse_only_current_exc, bool preserve_nexc)
  {
    //cout << "Down start" << endl;
    // Given an OriginIx2 and a BaseScanIx2 (by which we mean a set of Ix2 providing a basis for scanning),
    // this function returns a vector of the labels of the states obtained by each allowable 
    // one-step decrease of the quantum numbers.

    // Rules for moving quantum numbers:
    // - quantum numbers preserve their ordering (an Umklapp is thus made of many one-step movements)
    // - if a quantum number has moved as compared to BaseScanIx2, then it can only move further in the same direction

    // The ordering rule for allowable decreases is that we start from the topmost, leftmost quantum number.
    // The decreases can thus be applied to quantum numbers below the bottom-most, right-most left-displaced quantum nr.

    // The preserve_nexc boolean flags whether we maintain or increase the nr of ph pairs.
    // This is only checked if disperse_only_current_exc == false (otherwise it's irrelevant).

    Vect<Vect<int> > ScanIx2 = Return_Ix2_from_Label (ScanIx2_label, OriginIx2);

    // Determine the level and index of the bottom-most right-most left-moving quantum number sits:
    int exclevel = -1;
    int excindex = 0;
    bool excfound = false;

    //cout << "Looking for exclevel and excindex for " << endl << "\tBaseIx2 = " << BaseScanIx2 << endl << "\tScanIx2 = " << ScanIx2 << endl;
    do {
      exclevel++;
      if (exclevel == ScanIx2.size()) { // there isn't a single left-moving quantum number in ScanIx2
	break; 
      }
      for (int alpha = ScanIx2[exclevel].size() - 1; alpha >= 0; --alpha) { 
	//cout << exclevel << "\t" << alpha << "\t" << ScanIx2[exclevel][alpha] << "\t" << BaseScanIx2[exclevel][alpha] << "\t" << (ScanIx2[exclevel][alpha] < BaseScanIx2[exclevel][alpha]) << endl;
	if (ScanIx2[exclevel][alpha] < BaseScanIx2[exclevel][alpha]) {
	  excindex = alpha;
	  excfound = true;
	  break;
	}
      }
    } while (!excfound);
    // If we haven't found an excitation, then exclevel == ScanIx2.size() and excindex = 0;
    //if (!excfound) excindex = ScanIx2[exclevel].size() - 1; 

    //cout << "exclevel = " << exclevel << endl;
    //cout << "excindex = " << excindex << endl;

    // The quantum numbers which we can move left are thus those 
    // with (j < exclevel) and (j == exclelev and alpha >= excindex)

      
    int ndesc_possible = 1;
    if (!disperse_only_current_exc) {
      ndesc_possible = 0;
      for (int j = 0; j <= JSC::min(exclevel, ScanIx2.size() - 1); ++j) ndesc_possible += ScanIx2[j].size();
    }

    // Now construct the actual descendents:
    Vect<string> desclabelfound (ndesc_possible);
    int ndesc_found = 0;

    if (!disperse_only_current_exc && !preserve_nexc) 
      // List descendents with a new excitation at a lower level:
      for (int j = 0; j < exclevel; ++j)
	for (int alpha = 0; alpha < ScanIx2[j].size(); ++alpha) {
	  if (ScanIx2[j][alpha] <= BaseScanIx2[j][alpha] // candidate either undisplaced or left-moving
	      && !ScanIx2[j].includes(ScanIx2[j][alpha] - 2) // there is a hole to its left
	      && ScanIx2[j][alpha] - 2 >= Ix2_min[j]) { // this new quantum number fits within the limits
	    // We've found a descendent
	    ScanIx2[j][alpha] -= 2;
	    desclabelfound[ndesc_found] = Return_State_Label (ScanIx2, OriginIx2);
	    ScanIx2[j][alpha] += 2;
	    ndesc_found++;
	  }
	}
    
    // List descendents with a new excitation at exclevel:
    if (exclevel < ScanIx2.size()) { // excfound == true, excindex is now guaranteed < ScanIx2[exclevel].size()
      int alphamin = (disperse_only_current_exc ? excindex : excindex + 1);
      int alphamax = (disperse_only_current_exc ? excindex : ScanIx2[exclevel].size() - 1);
      //for (int alpha = ScanIx2[exclevel].size() - 1; alpha >= excindex; --alpha) {
      for (int alpha = alphamax; alpha >= alphamin; --alpha) {
	if (ScanIx2[exclevel][alpha] <= BaseScanIx2[exclevel][alpha]
	    && !ScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] - 2)
	    && ScanIx2[exclevel][alpha] - 2 >= Ix2_min[exclevel]) {
	  ScanIx2[exclevel][alpha] -= 2;
	  desclabelfound[ndesc_found] = Return_State_Label (ScanIx2, OriginIx2);
	  ScanIx2[exclevel][alpha] += 2;
	  //cout << "\tTesting .compare: " << Extract_nexc_Label(desclabelfound[ndesc_found]) << "\t" << Extract_nexc_Label(ScanIx2_label) << "\t" << (Extract_nexc_Label(desclabelfound[ndesc_found]).compare(Extract_nexc_Label(ScanIx2_label)) == 0) << "\tpreserve_nexc: " << preserve_nexc << endl;
	  if (disperse_only_current_exc 
	      //|| (preserve_nexc == BaseScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] - 2)))
	      || (preserve_nexc == (Extract_nexc_Label(desclabelfound[ndesc_found]).compare(Extract_nexc_Label(ScanIx2_label)) == 0)))
	      //|| (preserve_nexc == strcmp(Extract_nexc_Label(desclabelfound[ndesc_found]), Extract_nexc_Label(ScanIx2_label))))
	    ndesc_found++;
	}
      }
    } // if (exclevel < ScanIx2.size())

    // Resize desc:
    Vect<string> desclabelfound_resized(ndesc_found);
    for (int idesc = 0; idesc < ndesc_found; ++idesc)
      desclabelfound_resized[idesc] = desclabelfound[idesc];
    
    //cout << "Down done" << endl;
    //return(descIx2found);
    //return(desclabelfound);
    return(desclabelfound_resized);
  }


  Vect<string> Descendent_States_with_iK_Preserved (string ScanIx2_label, const Vect<Vect<int> >& OriginIx2, const Vect<Vect<int> >& BaseScanIx2, const Vect<int>& Ix2_min, const Vect<int>& Ix2_max, bool disperse_only_current_exc_up, bool preserve_nexc_up, bool disperse_only_current_exc_down, bool preserve_nexc_down)
  {
    // Returns the labels of all states which are obtained from ScanIx2 by stepping p one step, and down one step.
    Vect<string> labels_up = Descendent_States_with_iK_Stepped_Up (ScanIx2_label, OriginIx2, BaseScanIx2, Ix2_min, Ix2_max, disperse_only_current_exc_up, preserve_nexc_up);
    //cout << "labels_up = " << labels_up << endl;

    Vect<string> labels_found(0);
    for (int i = 0; i < labels_up.size(); ++i) {
      labels_found.Append (Descendent_States_with_iK_Stepped_Down (labels_up[i], OriginIx2, BaseScanIx2, Ix2_min, Ix2_max, disperse_only_current_exc_down, preserve_nexc_down));
    }

    return(labels_found);
  }



  // For symmetric state descending:

  Vect<string> Descendent_States_with_iK_Stepped_Up_rightIx2only 
  (string ScanIx2_label, const Vect<Vect<int> >& OriginIx2, const Vect<Vect<int> >& BaseScanIx2, 
   const Vect<int>& Ix2_min, const Vect<int>& Ix2_max, bool disperse_only_current_exc, bool preserve_nexc)
  {
    // This function raises an Ix2 in the right half of ScanIx2, and does the symmetric lowering in the lower half.

    // ASSUMPTIONS: OriginIx2 is a symmetric state.

    //cout << "Desc Up A" << endl;

    Vect<Vect<int> > ScanIx2 = Return_Ix2_from_Label (ScanIx2_label, OriginIx2);

    //cout << "Desc Up B" << endl;

    // Determine the level and index of the bottom-most left-most right-moving quantum number sits:
    int exclevel = -1;
    int excindex = 0;
    bool excfound = false;
    do {
      exclevel++;
      if (exclevel == ScanIx2.size()) { // there isn't a single right-moving quantum number in ScanIx2
	break; 
      }
      for (int alpha = (ScanIx2[exclevel].size() + 1)/2; alpha < ScanIx2[exclevel].size(); ++alpha) 
	if (ScanIx2[exclevel][alpha] > BaseScanIx2[exclevel][alpha]) {
	  excindex = alpha;
	  excfound = true;
	  break;
	}
    } while (!excfound);
    // If we haven't found an excitation, then exclevel == ScanIx2.size() and excindex = 0;

    //cout << "exclevel = " << exclevel << "\t" << "excindex = " << excindex << endl;

    // The quantum numbers which we can move right are thus those 
    // with (j < exclevel) and (j == exclelev and alpha <= excindex)

    int ndesc_possible = 1;
    if (!disperse_only_current_exc) {
      ndesc_possible = 0;
      for (int j = 0; j <= JSC::min(exclevel, ScanIx2.size() -1); ++j) ndesc_possible += ScanIx2[j].size();
    }

    // Now construct the actual descendents:
    Vect<string> desclabelfound (ndesc_possible);
    int ndesc_found = 0;

    if (!disperse_only_current_exc && !preserve_nexc) 
      // List descendents with a new excitation at a lower level:
      // this can only create an additional ph pair
      for (int j = 0; j < exclevel; ++j)
	for (int alpha = (ScanIx2[j].size() + 1)/2; alpha < ScanIx2[j].size(); ++alpha) {
	  if (ScanIx2[j][alpha] >= BaseScanIx2[j][alpha] // candidate either undisplaced or right-moving
	      && !ScanIx2[j].includes(ScanIx2[j][alpha] + 2) // there is a hole to its right
	      && ScanIx2[j][alpha] + 2 <= Ix2_max[j]) { // this new quantum number fits within the limits
	    // We've found a descendent
	    ScanIx2[j][alpha] += 2;
	    ScanIx2[j][ScanIx2[j].size() - 1 - alpha] -= 2;
	    desclabelfound[ndesc_found] = Return_State_Label (ScanIx2, OriginIx2);
	    ScanIx2[j][alpha] -= 2;
	    ScanIx2[j][ScanIx2[j].size() - 1 - alpha] += 2;
	    ndesc_found++;
	  }
	}

    // List descendents with a new excitation at exclevel:
    if (exclevel < ScanIx2.size()) { // excfound == true, excindex is now guaranteed < ScanIx2[exclevel].size()
      int alphamin = (disperse_only_current_exc ? excindex : 0);
      alphamin = JSC::max((ScanIx2[exclevel].size() + 1)/2, alphamin);
      int alphamax = (disperse_only_current_exc ? excindex : excindex - 1);
      //for (int alpha = 0; alpha <= excindex; ++alpha) {
      //cout << "alphamin = " << alphamin << "\talphamax = " << alphamax << endl;
      for (int alpha = alphamin; alpha <= alphamax; ++alpha) {
	if (ScanIx2[exclevel][alpha] >= BaseScanIx2[exclevel][alpha]
	    && !ScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] + 2)
	    && ScanIx2[exclevel][alpha] + 2 <= Ix2_max[exclevel]) {
	  // We've found a descendent
	  ScanIx2[exclevel][alpha] += 2;
	  ScanIx2[exclevel][ScanIx2[exclevel].size() - 1 - alpha] -= 2;
	  //cout << "Desc Up a" << endl;
	  //cout << "ScanIx2[exclevel] = " << ScanIx2[exclevel] << endl;
	  desclabelfound[ndesc_found] = Return_State_Label (ScanIx2, OriginIx2);
	  //cout << "Desc Up b" << endl;
	  ScanIx2[exclevel][alpha] -= 2;
	  ScanIx2[exclevel][ScanIx2[exclevel].size() - 1 - alpha] += 2;
	  // If we're dispersing a subleading Ix2, check whether we match the preserve_nexc condition:
	  // The descendent is acceptable if disperse_only_current_exc == true,
	  // or if preserve_nexc == true and nexc labels match,
	  // or if preserve_nexc == false and nexc labels don't match:
	  //if (!disperse_only_current_exc && (BaseScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] + 2) != (Extract_nexc_Label(desclabelfound[ndesc_found]).compare(Extract_nexc_Label(ScanIx2_label)) == 0))) JSCerror("Inconsistency in check in Descendents");
	  if (disperse_only_current_exc 
	      //|| (preserve_nexc == BaseScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] + 2)))
	      || (preserve_nexc == (Extract_nexc_Label(desclabelfound[ndesc_found]).compare(Extract_nexc_Label(ScanIx2_label)) == 0)))
	    ndesc_found++;
	}
      }
    } //if (exclevel < ScanIx2.size())

    //cout << "Desc Up D" << endl;

    // Resize desc:
    Vect<string> desclabelfound_resized(ndesc_found);
    for (int idesc = 0; idesc < ndesc_found; ++idesc)
      desclabelfound_resized[idesc] = desclabelfound[idesc];

    //cout << "Up done" << endl;

    //return(descIx2found);
    //return(desclabelfound);
    return(desclabelfound_resized);
  }


  Vect<string> Descendent_States_with_iK_Stepped_Down_rightIx2only 
  (string ScanIx2_label, const Vect<Vect<int> >& OriginIx2, const Vect<Vect<int> >& BaseScanIx2, 
   const Vect<int>& Ix2_min, const Vect<int>& Ix2_max, bool disperse_only_current_exc, bool preserve_nexc)
  {
    Vect<Vect<int> > ScanIx2 = Return_Ix2_from_Label (ScanIx2_label, OriginIx2);

    // Determine the level and index of the bottom-most right-most left-moving quantum number sits:
    int exclevel = -1;
    int excindex = 0;
    bool excfound = false;

    //cout << "Looking for exclevel and excindex for " << endl << "\tBaseIx2 = " << BaseScanIx2 << endl << "\tScanIx2 = " << ScanIx2 << endl;
    do {
      exclevel++;
      if (exclevel == ScanIx2.size()) { // there isn't a single left-moving quantum number in ScanIx2
	break; 
      }
      for (int alpha = ScanIx2[exclevel].size() - 1; alpha >= (ScanIx2[exclevel].size() + 1)/2 ; --alpha) { 
	//cout << exclevel << "\t" << alpha << "\t" << ScanIx2[exclevel][alpha] << "\t" << BaseScanIx2[exclevel][alpha] << "\t" << (ScanIx2[exclevel][alpha] < BaseScanIx2[exclevel][alpha]) << endl;
	if (ScanIx2[exclevel][alpha] < BaseScanIx2[exclevel][alpha]) {
	  excindex = alpha;
	  excfound = true;
	  break;
	}
      }
    } while (!excfound);
    // If we haven't found an excitation, then exclevel == ScanIx2.size() and excindex = 0;
    //if (!excfound) excindex = ScanIx2[exclevel].size() - 1; 

    //cout << "exclevel = " << exclevel << endl;
    //cout << "excindex = " << excindex << endl;

    // The quantum numbers which we can move left are thus those 
    // with (j < exclevel) and (j == exclevel and alpha >= excindex)

      
    int ndesc_possible = 1;
    if (!disperse_only_current_exc) {
      ndesc_possible = 0;
      for (int j = 0; j <= JSC::min(exclevel, ScanIx2.size() - 1); ++j) ndesc_possible += ScanIx2[j].size();
    }

    // Now construct the actual descendents:
    Vect<string> desclabelfound (ndesc_possible);
    int ndesc_found = 0;

    if (!disperse_only_current_exc && !preserve_nexc) 
      // List descendents with a new excitation at a lower level:
      for (int j = 0; j < exclevel; ++j)
	for (int alpha = (ScanIx2[j].size() + 1)/2; alpha < ScanIx2[j].size(); ++alpha) {
	  if (ScanIx2[j][alpha] <= BaseScanIx2[j][alpha] // candidate either undisplaced or left-moving
	      && !ScanIx2[j].includes(ScanIx2[j][alpha] - 2) // there is a hole to its left
	      && ScanIx2[j][alpha] - 2 >= Ix2_min[j]) { // this new quantum number fits within the limits
	    // We've found a descendent
	    ScanIx2[j][alpha] -= 2;
	    ScanIx2[j][ScanIx2[j].size() - 1 - alpha] += 2;
	    desclabelfound[ndesc_found] = Return_State_Label (ScanIx2, OriginIx2);
	    ScanIx2[j][alpha] += 2;
	    ScanIx2[j][ScanIx2[j].size() - 1 - alpha] -= 2;
	    ndesc_found++;
	  }
	}
    
    // List descendents with a new excitation at exclevel:
    if (exclevel < ScanIx2.size()) { // excfound == true, excindex is now guaranteed < ScanIx2[exclevel].size()
      int alphamin = (disperse_only_current_exc ? excindex : excindex + 1);
      alphamin = JSC::max((ScanIx2[exclevel].size() + 1)/2, alphamin);
      int alphamax = (disperse_only_current_exc ? excindex : ScanIx2[exclevel].size() - 1);
      //for (int alpha = ScanIx2[exclevel].size() - 1; alpha >= excindex; --alpha) {
      for (int alpha = alphamax; alpha >= alphamin; --alpha) {
	if (ScanIx2[exclevel][alpha] <= BaseScanIx2[exclevel][alpha]
	    && !ScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] - 2)
	    && ScanIx2[exclevel][alpha] - 2 >= Ix2_min[exclevel]) {
	  ScanIx2[exclevel][alpha] -= 2;
	  ScanIx2[exclevel][ScanIx2[exclevel].size() - 1 - alpha] += 2;
	  desclabelfound[ndesc_found] = Return_State_Label (ScanIx2, OriginIx2);
	  ScanIx2[exclevel][alpha] += 2;
	  ScanIx2[exclevel][ScanIx2[exclevel].size() - 1 - alpha] -= 2;
	  //cout << "\tTesting .compare: " << Extract_nexc_Label(desclabelfound[ndesc_found]) << "\t" << Extract_nexc_Label(ScanIx2_label) << "\t" << (Extract_nexc_Label(desclabelfound[ndesc_found]).compare(Extract_nexc_Label(ScanIx2_label)) == 0) << "\tpreserve_nexc: " << preserve_nexc << endl;
	  if (disperse_only_current_exc 
	      //|| (preserve_nexc == BaseScanIx2[exclevel].includes(ScanIx2[exclevel][alpha] - 2)))
	      || (preserve_nexc == (Extract_nexc_Label(desclabelfound[ndesc_found]).compare(Extract_nexc_Label(ScanIx2_label)) == 0)))
	      //|| (preserve_nexc == strcmp(Extract_nexc_Label(desclabelfound[ndesc_found]), Extract_nexc_Label(ScanIx2_label))))
	    ndesc_found++;
	}
      }
    } // if (exclevel < ScanIx2.size())

    // Resize desc:
    Vect<string> desclabelfound_resized(ndesc_found);
    for (int idesc = 0; idesc < ndesc_found; ++idesc)
      desclabelfound_resized[idesc] = desclabelfound[idesc];
    
    //cout << "Down done" << endl;
    //return(descIx2found);
    //return(desclabelfound);
    return(desclabelfound_resized);
  }



  // Specialization for Lieb-Liniger case:
  //Vect<string> Descendent_States_with_iK_Stepped_Up (string ScanIx2_label, const Vect<int>& OriginIx2, const Vect<int>& BaseScanIx2)
  Vect<string> Descendent_States_with_iK_Stepped_Up (string ScanIx2_label, const LiebLin_Bethe_State& OriginState, bool disperse_only_current_exc, bool preserve_nexc)
  {
    //Vect<Vect<int> > ScanIx2here(1);
    //ScanIx2here[0] = ScanIx2;
    Vect<Vect<int> > OriginIx2here(1);
    OriginIx2here[0] = OriginState.Ix2;
    Vect<Vect<int> > BaseScanIx2here(1);
    //BaseScanIx2here[0] = BaseScanIx2;
    BaseScanIx2here[0] = OriginState.Ix2;
    Vect<int> Ix2_min(1);
    Ix2_min[0] = LIEBLIN_Ix2_MIN;
    Vect<int> Ix2_max(1);
    Ix2_max[0] = LIEBLIN_Ix2_MAX;

    return (Descendent_States_with_iK_Stepped_Up (ScanIx2_label, OriginIx2here, BaseScanIx2here, Ix2_min, Ix2_max, disperse_only_current_exc, preserve_nexc));
  }
  // Specialization for Lieb-Liniger case:
  //Vect<string> Descendent_States_with_iK_Stepped_Down (string ScanIx2_label, const Vect<int>& OriginIx2, const Vect<int>& BaseScanIx2)
  Vect<string> Descendent_States_with_iK_Stepped_Down (string ScanIx2_label, const LiebLin_Bethe_State& OriginState, bool disperse_only_current_exc, bool preserve_nexc)
  {
    //Vect<Vect<int> > ScanIx2here(1);
    //ScanIx2here[0] = ScanIx2;
    Vect<Vect<int> > OriginIx2here(1);
    OriginIx2here[0] = OriginState.Ix2;
    Vect<Vect<int> > BaseScanIx2here(1);
    //BaseScanIx2here[0] = BaseScanIx2;
    BaseScanIx2here[0] = OriginState.Ix2;
    Vect<int> Ix2_min(1);
    Ix2_min[0] = LIEBLIN_Ix2_MIN;
    Vect<int> Ix2_max(1);
    Ix2_max[0] = LIEBLIN_Ix2_MAX;

    return (Descendent_States_with_iK_Stepped_Down (ScanIx2_label, OriginIx2here, BaseScanIx2here, Ix2_min, Ix2_max, disperse_only_current_exc, preserve_nexc));
  }

  // Specialization for Lieb-Liniger case:
  //Vect<string> Descendent_States_with_iK_Preserved (string ScanIx2_label, const Vect<int>& OriginIx2, const Vect<int>& BaseScanIx2)
  Vect<string> Descendent_States_with_iK_Preserved (string ScanIx2_label, const LiebLin_Bethe_State& OriginState, bool disperse_only_current_exc_up, bool preserve_nexc_up, bool disperse_only_current_exc_down, bool preserve_nexc_down)
  {
    //Vect<Vect<int> > ScanIx2here(1);
    //ScanIx2here[0] = ScanIx2;
    Vect<Vect<int> > OriginIx2here(1);
    OriginIx2here[0] = OriginState.Ix2;
    Vect<Vect<int> > BaseScanIx2here(1);
    //BaseScanIx2here[0] = BaseScanIx2;
    BaseScanIx2here[0] = OriginState.Ix2;
    Vect<int> Ix2_min(1);
    Ix2_min[0] = LIEBLIN_Ix2_MIN;
    Vect<int> Ix2_max(1);
    Ix2_max[0] = LIEBLIN_Ix2_MAX;

    return (Descendent_States_with_iK_Preserved (ScanIx2_label, OriginIx2here, BaseScanIx2here, Ix2_min, Ix2_max, disperse_only_current_exc_up, preserve_nexc_up, disperse_only_current_exc_down, preserve_nexc_down));
  }

  // Specialization for Lieb-Liniger case:
  Vect<string> Descendent_States_with_iK_Stepped_Up_rightIx2only 
  (string ScanIx2_label, const LiebLin_Bethe_State& OriginState, bool disperse_only_current_exc, bool preserve_nexc)
  {
    Vect<Vect<int> > OriginIx2here(1);
    OriginIx2here[0] = OriginState.Ix2;
    Vect<Vect<int> > BaseScanIx2here(1);
    //BaseScanIx2here[0] = BaseScanIx2;
    BaseScanIx2here[0] = OriginState.Ix2;
    Vect<int> Ix2_min(1);
    Ix2_min[0] = LIEBLIN_Ix2_MIN;
    Vect<int> Ix2_max(1);
    Ix2_max[0] = LIEBLIN_Ix2_MAX;

    return(Descendent_States_with_iK_Stepped_Up_rightIx2only (ScanIx2_label, OriginIx2here, BaseScanIx2here, Ix2_min, Ix2_max, disperse_only_current_exc, preserve_nexc));
  }

  Vect<string> Descendent_States_with_iK_Stepped_Down_rightIx2only 
  (string ScanIx2_label, const LiebLin_Bethe_State& OriginState, bool disperse_only_current_exc, bool preserve_nexc)
  {
    Vect<Vect<int> > OriginIx2here(1);
    OriginIx2here[0] = OriginState.Ix2;
    Vect<Vect<int> > BaseScanIx2here(1);
    //BaseScanIx2here[0] = BaseScanIx2;
    BaseScanIx2here[0] = OriginState.Ix2;
    Vect<int> Ix2_min(1);
    Ix2_min[0] = LIEBLIN_Ix2_MIN;
    Vect<int> Ix2_max(1);
    Ix2_max[0] = LIEBLIN_Ix2_MAX;

    return(Descendent_States_with_iK_Stepped_Down_rightIx2only (ScanIx2_label, OriginIx2here, BaseScanIx2here, Ix2_min, Ix2_max, disperse_only_current_exc, preserve_nexc));
  }


  // Specializations for Heis states:
  Vect<string> Descendent_States_with_iK_Stepped_Up 
  (string ScanIx2_label, const Heis_Bethe_State& OriginState, bool disperse_only_current_exc, bool preserve_nexc)
  {
    return(Descendent_States_with_iK_Stepped_Up (ScanIx2_label, OriginState.Ix2, OriginState.Ix2, OriginState.base.Ix2_min, OriginState.base.Ix2_max, disperse_only_current_exc, preserve_nexc));
  }
  Vect<string> Descendent_States_with_iK_Stepped_Down 
  (string ScanIx2_label, const Heis_Bethe_State& OriginState, bool disperse_only_current_exc, bool preserve_nexc)
  {
    return(Descendent_States_with_iK_Stepped_Down (ScanIx2_label, OriginState.Ix2, OriginState.Ix2, OriginState.base.Ix2_min, OriginState.base.Ix2_max, disperse_only_current_exc, preserve_nexc));
  }
  Vect<string> Descendent_States_with_iK_Preserved 
  (string ScanIx2_label, const Heis_Bethe_State& OriginState, bool disperse_only_current_exc_up, bool preserve_nexc_up, bool disperse_only_current_exc_down, bool preserve_nexc_down)
  {
    return(Descendent_States_with_iK_Preserved (ScanIx2_label, OriginState.Ix2, OriginState.Ix2, OriginState.base.Ix2_min, OriginState.base.Ix2_max, disperse_only_current_exc_up, preserve_nexc_up, disperse_only_current_exc_down, preserve_nexc_down));
  }
  Vect<string> Descendent_States_with_iK_Stepped_Up_rightIx2only 
  (string ScanIx2_label, const Heis_Bethe_State& OriginState, bool disperse_only_current_exc, bool preserve_nexc)
  {
    return(Descendent_States_with_iK_Stepped_Up_rightIx2only (ScanIx2_label, OriginState.Ix2, OriginState.Ix2, OriginState.base.Ix2_min, OriginState.base.Ix2_max, disperse_only_current_exc, preserve_nexc));
  }
  Vect<string> Descendent_States_with_iK_Stepped_Down_rightIx2only 
  (string ScanIx2_label, const Heis_Bethe_State& OriginState, bool disperse_only_current_exc, bool preserve_nexc)
  {
    return(Descendent_States_with_iK_Stepped_Down_rightIx2only (ScanIx2_label, OriginState.Ix2, OriginState.Ix2, OriginState.base.Ix2_min, OriginState.base.Ix2_max, disperse_only_current_exc, preserve_nexc));
  }
      


  Vect<bool> Is_Good_New_Hole_Position (const Vect<Vect<int> >& OriginIx2, State_Label_Data currentdata, int exclevel_newph)
  {
    // Given a state, returns the acceptable new hole positions.
    
    
    // Define the objects for the newstatedata:
    Vect<int> type_new = currentdata.type;
    Vect<int> M_new = currentdata.M;
    Vect<int> nexc_new = currentdata.nexc;
    nexc_new[exclevel_newph] += 1; // we drill one more particle-hole pair at this level 
    int index_new = (currentdata.nexc[exclevel_newph] + 1)/2; // we put the new p-h pair at index index_new.
    
    //int ntypespresent = ScanIx2.size();
    int ntypespresent = currentdata.type.size();
    Vect<Vect<int> > Ix2old_new(ntypespresent);
    Vect<Vect<int> > Ix2exc_new(ntypespresent);
    for (int it = 0; it < ntypespresent; ++it) Ix2old_new[it] = Vect<int>(JSC::max(nexc_new[it],1));
    for (int it = 0; it < ntypespresent; ++it) Ix2exc_new[it] = Vect<int>(JSC::max(nexc_new[it],1));
    
    // Copy earlier data in:
    for (int it = 0; it < ntypespresent; ++it) {
      for (int i = 0; i < currentdata.nexc[it]; ++i) {
	Ix2old_new[it][i + (it == exclevel_newph && i >= index_new)] = currentdata.Ix2old[it][i]; 
	Ix2exc_new[it][i + (it == exclevel_newph && i >= index_new)] = currentdata.Ix2exc[it][i]; 
      }
    }
    
    //cout << "Here 1" << endl;
    
    State_Label_Data descdatanewph (type_new, M_new, nexc_new, Ix2old_new, Ix2exc_new);
    
    // We now look for all possible hole positions,
    // the allowable ones being either at the edge of a block, or next to an existing hole excitation,
    // under the condition of obeying the `towards the block center' rule.

    Vect<bool> isgoodnewholepos(false, OriginIx2[descdatanewph.type[exclevel_newph] ].size()); 
    
    for (int ih = 0; ih < OriginIx2[descdatanewph.type[exclevel_newph] ].size(); ++ih) {
      
      int candidateIx2old = OriginIx2[descdatanewph.type[exclevel_newph] ][ih];
      
      // candidateIx2old is an acceptable position for the new hole provided the following conditions are fulfilled:
      // A- it is in OriginIx2
      // B- it follows the ordering rule, i.e. it sits in the middle of previous particle excitations, 
      //      namely between Ix2old[index_new - 1] (if this exists) and Ix2old[index_new] (if this exists)
      // C- it is next to an OriginIx2 vacancy or immediately right of Ix2old[index_new - 1] (if this exists) 
      //      or immediately left of Ix2old[index_new] (if this exists) 
      // D- it does not break the `towards the center' rule 
      //      (it will break the rule at this point if it is created away from an OriginIx2 boundary (and thus next to a preexisting excitation), 
      //      and if this excitation and it are not in the same sideblock (in other words: if there is a sideblock boundary between them)
      
      //cout << "candidateIx2old " << candidateIx2old << " being tested" << endl;
      // A
      if (!OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old)) {
	// is not contained in OriginIx2
	//cout << "candidateIx2old " << candidateIx2old << " rejected for reason A" << endl;
	continue; 
      }
      
      // B1
      if (currentdata.nexc[exclevel_newph] > 0 
	  && candidateIx2old <= currentdata.Ix2old[exclevel_newph][index_new - 1]) {
	// there is at least one hole exc to the left, and the candidate position isn't right of Ix2old[index_new - 1]
	//cout << "candidateIx2old " << candidateIx2old << " rejected for reason B1" << endl;
	continue;
      }
      
      // B2
      if (currentdata.nexc[exclevel_newph] > 1
	  && candidateIx2old >= currentdata.Ix2old[exclevel_newph][index_new]) {
	// there is at least one hole exc to the right, and the candidate position isn't left of Ix2old[index_new]
	//cout << "candidateIx2old " << candidateIx2old << " rejected for reason B2" << endl;
	continue;
      }

      // C- it is next to an OriginIx2 vacancy or immediately right of Ix2old[index_new - 1] (if this exists) 
      //      or immediately left of Ix2old[index_new] (if this exists) 
      if (!(!OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old + 2) || !OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old - 2)) 
	  // doesn't sit next to an OriginIx2 vacancy
	  && (currentdata.nexc[exclevel_newph] == 0 || candidateIx2old != currentdata.Ix2old[exclevel_newph][index_new - 1] + 2)
	  // doesn't sit immediately right of first hole excitation to the left 
	  && (currentdata.nexc[exclevel_newph] <= 1 || candidateIx2old != currentdata.Ix2old[exclevel_newph][index_new] - 2)
	  // doesn't sit immediately left of first hole  excitation to the right
	  ) {
	//cout << "candidateIx2old " << candidateIx2old << " rejected for reason C" << endl;
	continue;
      }

      // D- it does not break the `towards the center' rule
      // In other words, if created away from a block boundary but next to a preexisting hole, 
      // must be in same sideblock as this particle:
      
      // Determine the size of the block of OriginIx2 in which this hole sits:
      int nroccupiedtoleft = 0;
      while (OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old -2*(nroccupiedtoleft + 1)))
	nroccupiedtoleft++;
      int nroccupiedtoright = 0;
      while (OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old +2*(nroccupiedtoright + 1)))
	nroccupiedtoright++;
      
      // We can determine whether the new hole would be left- or right-moving
      bool hole_candidate_is_left_moving = nroccupiedtoleft >= nroccupiedtoright;
      bool hole_candidate_is_right_moving = nroccupiedtoleft < nroccupiedtoright;
      
      if (hole_candidate_is_left_moving 
	  && (currentdata.nexc[exclevel_newph] > 0 && candidateIx2old == currentdata.Ix2old[exclevel_newph][index_new - 1] + 2)
	  // it is created to the right of a preexisting hole excitation
	  && OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old + 2) // and is not sitting at the boundary
	  && (currentdata.nexc[exclevel_newph] <= 1 || candidateIx2old != currentdata.Ix2old[exclevel_newph][index_new] - 2)
	  // and is not sitting just left of another preexisting hole excitation which is also left moving
	  ) 
	{
	  //cout << "candidateIx2old " << candidateIx2old << " rejected for reason D1" << endl;
	  continue;
	}	      
      
      if (hole_candidate_is_right_moving 
	  && (currentdata.nexc[exclevel_newph] > 1 && candidateIx2old == currentdata.Ix2old[exclevel_newph][index_new] - 2)
	  // it is created to the left of a preexisting hole excitation
	  && OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old - 2) // and is not sitting at the boundary
	  && (currentdata.nexc[exclevel_newph] == 0 || candidateIx2old != currentdata.Ix2old[exclevel_newph][index_new - 1] + 2)
	  // and is not sitting just right of another preexisting hole excitation which is also right moving
	  ) 
	{
	  //cout << "candidateIx2old " << candidateIx2old << " rejected for reason D2" << endl;
	  continue;
	}	      
      
      
      // If we have reached this point, candidateIx2old is acceptable.
      
      isgoodnewholepos[ih] = true;
      
    } // for (int ih

    return(isgoodnewholepos);

  } // Vect<bool> Is_Good_New_Hole_Position




  //Vect<string> Descendents (const Bethe_State& ScanState, string type_required)
  Vect<string> Descendents (string ScanStateLabel, const Vect<Vect<int> >& ScanIx2, const Vect<Vect<int> >& OriginIx2, 
			    const Vect<int>& Ix2_min, const Vect<int>& Ix2_max, int type_required)
  {
    // This generates descendents for a fixed base.
    // As compared to the OriginState, ScanState thus has only (in-level) particle-hole excitations

    // ASSUMPTIONS:
    // ScanStateLabel is consistent with ScanIx2 and OriginIx2.

    // type_required == 0: inner (right)most hole moved
    // type_required == 1: inner (right)most particle moved
    // type_required == 2: new particle-hole pair added, keeping only closest p < h and p > h cases
    // type_required == 3: at level 0, move leftmost particle one step left, rightmost one step right (fixed iK logic using skeletons).
    // type_required == 4: generalized Umklapp: increase distance between latest-added p-h pair, staying on boundary of blocks

    //cout << "\tCalling Descendents for type = " << type_required << endl;

    bool cinbreaks = false;
    
    // Number of descendents: 
    //int ndesc_possible = (type_required == 0 ? 1 : ScanState.base.charge * 2); // upper bound guess; should be refined.
    //int ndesc_possible = (type_required == 0 ? 1 : 2* ScanIx2[0].size() * 2 * ScanIx2[0].size()); // upper bound guess; should be refined
    int ndesc_possible = 1;
    if (type_required == 1 || type_required == 2) {
      for (int i = 0; i < ScanIx2.size(); ++i)
	ndesc_possible = JSC::max (ndesc_possible, 2* ScanIx2[i].size() * 2 * ScanIx2[i].size()); // inexact, should be refined
    }
    //cout << "ndesc_possible = " << ndesc_possible << endl;
    
    //cout << "OriginIx2.size() = " << OriginIx2.size() << endl;
    //for (int i = 0; i < OriginIx2.size(); ++i) cout << "i = " << i << "\tOriginIx2[i].size() = " << OriginIx2[i].size() << endl;
    //for (int i = 0; i < OriginIx2.size(); ++i) cout << "i = " << i << "\tOriginIx2[i] = " << OriginIx2[i] << endl;
    //cout << "OriginIx2 = " << endl;
    //cout << OriginIx2 << endl;
 
    Vect<string> desclabelfound (ndesc_possible);
    Vect<int> desctypefound (ndesc_possible);
    int ndesc_found = 0;

    // First read the current state data:
    State_Label_Data currentdata = Read_State_Label (ScanStateLabel, OriginIx2);

    // Determine the level at which the highest current excitation sits
    //int exclevel = ScanIx2.size() - 1;
    int exclevel = currentdata.type.size() - 1;
    while (exclevel > 0 && currentdata.nexc[exclevel] == 0) exclevel--;
    //cout << "exclevel = " << exclevel << endl;

    //cout << "***** Looking for descendents of type " << type_required << " for state with OriginIx2, ScanIx2: *******" << endl;
    //cout << OriginIx2[currentdata.type[exclevel] ] << endl;
    //cout << ScanIx2[currentdata.type[exclevel] ] << endl;
    //if (currentdata.nexc[exclevel] > 0) {
    //cout << "with current excitations (holes, particles):" << endl;
    //cout << currentdata.Ix2old[exclevel] << endl;
    //cout << currentdata.Ix2exc[exclevel] << endl;
    //}

    if (type_required == 0) {

      // We move the inner(right)most hole of the highest excited level one more step towards 
      // the center of the block of Ix2 vacancies in which it sits.
      
      // Is there already an excitation at level 0? If so, move inner(right)most particle one step further.
      if (currentdata.nexc[0] > 0) {

	// Produce a new state_label_data:
	State_Label_Data descdata = Read_State_Label (ScanStateLabel, OriginIx2);
	// Identify the inner(right)most excitation:
	int innerindex = currentdata.nexc[exclevel]/2;

	int newholepos = currentdata.Ix2old[exclevel][innerindex];

	// Determine the size of the block of OriginState.Ix2 in which this hole sits:
	int nroccupiedtoleft = 0;
	while (OriginIx2[descdata.type[exclevel] ].includes (currentdata.Ix2old[exclevel][innerindex] -2*(nroccupiedtoleft + 1)))
	  nroccupiedtoleft++;
	int nroccupiedtoright = 0;
	while (OriginIx2[descdata.type[exclevel] ].includes (currentdata.Ix2old[exclevel][innerindex] +2*(nroccupiedtoright + 1)))
	  nroccupiedtoright++;

	if (nroccupiedtoleft - nroccupiedtoright + 2 < 0) newholepos += 2;
	else if (nroccupiedtoleft - nroccupiedtoright - 2 >= 0) newholepos -= 2;

	if (newholepos != currentdata.Ix2old[exclevel][innerindex] // we have successfully moved the hole
	    && !currentdata.Ix2old[exclevel].includes(newholepos) // new hole position is not already taken
	    ) 
	  { // we have found a descendent
	    descdata.Ix2old[exclevel][innerindex] = newholepos;
	    desclabelfound[ndesc_found] = Return_State_Label (descdata, OriginIx2);
	    desctypefound[ndesc_found] = 0;
	    //cout << "For state with OriginIx2, ScanIx2:" << endl;
	    //cout << OriginIx2[exclevel] << endl;
	    //cout << ScanIx2[exclevel] << endl;
	    //if (currentdata.nexc[exclevel] > 0) {
	    //cout << "with current excitations (holes, particles):" << endl;
	    //cout << currentdata.Ix2old[exclevel] << endl;
	    //cout << currentdata.Ix2exc[exclevel] << endl;
	    //}
	    //cout << "\tFound new descendent of type " << desctypefound[ndesc_found] << ": label " << desclabelfound[ndesc_found] << "\tholes " << descdata.Ix2old[exclevel] << "\tpart " << descdata.Ix2exc[exclevel] << endl;
	    ndesc_found++;
	    if (cinbreaks) { char a;    cin >> a;}
	  }

      } // if (currentdata.nexc[exclevel] > 0)

    } // if (type_required == 0)


    if (type_required == 1) {

      //cout << "exclevel = " << exclevel << endl;
      //cout << "currentdata.nexc[exclevel] = " << currentdata.nexc[exclevel] << endl;

      // We move the inner(right)most particle of the highest excited level one more step towards 
      // the center of the block of Ix2 vacancies in which it sits.
      
      // Is there already an excitation at level 0? If so, move inner(right)most particle one step further.
      if (currentdata.nexc[exclevel] > 0) {

	// Produce a new state_label_data:
	State_Label_Data descdata = Read_State_Label (ScanStateLabel, OriginIx2);
	// Identify the inner(right)most excitation:
	int innerindex = currentdata.nexc[exclevel]/2;

	int partpos = currentdata.Ix2exc[exclevel][innerindex];

	// Determine the size of the block of OriginState.Ix2 vacancies in which this particle sits:
	bool partpos_is_left_of_all_OriginIx2 = (partpos <= OriginIx2[descdata.type[exclevel] ].min());
	bool partpos_is_right_of_all_OriginIx2 = (partpos >= OriginIx2[descdata.type[exclevel] ].max());
	int nremptytoleft = 0;
	if (partpos_is_left_of_all_OriginIx2) nremptytoleft = (partpos - Ix2_min[descdata.type[exclevel] ])/2;
	else while (!OriginIx2[descdata.type[exclevel] ].includes (partpos -2*(nremptytoleft + 1)))
	       nremptytoleft++;
	int nremptytoright = 0;
	if (partpos_is_right_of_all_OriginIx2) nremptytoright = (Ix2_max[descdata.type[exclevel] ] - partpos)/2;
	else while (!OriginIx2[descdata.type[exclevel] ].includes (partpos +2*(nremptytoright + 1)))
	       nremptytoright++;
	
	if (!partpos_is_left_of_all_OriginIx2 && (partpos_is_right_of_all_OriginIx2 || nremptytoleft - nremptytoright + 2 < 0)) partpos += 2;
	else if (!partpos_is_right_of_all_OriginIx2 && (partpos_is_left_of_all_OriginIx2 || nremptytoleft - nremptytoright - 2 >= 0)) partpos -= 2;

	if (partpos != currentdata.Ix2exc[exclevel][innerindex] // we have successfully moved the particle
	    && !OriginIx2[descdata.type[exclevel] ].includes(partpos) // it's actually a new particle position
	    && !currentdata.Ix2exc[exclevel].includes(partpos) // new particle position is not already taken
	    && partpos >= Ix2_min[descdata.type[exclevel] ]
	    && partpos <= Ix2_max[descdata.type[exclevel] ]
	    ) 
	  { // we have found a descendent
	    descdata.Ix2exc[exclevel][innerindex] = partpos;
	    desclabelfound[ndesc_found] = Return_State_Label (descdata, OriginIx2);
	    desctypefound[ndesc_found] = 1;
	    //cout << "For state with OriginIx2, ScanIx2:" << endl;
	    //cout << OriginIx2[exclevel] << endl;
	    //cout << ScanIx2[exclevel] << endl;
	    //if (currentdata.nexc[exclevel] > 0) {
	    //cout << "with current excitations (holes, particles):" << endl;
	    //cout << currentdata.Ix2old[exclevel] << endl;
	    //cout << currentdata.Ix2exc[exclevel] << endl;
	    //}
	    //cout << "\tFound new descendent of type " << desctypefound[ndesc_found] << ": label " << desclabelfound[ndesc_found] << "\tholes " << descdata.Ix2old[exclevel] << "\tpart " << descdata.Ix2exc[exclevel] << endl;
	    ndesc_found++;
	    if (cinbreaks) { char a;    cin >> a;}
	  }

      } // if (currentdata.nexc[exclevel] > 0) 


    } // if (type_required == 1)

  
    if (type_required == 2) {

      // Now add a new p-h pair at the inner(right)most position, at each level from exclevel upwards,
      // putting the particle and hole to each available positions at the edge of vacancy/occupancy blocks.

      //cout << "Trying for type 2 descendent. exclevel = " << exclevel << "\tcurrentdata.nexc.size() = " << currentdata.nexc.size() << endl;

      //for (int exclevel_newph = exclevel; exclevel_newph < ScanIx2.size(); ++exclevel_newph) {
      for (int exclevel_newph = exclevel; exclevel_newph < currentdata.nexc.size(); ++exclevel_newph) {

	if (ScanIx2[currentdata.type[exclevel_newph] ].size() <= currentdata.nexc[exclevel_newph]) continue; // no space for another p-h

	// Define the objects for the newstatedata:
	Vect<int> type_new = currentdata.type;
	Vect<int> M_new = currentdata.M;
	Vect<int> nexc_new = currentdata.nexc;
	nexc_new[exclevel_newph] += 1; // we drill one more particle-hole pair at this level 
	int index_new = (currentdata.nexc[exclevel_newph] + 1)/2; // we put the new p-h pair at index index_new.
      
	//int ntypespresent = ScanIx2.size();
	int ntypespresent = currentdata.type.size();
	Vect<Vect<int> > Ix2old_new(ntypespresent);
	Vect<Vect<int> > Ix2exc_new(ntypespresent);
	for (int it = 0; it < ntypespresent; ++it) Ix2old_new[it] = Vect<int>(JSC::max(nexc_new[it],1));
	for (int it = 0; it < ntypespresent; ++it) Ix2exc_new[it] = Vect<int>(JSC::max(nexc_new[it],1));

	// Copy earlier data in:
	for (int it = 0; it < ntypespresent; ++it) {
	  for (int i = 0; i < currentdata.nexc[it]; ++i) {
	    Ix2old_new[it][i + (it == exclevel_newph && i >= index_new)] = currentdata.Ix2old[it][i]; 
	    Ix2exc_new[it][i + (it == exclevel_newph && i >= index_new)] = currentdata.Ix2exc[it][i]; 
	  }
	}

	//cout << "Here 1" << endl;

	State_Label_Data descdatanewph (type_new, M_new, nexc_new, Ix2old_new, Ix2exc_new);

	// We now look for all possible hole positions,
	// the allowable ones being either at the edge of a block, or next to an existing hole excitation,
	// under the condition of obeying the `towards the block center' rule.
	
	Vect<bool> isgoodnewholepos = Is_Good_New_Hole_Position (OriginIx2, currentdata, exclevel_newph);
	/* The function call above replaces the whole block here:	
	Vect<bool> isgoodnewholepos(false, OriginIx2[descdatanewph.type[exclevel_newph] ].size()); 

	for (int ih = 0; ih < OriginIx2[descdatanewph.type[exclevel_newph] ].size(); ++ih) {
	  
	  int candidateIx2old = OriginIx2[descdatanewph.type[exclevel_newph] ][ih];

	  // candidateIx2old is an acceptable position for the new hole provided the following conditions are fulfilled:
	  // A- it is in OriginIx2
	  // B- it follows the ordering rule, i.e. it sits in the middle of previous particle excitations, 
	  //      namely between Ix2old[index_new - 1] (if this exists) and Ix2old[index_new] (if this exists)
	  // C- it is next to an OriginIx2 vacancy or immediately right of Ix2old[index_new - 1] (if this exists) 
	  //      or immediately left of Ix2old[index_new] (if this exists) 
	  // D- it does not break the `towards the center' rule 
	  //      (it will break the rule at this point if it is created away from an OriginIx2 boundary (and thus next to a preexisting excitation), 
	  //      and if this excitation and it are not in the same sideblock (in other words: if there is a sideblock boundary between them)

	  //cout << "candidateIx2old " << candidateIx2old << " being tested" << endl;
	  // A
	  if (!OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old)) {
	    // is not contained in OriginIx2
	    //cout << "candidateIx2old " << candidateIx2old << " rejected for reason A" << endl;
	    continue; 
	  }
	  
	  // B1
	  if (currentdata.nexc[exclevel_newph] > 0 
	      && candidateIx2old <= currentdata.Ix2old[exclevel_newph][index_new - 1]) {
	    // there is at least one hole exc to the left, and the candidate position isn't right of Ix2old[index_new - 1]
	    //cout << "candidateIx2old " << candidateIx2old << " rejected for reason B1" << endl;
	    continue;
	  }
	  
	  // B2
	  if (currentdata.nexc[exclevel_newph] > 1
	      && candidateIx2old >= currentdata.Ix2old[exclevel_newph][index_new]) {
	    // there is at least one hole exc to the right, and the candidate position isn't left of Ix2old[index_new]
	    //cout << "candidateIx2old " << candidateIx2old << " rejected for reason B2" << endl;
	    continue;
	  }
	  
	  // C- it is next to an OriginIx2 vacancy or immediately right of Ix2old[index_new - 1] (if this exists) 
	  //      or immediately left of Ix2old[index_new] (if this exists) 
	  if (!(!OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old + 2) || !OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old - 2)) 
	      // doesn't sit next to an OriginIx2 vacancy
	      && (currentdata.nexc[exclevel_newph] == 0 || candidateIx2old != currentdata.Ix2old[exclevel_newph][index_new - 1] + 2)
	      // doesn't sit immediately right of first hole excitation to the left 
	      && (currentdata.nexc[exclevel_newph] <= 1 || candidateIx2old != currentdata.Ix2old[exclevel_newph][index_new] - 2)
	      // doesn't sit immediately left of first hole  excitation to the right
	      ) {
	    //cout << "candidateIx2old " << candidateIx2old << " rejected for reason C" << endl;
	    continue;
	  }
	  
	  // D- it does not break the `towards the center' rule
	  // In other words, if created away from a block boundary but next to a preexisting hole, 
	  // must be in same sideblock as this particle:

	  // Determine the size of the block of OriginIx2 in which this hole sits:
	  int nroccupiedtoleft = 0;
	  while (OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old -2*(nroccupiedtoleft + 1)))
	    nroccupiedtoleft++;
	  int nroccupiedtoright = 0;
	  while (OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old +2*(nroccupiedtoright + 1)))
	    nroccupiedtoright++;

	  // We can determine whether the new hole would be left- or right-moving
	  bool hole_candidate_is_left_moving = nroccupiedtoleft >= nroccupiedtoright;
	  bool hole_candidate_is_right_moving = nroccupiedtoleft < nroccupiedtoright;

	  if (hole_candidate_is_left_moving 
	      && (currentdata.nexc[exclevel_newph] > 0 && candidateIx2old == currentdata.Ix2old[exclevel_newph][index_new - 1] + 2)
		  // it is created to the right of a preexisting hole excitation
	      && OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old + 2) // and is not sitting at the boundary
	      && (currentdata.nexc[exclevel_newph] <= 1 || candidateIx2old != currentdata.Ix2old[exclevel_newph][index_new] - 2)
	      // and is not sitting just left of another preexisting hole excitation which is also left moving
	      ) 
	    {
	      //cout << "candidateIx2old " << candidateIx2old << " rejected for reason D1" << endl;
	      continue;
	    }	      

	  if (hole_candidate_is_right_moving 
	      && (currentdata.nexc[exclevel_newph] > 1 && candidateIx2old == currentdata.Ix2old[exclevel_newph][index_new] - 2)
		  // it is created to the left of a preexisting hole excitation
	      && OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2old - 2) // and is not sitting at the boundary
	      && (currentdata.nexc[exclevel_newph] == 0 || candidateIx2old != currentdata.Ix2old[exclevel_newph][index_new - 1] + 2)
	      // and is not sitting just right of another preexisting hole excitation which is also right moving
	      ) 
	    {
	      //cout << "candidateIx2old " << candidateIx2old << " rejected for reason D2" << endl;
	      continue;
	    }	      


	  // If we have reached this point, candidateIx2old is acceptable.
	    
	  isgoodnewholepos[ih] = true;

	} // for (int ih
	*/

	
	      
	//cout << "Here 2" << endl;

	//cout << "\tisgoodnewholdpos = " << isgoodnewholepos << endl;

	//cout << "Looking for a new particle position: OriginIx2, ScanIx2:" << endl;
	//cout << OriginIx2[exclevel_newph] << endl;
	//cout << ScanIx2[exclevel_newph] << endl;
	//if (currentdata.nexc[exclevel_newph] > 0) {
	//  cout << "Current excitations (holes, particles):" << endl;
	//  cout << currentdata.Ix2old[exclevel_newph] << endl;
	//  cout << currentdata.Ix2exc[exclevel_newph] << endl;
	//}

	// We now look for all possible particle positions, 
	// the allowable ones being either at the edge of a block, or next to an existing particle excitation,
	// under the condition of obeying the `towards the block center' rule.
	
	// Determine range of possible Ix2exc:
	int Ix2excmin = ScanIx2[descdatanewph.type[exclevel_newph] ].min() - 2;
	if (currentdata.nexc[exclevel_newph] > 0) 
	  Ix2excmin = JSC::min (Ix2excmin, currentdata.Ix2exc[exclevel_newph][index_new - 1] + 2);
	Ix2excmin = JSC::max (Ix2excmin, Ix2_min[descdatanewph.type[exclevel_newph] ]);
	int Ix2excmax = ScanIx2[descdatanewph.type[exclevel_newph] ].max() + 2;
	if (currentdata.nexc[exclevel_newph] > 1) 
	  Ix2excmax = JSC::max (Ix2excmax, currentdata.Ix2exc[exclevel_newph][index_new] - 2);
	Ix2excmax = JSC::min (Ix2excmax, Ix2_max[descdatanewph.type[exclevel_newph] ]);
	
	Vect<bool> isgoodnewpartpos(false, (Ix2excmax - Ix2excmin)/2 + 1);
	
	for (int candidateIx2exc = Ix2excmin; candidateIx2exc <= Ix2excmax; candidateIx2exc += 2) {

	  //cout << "Here a" << endl;

	  // candidateIx2exc is an acceptable position for the new particle provided the following conditions are fulfilled:
	  // A- it is not in OriginIx2
	  // B- it follows the ordering rule, i.e. it sits in the middle of previous particle excitations, 
	  //      namely between Ix2exc[index_new - 1] (if this exists) and Ix2exc[index_new] (if this exists)
	  // C- it is next to an OriginIx2 or immediately right of Ix2exc[index_new - 1] (if this exists) 
	  //      or immediately left of Ix2exc[index_new] (if this exists) 
	  // D- it does not break the `towards the center' rule 
	  //      (it will break the rule at this point if it is created away from an OriginIx2 boundary (and thus next to a preexisting excitation), 
	  //      and if this excitation and it are not in the same sideblock (in other words: if there is a sideblock boundary between them)

	  //cout << "candidateIx2exc " << candidateIx2exc << " being tested" << endl;
	  // A
	  if (OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2exc)) {
	    // is contained in OriginIx2
	    //cout << "candidateIx2exc " << candidateIx2exc << " rejected for reason A" << endl;
	    continue; 
	  }
	  
	  //cout << "Here b" << endl;

	  // B1
	  if (currentdata.nexc[exclevel_newph] > 0 
	      && candidateIx2exc <= currentdata.Ix2exc[exclevel_newph][index_new - 1]) {
	    // there is at least one particle exc to the left, and the candidate position isn't right of Ix2exc[index_new - 1]
	    //cout << "candidateIx2exc " << candidateIx2exc << " rejected for reason B1" << endl;
	    continue;
	  }
	  
	  //cout << "Here c" << endl;

	  // B2
	  if (currentdata.nexc[exclevel_newph] > 1
	      && candidateIx2exc >= currentdata.Ix2exc[exclevel_newph][index_new]) {
	    // there is at least one particle exc to the right, and the candidate position isn't left of Ix2exc[index_new]
	    //cout << "candidateIx2exc " << candidateIx2exc << " rejected for reason B2" << endl;
	    continue;
	  }
	  
	  //cout << "Here d" << endl;

	  // C- it is next to an OriginIx2 or immediately right of Ix2exc[index_new - 1] (if this exists) 
	  //      or immediately left of Ix2exc[index_new] (if this exists) 
	  if (!(OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2exc + 2) || OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2exc - 2)) 
	      // doesn't sit next to an OriginIx2
	      && (currentdata.nexc[exclevel_newph] == 0 || candidateIx2exc != currentdata.Ix2exc[exclevel_newph][index_new - 1] + 2)
	      // doesn't sit immediately right of first particle excitation to the left 
	      && (currentdata.nexc[exclevel_newph] <= 1 || candidateIx2exc != currentdata.Ix2exc[exclevel_newph][index_new] - 2)
	      // doesn't sit immediately left of first particle excitation to the right
	      ) {
	    //cout << "candidateIx2exc " << candidateIx2exc << " rejected for reason C" << endl;
	    continue;
	  }
	  
	  //cout << "Here e" << endl;

	  // D- it does not break the `towards the center' rule
	  // In other words, if created away from a block boundary but next to a preexisting particle, 
	  // must be in same sideblock as this particle:

	  // Determine the size of the block of OriginIx2 vacancies in which this particle sits:
	  bool candidate_is_left_of_all_OriginIx2 = (candidateIx2exc < OriginIx2[descdatanewph.type[exclevel_newph] ].min()); // this makes it left-moving
	  bool candidate_is_right_of_all_OriginIx2 = (candidateIx2exc > OriginIx2[descdatanewph.type[exclevel_newph] ].max()); // this makes it right-moving
	  int nremptytoleft = 0;
	  if (candidate_is_left_of_all_OriginIx2) nremptytoleft = (candidateIx2exc - Ix2_min[descdatanewph.type[exclevel_newph] ])/2;
	  else while (!OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2exc -2*(nremptytoleft + 1)))
		 nremptytoleft++;
	  int nremptytoright = 0;
	  if (candidate_is_right_of_all_OriginIx2) nremptytoright = (Ix2_max[descdatanewph.type[exclevel_newph] ] - candidateIx2exc)/2;
	  else while (!OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2exc +2*(nremptytoright + 1)))
		 nremptytoright++;
	  // We can determine whether the new particle would be left- or right-moving
	  bool candidate_is_left_moving = candidate_is_left_of_all_OriginIx2 
	    || (!candidate_is_right_of_all_OriginIx2 && nremptytoleft >= nremptytoright);
	  bool candidate_is_right_moving = candidate_is_right_of_all_OriginIx2 
	    || (!candidate_is_left_of_all_OriginIx2 && nremptytoleft < nremptytoright);
	  // Consistency checks:
	  if (candidate_is_left_moving && candidate_is_right_moving) JSCerror("New particle moving left and right at same time");
	  if (!candidate_is_left_moving && !candidate_is_right_moving) JSCerror("New particle not moving either left or right");

	  //cout << "Here f" << endl;

	  if (candidate_is_left_moving 
	      && (currentdata.nexc[exclevel_newph] > 0 && candidateIx2exc == currentdata.Ix2exc[exclevel_newph][index_new - 1] + 2)
		  // it is created to the right of a preexisting particle excitation
	      && !OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2exc + 2) // and is not sitting at the boundary
	      && (currentdata.nexc[exclevel_newph] <= 1 || candidateIx2exc != currentdata.Ix2exc[exclevel_newph][index_new] - 2)
	      // and is not sitting just left of another preexisting particle excitation which is also left moving
	      ) 
	    {
	      //cout << "candidateIx2exc " << candidateIx2exc << " rejected for reason D1" << endl;
	      continue;
	    }	      

	  //cout << "Here g" << endl;

	  if (candidate_is_right_moving 
	      && (currentdata.nexc[exclevel_newph] > 1 && candidateIx2exc == currentdata.Ix2exc[exclevel_newph][index_new] - 2)
		  // it is created to the left of a preexisting particle excitation
	      && !OriginIx2[descdatanewph.type[exclevel_newph] ].includes (candidateIx2exc - 2) // and is not sitting at the boundary
	      && (currentdata.nexc[exclevel_newph] == 0 || candidateIx2exc != currentdata.Ix2exc[exclevel_newph][index_new - 1] + 2)
	      // and is not sitting just right of another preexisting particle excitation which is also right moving
	      ) 
	    {
	      //cout << "candidateIx2exc " << candidateIx2exc << " rejected for reason D2" << endl;
	      continue;
	    }	      

	  //cout << "\tFound a possible exc position at " << candidateIx2exc << endl;

	  // If we have reached this point, candidateIx2exc is acceptable.
	  // Immediately construct all descendents with this part position:
	  
	  //cout << "Here h" << endl;

	  // We now select the hole position closest to left or right:
	  int ihclosestleft = -1;
	  int ihclosestright = -1;
	  for (int ih = 0; ih < OriginIx2[descdatanewph.type[exclevel_newph] ].size(); ++ih) 
	    if (isgoodnewholepos[ih]) 
	      {
		/*
		//cout << "Here alpha" << endl;
		descdatanewph.Ix2old[exclevel_newph][index_new] = OriginIx2[descdatanewph.type[exclevel_newph] ][ih];
		//cout << "Here beta" << endl;
		descdatanewph.Ix2exc[exclevel_newph][index_new] = candidateIx2exc;
		//cout << "Here gamma" << endl;
		desclabelfound[ndesc_found] = Return_State_Label (descdatanewph, OriginIx2);
		//cout << "Here delta" << endl;
		desctypefound[ndesc_found] = 2;
		//cout << "For state with OriginIx2, ScanIx2:" << endl;
		//cout << OriginIx2[exclevel_newph] << endl;
		//cout << ScanIx2[exclevel_newph] << endl;
		//if (currentdata.nexc[exclevel_newph] > 0) {
		//  cout << "with current excitations (holes, particles):" << endl;
		//  cout << currentdata.Ix2old[exclevel_newph] << endl;
		//  cout << currentdata.Ix2exc[exclevel_newph] << endl;
		//}
		//cout << "\tFound new descendent of type " << desctypefound[ndesc_found] << ": label " << desclabelfound[ndesc_found] << "\tholes " << descdatanewph.Ix2old[exclevel_newph] << "\tpart " << descdatanewph.Ix2exc[exclevel_newph] << endl;
		ndesc_found++;

		if (cinbreaks) { char a;    cin >> a;}
		*/
		// New in ++G_3.1:
		if (OriginIx2[descdatanewph.type[exclevel_newph] ][ih] - candidateIx2exc < 0) // new hole is left of new particle
		  if (ihclosestleft == -1 // we hadn't found a hole previously
		      || ihclosestleft >= 0 // we had aleady found a new hole to the left
		      && OriginIx2[descdatanewph.type[exclevel_newph] ][ih] > OriginIx2[descdatanewph.type[exclevel_newph] ][ihclosestleft]) // 
		    ihclosestleft = ih;
		if (OriginIx2[descdatanewph.type[exclevel_newph] ][ih] - candidateIx2exc > 0) // new hole is right of new particle
		  if (ihclosestright == -1 // we hadn't found a hole previously
		      || ihclosestright >= 0 // we had aleady found a new hole to the right
		      && OriginIx2[descdatanewph.type[exclevel_newph] ][ih] < OriginIx2[descdatanewph.type[exclevel_newph] ][ihclosestright]) // 
		    ihclosestright = ih;
	      } // if (isgoodnewholepos[ih])
	  // for ih

	  if (ihclosestleft >= 0) { // Found a descendent with new hole to left of new particle
	    descdatanewph.Ix2old[exclevel_newph][index_new] = OriginIx2[descdatanewph.type[exclevel_newph] ][ihclosestleft];
	    descdatanewph.Ix2exc[exclevel_newph][index_new] = candidateIx2exc;
	    desclabelfound[ndesc_found] = Return_State_Label (descdatanewph, OriginIx2);
	    desctypefound[ndesc_found] = 2;
	    ndesc_found++;
	  }
	  
	  if (ihclosestright >= 0) { // Found a descendent with new hole to right of new particle
	    descdatanewph.Ix2old[exclevel_newph][index_new] = OriginIx2[descdatanewph.type[exclevel_newph] ][ihclosestright];
	    descdatanewph.Ix2exc[exclevel_newph][index_new] = candidateIx2exc;
	    desclabelfound[ndesc_found] = Return_State_Label (descdatanewph, OriginIx2);
	    desctypefound[ndesc_found] = 2;
	    ndesc_found++;
	  }

	} // for (int candidateIx2exc)

	//cout << "Here 3" << endl;

      } // for (int exclevel_newph)

    } // if (type_required == 2) 


    if (type_required == 3) {

      // We here only move the leftmost Ix2[0] left, and rightmost Ix2[0] right (this preserves iK)

      if (ScanIx2[0][0] > Ix2_min[0] && ScanIx2[0][ScanIx2[0].size() - 1] < Ix2_max[0]) {
	Vect<Vect<int> > descIx2 = ScanIx2;
	descIx2[0][0] -= 2;
	descIx2[0][descIx2[0].size() - 1] += 2;
	desclabelfound[ndesc_found] = Return_State_Label (descIx2, OriginIx2);
	desctypefound[ndesc_found] = 3;
	ndesc_found++;
      }

    } // if (type_required == 3)


    if (type_required == 4) {

      // We take the latest particle-hole pair, and move the hole to a further generalized Umklapp position

      State_Label_Data descdata = currentdata;

      // Start by finding the possible hole positions of the state without the innermost p-h pair:
      Vect<int> nexc_new = currentdata.nexc;
      if (currentdata.nexc[exclevel] < 1) JSCerror("Should not call type 3 descendent if there are no ph exc at exclevel");
      nexc_new[exclevel] -= 1; // we remove the innermost ph pair

      int innerindex = currentdata.nexc[exclevel]/2; // index of ph pair we remove

      int ntypespresent = currentdata.type.size();
      Vect<Vect<int> > Ix2old_new(ntypespresent);
      Vect<Vect<int> > Ix2exc_new(ntypespresent);
      for (int it = 0; it < ntypespresent; ++it) Ix2old_new[it] = Vect<int>(JSC::max(nexc_new[it],1));
      for (int it = 0; it < ntypespresent; ++it) Ix2exc_new[it] = Vect<int>(JSC::max(nexc_new[it],1));
      
      // Copy earlier data in:
      for (int it = 0; it < ntypespresent; ++it) {
	for (int i = 0; i < nexc_new[it]; ++i) {
	  Ix2old_new[it][i] = currentdata.Ix2old[it][i + (it == exclevel && i >= innerindex)]; 
	  Ix2exc_new[it][i] = currentdata.Ix2exc[it][i + (it == exclevel && i >= innerindex)]; 
	}
      }

      State_Label_Data data_latest_ph_removed (currentdata.type, currentdata.M, nexc_new, Ix2old_new, Ix2exc_new);
     
      Vect<bool> isgoodnewholepos = Is_Good_New_Hole_Position (OriginIx2, data_latest_ph_removed, exclevel);

      // Move hole further away from its corresponding particle, preserving the direction
      bool holemovingleft = currentdata.Ix2old[exclevel][innerindex] < currentdata.Ix2exc[exclevel][innerindex];

      int inewhole;
      if (holemovingleft) {
	inewhole = 0;
	while (OriginIx2[currentdata.type[exclevel] ][inewhole] != currentdata.Ix2old[exclevel][innerindex]) inewhole++;
	do {
	  inewhole--;
	} while (inewhole >= 0 && !isgoodnewholepos[inewhole]);
      }

      if (!holemovingleft) {
	inewhole = OriginIx2[currentdata.type[exclevel] ].size() - 1;
	while (OriginIx2[currentdata.type[exclevel] ][inewhole] != currentdata.Ix2old[exclevel][innerindex]) inewhole--;
	do {
	  inewhole++;
	} while (inewhole <= OriginIx2[currentdata.type[exclevel] ].size() - 1 && !isgoodnewholepos[inewhole]);
      }

      if (inewhole >= 0 && inewhole <= OriginIx2[currentdata.type[exclevel] ].size() - 1) { // Found a descendent
	descdata.Ix2old[exclevel][innerindex] = OriginIx2[currentdata.type[exclevel] ][inewhole];
	desclabelfound[ndesc_found] = Return_State_Label (descdata, OriginIx2);
	desctypefound[ndesc_found] = 4;
	ndesc_found++;
      }

    } // if (type_required == 4)




    Vect<string> desclabelfoundresized (ndesc_found);
    for (int i = 0; i < ndesc_found; ++i) {
      desclabelfoundresized[i] = desclabelfound[i];
    }

    //cout << "For state with label " << ScanStateLabel << ", found " << ndesc_found << " descendents of type " << type_required << endl;
    //cout << desclabelfoundresized << endl;
    if (cinbreaks) { char a;    cin >> a;}

    return(desclabelfoundresized);

  } // Vect<string> Descendents



  // Specialization for LiebLin gas:

  Vect<string> Descendents (const LiebLin_Bethe_State& ScanState, const LiebLin_Bethe_State& OriginState, int type_required)
  //string ScanStateLabel, const Vect<Vect<int> >& ScanIx2, const Vect<Vect<int> >& OriginIx2, 
  //			    const Vect<int>& Ix2_min, const Vect<int>& Ix2_max, int type_required)
  {
    Vect<Vect<int> > ScanIx2here(1);
    ScanIx2here[0] = ScanState.Ix2;
    Vect<Vect<int> > OriginIx2here(1);
    OriginIx2here[0] = OriginState.Ix2;
    Vect<int> Ix2_min(1);
    Ix2_min[0] = LIEBLIN_Ix2_MIN;
    Vect<int> Ix2_max(1);
    Ix2_max[0] = LIEBLIN_Ix2_MAX;
    Vect<string> desc_found;

    // If the state is an outer skeleton, we use a slightly modified OriginIx2 (in which the outermost 
    // rapidities at level zero are put to the outer skeleton position convention) for computing the descendents.

    //if (ScanState.Is_Outer_Skeleton()) { // since the state is an outer skeleton, we first modify the OriginIx2
    //if (ScanState.N >= 2 && ScanState.Ix2[0] == LIEBLIN_Ix2_MIN + (ScanState.N % 2) + 1 && ScanState.Ix2[ScanState.N-1] == LIEBLIN_Ix2_MAX - (ScanState.N % 2) - 1) { // since the state is an outer skeleton, we first modify the OriginIx2
    if (Is_Outer_Skeleton(ScanState)) {
      //cout << "\tDetected outer skeleton in Descendents: " << ScanState.label << endl;
      /* First (failed) attempt
      OriginIx2here[0][0] = LIEBLIN_Ix2_MIN + (ScanState.N % 2) + 3; // The OriginIx2here endpoints are set one step internally next to the Skeleton position conventions
      OriginIx2here[0][OriginState.N - 1] = LIEBLIN_Ix2_MAX - (ScanState.N % 2) - 3;
      // Obtain the descendends using this modified OriginIx2:
      Vect<string> desc_mod = Descendents (ScanState.label, ScanIx2here, OriginIx2here, Ix2_min, Ix2_max, type_required);    
      // Translate the obtained descendends into ones corresponding to OriginIx2:
      LiebLin_Bethe_State LabellingState = ScanState;
      for (int idesc = 0; idesc < desc_mod.size(); ++idesc) {
	LabellingState.Set_to_Label (desc_mod[idesc], OriginIx2here[0]); // Set quantum numbers according to OriginIx2here
	LabellingState.Set_Label_from_Ix2 (OriginState.Ix2); // Now reset label on OriginIx2 from the correctly set Ix2
	desc_mod[idesc] = LabellingState.label;
	cout << "\tidesc " << idesc << "\tlabel " << LabellingState.label << "\tIx2 " << LabellingState.Ix2 << endl;
      }
      return(desc_mod);
      */
      // Construct and descend a state with 2 less particles:
      if (ScanState.N < 3) JSCerror("Skeleton descendent logic at fixed iK not implemented for N < 3.");
      LiebLin_Bethe_State ReducedScanState (ScanState.c_int, ScanState.L, ScanState.N - 2);
      LiebLin_Bethe_State ReducedOriginState (ScanState.c_int, ScanState.L, ScanState.N - 2);
      for (int i = 0; i < ScanState.N - 2; ++i) {
	ReducedScanState.Ix2[i] = ScanState.Ix2[i+1];
	ReducedOriginState.Ix2[i] = OriginState.Ix2[i+1];
      }
      ReducedScanState.Set_Label_from_Ix2 (ReducedOriginState.Ix2);
      Vect<string> desc_mod = Descendents (ReducedScanState, ReducedOriginState, type_required);          
      // Now translate results back to ScanState: we return outer skeleton states 
      LiebLin_Bethe_State LabellingState = ScanState;
      for (int idesc = 0; idesc < desc_mod.size(); ++idesc) {
	ReducedScanState.Set_to_Label (desc_mod[idesc], ReducedOriginState.Ix2);
	// Now set all quantum numbers of LabellingState:
	LabellingState.Ix2[0] = LIEBLIN_Ix2_MIN + (ScanState.N % 2) + 1;
	for (int i = 0; i < ScanState.N - 2; ++i) LabellingState.Ix2[i+1] = ReducedScanState.Ix2[i];
	LabellingState.Ix2[ScanState.N - 1] = LIEBLIN_Ix2_MAX - (ScanState.N % 2) - 1;
	LabellingState.Set_Label_from_Ix2 (OriginState.Ix2); // Now reset label on OriginIx2 from the correctly set Ix2
	desc_mod[idesc] = LabellingState.label;
	//cout << "\tidesc " << idesc << "\tlabel " << LabellingState.label << "\tIx2 " << LabellingState.Ix2 << endl;
      }
      //cout << "\tFound " << desc_mod.size() << "descendents, " << endl;
      //cout << "\t" << desc_mod << endl;
      desc_found = desc_mod;
    }

    // If not outer skeleton, just return straight descendents
    else desc_found = Descendents (ScanState.label, ScanIx2here, OriginIx2here, Ix2_min, Ix2_max, type_required);  

    //return(Descendents (ScanState.label, ScanIx2here, OriginIx2here, Ix2_min, Ix2_max, type_required));  
    return(desc_found);
  }

  // Specialization for Heisenberg:

  Vect<string> Descendents (const Heis_Bethe_State& ScanState, const Heis_Bethe_State& OriginState, int type_required)
  {
    return(Descendents (ScanState.label, ScanState.Ix2, OriginState.Ix2, ScanState.base.Ix2_min, ScanState.base.Ix2_max, type_required));
  }






  // ABACUS++T version:
  /*
  //Vect<string> Descendents (const LiebLin_Bethe_State& ScanState, const LiebLin_Bethe_State& OriginState)
  Vect<string> Descendents (const LiebLin_Bethe_State& ScanState, const LiebLin_Bethe_State& OriginState, int type_required)
  //Descendent_Data Descendents (const LiebLin_Bethe_State& ScanState, const LiebLin_Bethe_State& OriginState, int type_required)
  {

    // THESE EXPLANATIONS ARE DEPRECATED 

    // A given state is labeled by pairs (Ix2old, Ix2exc) of modified quantum numbers,
    // say (Ix2old[0][0], Ix2exc[0][0]) ... (Ix2old[0][nexc[0] ], Ix2exc[0][nexc[0] ]) at level 0
    // with the constraint that Ix2old and Ix2exc are strictly ordered.
    // Higher levels are treated similarly.

    // By definition, the descendents are constructed according to two possibilities:

    // First, if type_required == 0, by taking the inner(right)most index particle (so at index (nexc[i]+1)/2)
    // and moving it one step further in the direction it was moving, if possible.
    // The expected data_value in this case is simply data_value_ref of the ancestor.

    // Second, if type_required == 1, by adding another Ix2old Ix2exc pair,
    // letting the hole take all possible values of Ix2old between Ix2old[i][(nexc[i]+1)/2 ] and Ix2old[i][(nexc[i]+1)/2 + 1]
    // and putting the corresponding Ix2exc to the first available position on the left or right (if any),
    // without annihilating another hole. In other words, we add a Ix2old Ix2exc pair in the (right)middle, respecting the ordering rule.
    // The expected data_value in this case is the ancestor's data_value multiplied by the cost of adding a particle-hole (ph_cost).

    // IMPORTANT NOTE: THE IMPLEMENTATION IS FOR NOW ONLY VALID FOR LIEB-LINIGER.
    // The logic is however extensible to the case of multiple base levels.

    // Number of descendents: 
    int ndesc_possible = (type_required == 0 ? 1 : ScanState.N * 2); // upper bound guess; should be refined.
    Vect<string> desclabelfound (ndesc_possible);
    Vect<int> desctypefound (ndesc_possible);
    int ndesc_found = 0;

    // First read the current state data:
    State_Label_Data currentdata = Read_State_Label (ScanState.label, OriginState.Ix2);

    if (type_required == 0) {
      // We move the inner(right)most hole one more step towards the center of the block of Ix2 in which it sits.

      // Is there already an excitation at level 0? If so, move inner(right)most particle one step further.
      if (currentdata.nexc[0] > 0) {

	//cout << "Here a" << endl;

	// Produce a new state_label_data:
	State_Label_Data descdata = Read_State_Label (ScanState.label, OriginState.Ix2);
	// Identify the inner(right)most excitation:
	int innerindex = currentdata.nexc[0]/2;

	int newholepos = currentdata.Ix2old[0][innerindex];

	// Determine the size of the block of OriginState.Ix2 in which this hole sits:
	int nroccupiedtoleft = 0;
	while (OriginState.Ix2.includes (currentdata.Ix2old[0][innerindex] -2*(nroccupiedtoleft + 1)))
	//while (ScanState.Ix2.includes (currentdata.Ix2old[0][innerindex] -2*(nroccupiedtoleft + 1)))
	//while (OriginState.Ix2.includes (currentdata.Ix2old[0][innerindex] -2*(nroccupiedtoleft + 1)) 
	//&& ScanState.Ix2.includes (currentdata.Ix2old[0][innerindex] -2*(nroccupiedtoleft + 1)))
	  ////&& !currentdata.Ix2old[0].includes(currentdata.Ix2old[0][innerindex] -2*(nroccupiedtoleft + 1))) 
	  nroccupiedtoleft++;
	int nroccupiedtoright = 0;
	while (OriginState.Ix2.includes (currentdata.Ix2old[0][innerindex] +2*(nroccupiedtoright + 1)))
	//while (ScanState.Ix2.includes (currentdata.Ix2old[0][innerindex] +2*(nroccupiedtoright + 1)))
	//while (OriginState.Ix2.includes (currentdata.Ix2old[0][innerindex] +2*(nroccupiedtoright + 1))
	//     && ScanState.Ix2.includes (currentdata.Ix2old[0][innerindex] +2*(nroccupiedtoright + 1)))
	  ////&& !currentdata.Ix2old[0].includes(currentdata.Ix2old[0][innerindex] +2*(nroccupiedtoright + 1))) 
	  nroccupiedtoright++;

	//cout << "Here b" << endl;

	//cout << "holeIx2 = " << currentdata.Ix2old[0][innerindex] << "\tnroccupiedtoleft = " << nroccupiedtoleft << "\tnroccupiedtoright = " << nroccupiedtoright << endl;
	// Move the hole further in: 
	// Requirements for end configuration: hole must be further near middle, so
	// if nroccupiedtoleft < nroccupiedtoright, move towards the right.
	// if nroccupiedtoleft > nroccupiedtoright, move towards the left.
	// If the two are then equal, choose hole moved from right to left.
	// if moved towards the right: nroccupiedtoleft + 1 < nroccupiedtoright - 1
	// if moved towards the left:  nroccupiedtoleft - 1 <= nroccupiedtoright + 1

	if (//nroccupiedleft < nroccupiedtoright && // redundant from next condition
	    nroccupiedtoleft - nroccupiedtoright + 2 < 0) newholepos += 2;
	else if (// nroccupiedright > nroccupiedtoright && // redundant from next condition
		 nroccupiedtoleft - nroccupiedtoright - 2 >= 0) newholepos -= 2;

	if (newholepos != currentdata.Ix2old[0][innerindex] // we have successfully moved the hole
	    && !currentdata.Ix2old[0].includes(newholepos) // new hole position is not already taken
	    ) 
	  { // we have found a descendent
	    descdata.Ix2old[0][innerindex] = newholepos;
	    desclabelfound[ndesc_found] = Return_State_Label (descdata, OriginState.Ix2);
	    desctypefound[ndesc_found++] = 0;
	  }

	//cout << "Here c" << endl;

      } // if (nexc > 0)
    } // if (type_required == 0)


    if (type_required == 1) {
      // Is there already an excitation at level 0? If so, move inner(right)most particle one step further.
      if (currentdata.nexc[0] > 0) {
	// Produce a new state_label_data:
	State_Label_Data descdata = Read_State_Label (ScanState.label, OriginState.Ix2);
	// Identify the inner(right)most excitation:
	int innerindex = currentdata.nexc[0]/2;

	int newpartpos = currentdata.Ix2exc[0][innerindex];

	// Determine the size of the vacancy block of OriginState.Ix2 in which this particle sits:
	int nremptytoleft = 0;
	while (!OriginState.Ix2.includes (currentdata.Ix2exc[0][innerindex] -2*(nremptytoleft+1)))
	  nremptytoleft++;
	int nremptytoright = 0;
	while (!OriginState.Ix2.includes (currentdata.Ix2exc[0][innerindex] + 2*(nremptytoright+1)))
	  nremptytoright++;
	// BUG HERE: doesn't terminate if there is no particle to the left or right...


	if (nremptytoleft - nremptytoright + 2 < 0) newpartpos += 2;
	else if (nremptytoleft - nremptytoright + 2 >= 0) newpartpos -= 2;

	if (newpartpos != currentdata.Ix2exc[0][innerindex] // we have successfully moved the hole
	    && !currentdata.Ix2exc[0].includes(newpartpost) // new particle position is not already taken
	    )
	  { // we have found a descendent
	    descdata.Ix2exc[0][innerindex] = newpartpos;
	    desclabelfound[ndesc_found] = Return_State_Label (descdata, OriginState.Ix2);
	    desctypefound[ndesc_found++] = 1;
	  }

	// FROM ++T, DEPRECATED IN ++G:
	// Can this Ix2exc be moved one more step in the same direction? Try it out.
	////descdata.Ix2exc[0][innerindex] += (currentdata.Ix2exc[0][innerindex] > currentdata.Ix2old[0][innerindex] ? 2 : -2);
	//int newpartpos = currentdata.Ix2exc[0][innerindex] + (currentdata.Ix2exc[0][innerindex] > currentdata.Ix2old[0][innerindex] ? 2 : -2);
	// Move until we get a consistent new configuration:
	// First, this quantum number was not already occupied
	// Second, this quantum number does not annihilate one of the holes:
	//while (ScanState.Ix2.includes(newpartpos) || currentdata.Ix2old[0].includes(newpartpos)) 
	// newpartpos += (currentdata.Ix2exc[0][innerindex] > currentdata.Ix2old[0][innerindex] ? 2 : -2);
	
	// If this new particle position is to the right of the Ix2exc of one index lower, we've got a descendent:
	//if (innerindex == 0 || newpartpos > currentdata.Ix2exc[0][innerindex - 1]
	//  //&& abs(newpartpos) <= 13 // TEMPORARY LIMIT: INDUCE FINITE NR OF STATES
	  //  && (innerindex == currentdata.nexc[0] - 1 || newpartpos < currentdata.Ix2exc[0][innerindex + 1])) {
	  // We have a descendent,
	  ////cout << "\tFound increased displacement at innerindex " << innerindex << endl;
	  ////cout << descdata.Ix2exc[0][innerindex] << " to " << newpartpos << endl;
	  //descdata.Ix2exc[0][innerindex] = newpartpos;
	  //desclabelfound[ndesc_found] = Return_State_Label (descdata, OriginState.Ix2);
	  //desctypefound[ndesc_found++] = 1;
	//}
	
      } //if (currentdata.nexc[0] > 0)
    } // if (type_required == 1)


    if (type_required == 2) {
      // Now add a new p-h pair at the inner(right)most position, scanning the hole position immediately 
      // but putting the particle to the first available left and right positions (if any).

      // type_required == 2 means that the hole must be either at the edge of an OriginIx2 domain, or next to an existing hole.

      // Define the objects for the newstatedata:
      Vect<int> type_new = currentdata.type;
      Vect<int> M_new = currentdata.M;
      Vect<int> nexc_new = currentdata.nexc;
      nexc_new[0] += 1; // we drill one more particle-hole pair at level 0
      int index_new = (currentdata.nexc[0] + 1)/2; // we put the new p-h pair at index index_new.
      
      int ntypespresent = 1; // only one type for LiebLin
      Vect<Vect<int> > Ix2old_new(ntypespresent);
      Vect<Vect<int> > Ix2exc_new(ntypespresent);
      for (int it = 0; it < ntypespresent; ++it) Ix2old_new[it] = Vect<int>(JSC::max(nexc_new[it],1));
      for (int it = 0; it < ntypespresent; ++it) Ix2exc_new[it] = Vect<int>(JSC::max(nexc_new[it],1));
      
      // Copy earlier data in:
      for (int it = 0; it < ntypespresent; ++it) {
	for (int i = 0; i < currentdata.nexc[it]; ++i) {
	  Ix2old_new[it][i + (i >= index_new)] = currentdata.Ix2old[it][i]; 
	  Ix2exc_new[it][i + (i >= index_new)] = currentdata.Ix2exc[it][i]; 
	}
	//cout << "current Ix2old: " << currentdata.Ix2old[0] << endl;
	//cout << "current Ix2exc: " << currentdata.Ix2exc[0] << endl;
	//cout << "allocated new Ix2old: " << Ix2old_new[0] << endl;
	//cout << "allocated new Ix2exc: " << Ix2exc_new[0] << endl;
      }
      
      State_Label_Data descdatanewph (type_new, M_new, nexc_new, Ix2old_new, Ix2exc_new);
      
      //cout << "\ttype: " << descdatanewph.type << endl;
      //cout << "\tM: " << descdatanewph.M << endl;
      //cout << "\tnexc: " << descdatanewph.nexc << endl;
      //cout << "\tIx2old: " << descdatanewph.Ix2old[0] << endl;
      //cout << "\tIx2new: " << descdatanewph.Ix2exc[0] << endl;
      
      // Now find all the possible positions for the new hole:
      // The new hole should be between Ix2old_new[0][innerindex - 1] and Ix2old_new[0][innerindex + 1]
      // if those exist (i.e., modulo lack of constraint at the left/right boundaries).
      int newpartpos;
      for (int i = 0; i < ScanState.N; ++i) {

	// Determine the size of the block of OriginState.Ix2 in which this hole sits:
	int nroccupiedtoleft = 0;
	while (OriginState.Ix2.includes (ScanState.Ix2[i] -2*(nroccupiedtoleft + 1)))
	  nroccupiedtoleft++;
	int nroccupiedtoright = 0;
	while (OriginState.Ix2.includes (ScanState.Ix2[i] +2*(nroccupiedtoright + 1)))
	  nroccupiedtoright++;
	// This will be used in the next conditions to check whether the new hole would sit in a left or right block.

	if ((currentdata.nexc[0] == 0 // no earlier p-h in state so no left boundary for new hole; index_new == 0
	     || ScanState.Ix2[i] > currentdata.Ix2old[0][index_new - 1])
	    && (currentdata.nexc[0] <= 1 // at most one p-h in state so no right boundary for new hole; index_new <= 1
		|| ScanState.Ix2[i] < currentdata.Ix2old[0][index_new])
	    //&& (currentdata.nexc[0] == 0 || !currentdata.Ix2exc[0].includes(ScanState.Ix2[i])) 
	    // new hole is not one of the already existing p excitations // FULFILLED BY ABOVE CONDITIONS IF EXC ARE ORDERED
	    //&& (!ScanState.Ix2.includes(ScanState.Ix2[i] - 2) || !ScanState.Ix2.includes(ScanState.Ix2[i] + 2)) 
	    && (
		// New hole must be created on the boundary of an OriginState Ix2 domain, 
		// or next to a preexisting hole in the domain.
		(!OriginState.Ix2.includes(ScanState.Ix2[i] - 2)) 
		// new hole is on the left boundary of an OriginState Ix2 domain.
		|| !OriginState.Ix2.includes(ScanState.Ix2[i] + 2) 
		// new hole is on the right boundary of an OriginState Ix2 domain.
		//|| (currentdata.nexc[0] > 0 && ScanState.Ix2[i] == currentdata.Ix2old[0][index_new - 1] + 2)
		|| type_required == 2 && (currentdata.nexc[0] > 0 && nroccupiedtoleft < nroccupiedtoright 
		    && ScanState.Ix2[i] == currentdata.Ix2old[0][index_new - 1] + 2)
		// new hole is in a left domain and immediately to the right of first hole to its left
		//|| (currentdata.nexc[0] > 1 && ScanState.Ix2[i] == currentdata.Ix2old[0][index_new] - 2)
		|| type_required == 2 && (currentdata.nexc[0] > 1 && nroccupiedtoleft >= nroccupiedtoright 
		    && ScanState.Ix2[i] == currentdata.Ix2old[0][index_new] - 2)
		// new hole is in a right domain and immediately to the left of first hole to its right
		)
	    ) {

	  bool acceptable_hole = true;

	  if (currentdata.nexc[0] > 0 && currentdata.Ix2exc[0].includes(ScanState.Ix2[i])) acceptable_hole = false;

	  // Determine the size of the block of Ix2 in which this hole sits:
	  int nroccupiedtoleft = 0;
	  while (OriginState.Ix2.includes (ScanState.Ix2[i] -2*(nroccupiedtoleft + 1)))
	    nroccupiedtoleft++;
	  int nroccupiedtoright = 0;
	  while (OriginState.Ix2.includes (ScanState.Ix2[i] +2*(nroccupiedtoright + 1)))
	    nroccupiedtoright++;
	  
	  // The hole is unacceptable if it breaks the `towards the block center' rule:
	  if (nroccupiedtoleft < nroccupiedtoright // new hole would be in left part of block
	      && OriginState.Ix2.includes(ScanState.Ix2[i] - 2) // and is not at the boundary
	      && currentdata.nexc[0] > 0 && ScanState.Ix2[i] != currentdata.Ix2old[0][index_new - 1] + 2)
	    // and it's not immediately to the left of a preexisting hole
	    acceptable_hole = false;
	  if (nroccupiedtoleft >= nroccupiedtoright // new hole could be in right part of block
	      && OriginState.Ix2.includes(ScanState.Ix2[i] + 2) // and is not at the boundary
	      && currentdata.nexc[0] > 2 && ScanState.Ix2[i] != currentdata.Ix2old[0][index_new] - 2)
	    // and it's not immediately to the right of a preexisting hole
	    acceptable_hole = false;

	  if (acceptable_hole) {
	    //cout << "Found a possible hole position at index " << i << " and Ix2 " << ScanState.Ix2[i] << endl;
	    
	    // ScanState.Ix2[i] is an allowable new hole position.


	    // We now look for all possible particle positions, 
	    // the allowable ones being either at the edge of a block, or next to an existing particle:


	    // Find the first available particle position to the left:
	    newpartpos = ScanState.Ix2[i] - 2;
	    // This must be lower than the Ix2exc of one index higher (if any):
	    if (currentdata.nexc[0] >= 2) newpartpos = JSC::min(newpartpos, currentdata.Ix2exc[0][index_new] - 2);
	    // The new particle position must not already be occupied in ScanState, 
	    // and must not be at one of the already specified holes
	    while (ScanState.Ix2.includes(newpartpos) 
		   || (currentdata.nexc[0] > 0 && currentdata.Ix2old[0].includes(newpartpos))) newpartpos -= 2;
	    // If this new particle position is to the right of the Ix2exc of one index lower (if any), we've got a descendent:
	    if (currentdata.nexc[0] == 0 || newpartpos > currentdata.Ix2exc[0][index_new - 1]) {
	      //cout << "\tFound a particle position (moving L) at " << newpartpos << " with index " << index_new << endl;
	      //if (abs(newpartpos) <= 13) { // TEMPORARY LIMIT: INDUCE FINITE NR OF STATES
	      descdatanewph.Ix2old[0][index_new] = ScanState.Ix2[i];
	      descdatanewph.Ix2exc[0][index_new] = newpartpos;
	      //cout << "ScanState.Ix2[i] = " << ScanState.Ix2[i] << "\tnewpartpos = " << newpartpos << endl;
	      //cout << "\tIx2old: " << descdatanewph.Ix2old[0] << endl;
	      //cout << "\tIx2new: " << descdatanewph.Ix2exc[0] << endl;
	      desclabelfound[ndesc_found] = Return_State_Label (descdatanewph, OriginState.Ix2);
	      desctypefound[ndesc_found++] = type_required;
	      //cout << "\tLabel found = " << desclabelfound[ndesc_found - 1] << endl;
	      //} // TEMP
	    }


	    // Now find the first particle position to the right:
	    newpartpos = ScanState.Ix2[i] + 2;
	    // This must be higher than the Ix2exc of one index lower (if any):
	    if (index_new > 0) newpartpos = JSC::max(newpartpos, currentdata.Ix2exc[0][index_new - 1] + 2);
	    // The new particle position must not already be occupied in ScanState, 
	    // and must not be at one of the already specified holes
	    while (ScanState.Ix2.includes(newpartpos) 
		   || (currentdata.nexc[0] > 0 && currentdata.Ix2old[0].includes(newpartpos))) newpartpos += 2;
	    // If this new particle position is to the left of the Ix2exc of one index higher (if any), we've got a descendent:
	    if (currentdata.nexc[0] <= 1 || newpartpos < currentdata.Ix2exc[0][index_new]) {
	      //// If this new particle position is to the left of the Ix2exc of one index higher (if any), got descendent:
	      //if (currentdata.nexc[0] <= 1 || newpartpos < currentdata.Ix2exc[0][index_new]) {
	      //cout << "\tFound a particle position (moving R) at " << newpartpos << " with index " << index_new << endl;

	      //if (abs(newpartpos) <= 13) { // TEMPORARY LIMIT: INDUCE FINITE NR OF STATES
	      descdatanewph.Ix2old[0][index_new] = ScanState.Ix2[i];
	      descdatanewph.Ix2exc[0][index_new] = newpartpos;
	      //cout << "ScanState.Ix2[i] = " << ScanState.Ix2[i] << "\tnewpartpos = " << newpartpos << endl;
	      //cout << "\tIx2old: " << descdatanewph.Ix2old[0] << endl;
	      //cout << "\tIx2new: " << descdatanewph.Ix2exc[0] << endl;
	      desclabelfound[ndesc_found] = Return_State_Label (descdatanewph, OriginState.Ix2);
	      desctypefound[ndesc_found++] = type_required;
	      //cout << "\tLabel found = " << desclabelfound[ndesc_found - 1] << endl;
	      //} // TEMP
	    }
	  } // if (acceptable_hole
	}
      } // for i
    } // if (type_required == 2 || type_required == 3)


    Vect<string> desclabelfoundresized (ndesc_found);
    //Vect_DP desctypefoundresized (ndesc_found);
    for (int i = 0; i < ndesc_found; ++i) {
      desclabelfoundresized[i] = desclabelfound[i];
      //desctypefoundresized[i] = descvalue[i];
    }

    //for (int i = 0; i < ndesc_found; ++i) if (desclabelfound[i] == "16_2_13@17:15@19") cout << "16_2_13@17:15@19 descended from " << ScanState.label << endl;

    //if (ScanState.label == "16_0_") cout << "State 16_0_: descendents " << desclabelfound << endl;
    //if (ScanState.label == "16_1_13@17") cout << "State 16_1_13@17: descendents " << desclabelfound << endl;

    //cout << "Found " << ndesc_found << " descendents, " << desclabelfoundresized << endl;

    //Descendent_Data descdata();

    //descdata.label = desclabelfoundresized;
    //descdata.type = desctypefoundresized;
    
    return(desclabelfoundresized);
    //return(descdata);
  }


  //Vect<string> Descendents (const Heis_Bethe_State& ScanState, const Heis_Bethe_State& OriginState)
  Vect<string> Descendents (const Heis_Bethe_State& ScanState, const Heis_Bethe_State& OriginState, int type_required)
  //Descendent_Data Descendents (const Heis_Bethe_State& ScanState, const Heis_Bethe_State& OriginState, int type_required);
  {
    // NOT IMPLEMENTED YET
    JSCerror("Descendents for Heis not implemented yet.");
    return(Vect<string>());
  }

  */

} // namespace JSC

