/**********************************************************

This software is part of J.-S. Caux's ABACUS++ library.

Copyright (c).

-----------------------------------------------------------

File:  TBA_XXZ.cc

Purpose:  TBA for the gapless XXZ antiferromagnet.



***********************************************************/

#include "JSC.h"

using namespace std;
using namespace JSC;

namespace JSC {

  // First, define some useful kernels:

  DP XXZ_phi1_kernel (DP zeta, DP lambda)
  {
    return(2.0 * atan(tanh(lambda)/tan(0.5*zeta)));
  }

  DP XXZ_phi2_kernel (DP zeta, DP lambda)
  {
    return(2.0 * atan(tanh(lambda)/tan(zeta)));
  }

  DP XXZ_a1_kernel (DP sinzeta, DP coszeta, DP lambda) 
  {
    DP expmin2abslambda = exp(-2.0 * fabs(lambda));
    
    DP answer = sinzeta * expmin2abslambda
      /(PI * (0.5 * (1.0 + expmin2abslambda * expmin2abslambda) - coszeta * expmin2abslambda));

    return(answer);
  }

  DP XXZ_da1dlambda_kernel (DP sinzeta, DP coszeta, DP lambda) 
  {
    DP expmin2abslambda = exp(-2.0 * fabs(lambda));
    DP signlambda = lambda >= 0.0 ? 1.0 : -1.0;
    DP answer = -sinzeta * signlambda * expmin2abslambda * (1.0 - expmin2abslambda * expmin2abslambda)
      /(PI * pow((0.5 * (1.0 + expmin2abslambda * expmin2abslambda) 
		  - coszeta * expmin2abslambda), 2.0));

    return(answer);
  }

  DP XXZ_a2_kernel (DP sin2zeta, DP cos2zeta, DP lambda) 
  {
    DP expmin2abslambda = exp(-2.0 * fabs(lambda));
    
    DP answer = sin2zeta * expmin2abslambda
      /(PI * (0.5 * (1.0 + expmin2abslambda * expmin2abslambda) - cos2zeta * expmin2abslambda));

    return(answer);
  }

  void Iterate_XXZ_rhotot_GS (DP zeta, DP B, Root_Density& rhotot_GS)
  {
    // Performs an iteration of the TBA equation for rhotot_GS:

    // First, copy the actual values into the previous ones:
    for (int i = 0; i < rhotot_GS.Npts; ++i) rhotot_GS.prev_value[i] = rhotot_GS.value[i];

    // Calculate the new values:
    DP conv = 0.0;
    DP sinzeta = sin(zeta);
    DP sin2zeta = sin(2.0*zeta);
    DP coszeta = cos(zeta);
    DP cos2zeta = cos(2.0*zeta);
    for (int i = 0; i < rhotot_GS.Npts; ++i) {
      // First, calculate the convolution
      conv = 0.0;      
      for (int j = 0; j < rhotot_GS.Npts; ++j)
	conv += fabs(rhotot_GS.lambda[j]) > B ? 0.0 : 
	  XXZ_a2_kernel (sin2zeta, cos2zeta, rhotot_GS.lambda[i] - rhotot_GS.lambda[j]) * rhotot_GS.prev_value[j] * rhotot_GS.dlambda[j];
      rhotot_GS.value[i] = XXZ_a1_kernel(sinzeta, coszeta, rhotot_GS.lambda[i]) - conv;
    }      

    // Calculate the sum of differences:
    rhotot_GS.diff = 0.0;
    for (int i = 0; i < rhotot_GS.Npts; ++i) 
      rhotot_GS.diff += fabs(rhotot_GS.value[i] - rhotot_GS.prev_value[i]);

    return;
  }

  Root_Density XXZ_rhotot_GS (DP Delta, DP B, DP lambdamax, int Npts, DP req_prec) 
  {

    // This function returns a Root_Density object corresponding to the ground state
    // root distribution of the XXZ model.

    if (Delta >= 1.0 || Delta <= 0.0) JSCerror("Delta out of bounds in XXZ_rhotot_GS.");

    DP zeta = acos(Delta);

    Root_Density rhotot_GS(Npts, lambdamax);

    // We now attempt to find a solution...

    int niter = 0;
    do {
      Iterate_XXZ_rhotot_GS (zeta, B, rhotot_GS);
      //cout << rhotot_GS.diff << endl;
      niter ++;
    } while (rhotot_GS.diff > req_prec);

    //cout << "rhotot: niter = " << niter << endl;
    return(rhotot_GS);
  }

  DP Return_GS_Sz_tot_value (DP B, Root_Density& rhotot_GS)
  {
    DP integral = 0.0;
    for (int i = 0; i < rhotot_GS.Npts; ++i)
      integral += fabs(rhotot_GS.lambda[i]) > B ? 0.0 : rhotot_GS.value[i] * rhotot_GS.dlambda[i];

    return(0.5 - integral);
  }

  void Iterate_XXZ_eps_GS (DP zeta, DP Hz, Root_Density& eps_GS)
  {
    // Performs an iteration of the TBA equation for eps_GS:

    // First, copy the actual values into the previous ones:
    for (int i = 0; i < eps_GS.Npts; ++i) eps_GS.prev_value[i] = eps_GS.value[i];

    // Calculate the new values:
    DP conv = 0.0;
    DP sinzeta = sin(zeta);
    DP sin2zeta = sin(2.0*zeta);
    DP coszeta = cos(zeta);
    DP cos2zeta = cos(2.0*zeta);
    for (int i = 0; i < eps_GS.Npts; ++i) {
      // First, calculate the convolution
      conv = 0.0;
      for (int j = 0; j < eps_GS.Npts; ++j)
	conv += eps_GS.prev_value[j] > 0.0 ? 0.0 : 
	  XXZ_a2_kernel (sin2zeta, cos2zeta, eps_GS.lambda[i] - eps_GS.lambda[j]) 
	  * eps_GS.prev_value[j] * eps_GS.dlambda[j];
      eps_GS.value[i] = Hz - PI * sinzeta * XXZ_a1_kernel(sinzeta, coszeta, eps_GS.lambda[i]) - conv;
      //cout << i << "\t" << eps_GS.lambda[i] << "\t" << eps_GS.value[i] << "\t" << conv << endl;
    }      

    // Calculate the sum of differences:
    eps_GS.diff = 0.0;
    for (int i = 0; i < eps_GS.Npts; ++i) 
      eps_GS.diff += fabs(eps_GS.value[i] - eps_GS.prev_value[i]);

    return;
  }

  Root_Density XXZ_eps_GS (DP Delta, DP Hz, DP lambdamax, int Npts, DP req_prec) 
  {

    // This function returns a Root_Density object corresponding to the ground state
    // epsilon function of the XXZ model.

    if (Delta >= 1.0 || Delta <= 0.0) JSCerror("Delta out of bounds in XXZ_eps_GS.");

    DP zeta = acos(Delta);

    Root_Density eps_GS(Npts, lambdamax);

    // We now attempt to find a solution...

    int niter = 0;
    do {
      Iterate_XXZ_eps_GS (zeta, Hz, eps_GS);
      //cout << eps_GS.diff << endl;
      niter++;
      //cout << niter << "\t" << eps_GS.diff << endl;
    } while (eps_GS.diff > req_prec);

  //cout << "eps: niter = " << niter << endl;
    return(eps_GS);
  }

  void Iterate_XXZ_depsdlambda_GS (DP zeta, DP B, Root_Density& depsdlambda_GS)
  {
    // Performs an iteration of the TBA equation for depsdlambda_GS:

    // First, copy the actual values into the previous ones:
    for (int i = 0; i < depsdlambda_GS.Npts; ++i) depsdlambda_GS.prev_value[i] = depsdlambda_GS.value[i];

    // Calculate the new values:
    DP conv = 0.0;
    DP sinzeta = sin(zeta);
    DP sin2zeta = sin(2.0*zeta);
    DP coszeta = cos(zeta);
    DP cos2zeta = cos(2.0*zeta);
    for (int i = 0; i < depsdlambda_GS.Npts; ++i) {
      // First, calculate the convolution
      conv = 0.0;
      for (int j = 0; j < depsdlambda_GS.Npts; ++j)
	conv += fabs(depsdlambda_GS.lambda[j]) > B ? 0.0 : 
	  XXZ_a2_kernel (sin2zeta, cos2zeta, depsdlambda_GS.lambda[i] - depsdlambda_GS.lambda[j]) 
	  * depsdlambda_GS.prev_value[j] * depsdlambda_GS.dlambda[j];
      depsdlambda_GS.value[i] = -PI * sinzeta * XXZ_da1dlambda_kernel(sinzeta, coszeta, depsdlambda_GS.lambda[i]) - conv;
    }      

    // Calculate the sum of differences:
    depsdlambda_GS.diff = 0.0;
    for (int i = 0; i < depsdlambda_GS.Npts; ++i) 
      depsdlambda_GS.diff += fabs(depsdlambda_GS.value[i] - depsdlambda_GS.prev_value[i]);

    return;
  }

  Root_Density XXZ_depsdlambda_GS (DP Delta, DP B, DP lambdamax, int Npts, DP req_prec) 
  {

    // This function returns a Root_Density object corresponding to the ground state
    // (d/d\lambda) epsilon function of the XXZ model.

    if (Delta >= 1.0 || Delta <= 0.0) JSCerror("Delta out of bounds in XXZ_rho_GS.");

    DP zeta = acos(Delta);

    Root_Density depsdlambda_GS(Npts, lambdamax);

    // We now attempt to find a solution...

    int niter = 0;
    do {
      Iterate_XXZ_depsdlambda_GS (zeta, B, depsdlambda_GS);
      //cout << depsdlambda_GS.diff << endl;
      niter++;
    } while (depsdlambda_GS.diff > req_prec);

    //cout << "depsdlambda: niter = " << niter << endl;
    return(depsdlambda_GS);
  }

  void Iterate_XXZ_b2BB_lambda_B (DP zeta, DP B, Root_Density& b2BB)
  {
    // Calculates the vector corresponding to b2BB (lambda,B)

    // First, copy the actual values into the previous ones:
    for (int i = 0; i < b2BB.Npts; ++i) b2BB.prev_value[i] = b2BB.value[i];

    // Calculate the new values:
    DP conv = 0.0;
    DP sin2zeta = sin(2.0*zeta);
    DP cos2zeta = cos(2.0*zeta);
    for (int i = 0; i < b2BB.Npts; ++i) {
      if (fabs(b2BB.lambda[i]) > B) b2BB.value[i] = 0.0;
      else {
	// First, calculate the convolution
	conv = 0.0;
	for (int j = 0; j < b2BB.Npts; ++j)
	  conv += fabs(b2BB.lambda[j]) > B ? 0.0 : 
	    XXZ_a2_kernel (sin2zeta, cos2zeta, b2BB.lambda[i] - b2BB.lambda[j]) 
	    * b2BB.prev_value[j] * b2BB.dlambda[j];
	b2BB.value[i] = -XXZ_a2_kernel(sin2zeta, cos2zeta, b2BB.lambda[i] - B) 
	  - conv;
      }      
    }
    // Calculate the sum of differences:
    b2BB.diff = 0.0;
    for (int i = 0; i < b2BB.Npts; ++i) 
      b2BB.diff += fabs(b2BB.value[i] - b2BB.prev_value[i]);

    return;
  }

  Root_Density XXZ_b2BB_lambda_B (DP Delta, DP B, DP lambdamax, int Npts, DP req_prec) 
  {

    // This function returns a Root_Density object corresponding to the inverse
    // kernel b2BB (lambda, B) used in Kbar function.

    if (Delta >= 1.0 || Delta <= 0.0) JSCerror("Delta out of bounds in XXZ_Kbackflow_GS.");

    DP zeta = acos(Delta);

    Root_Density b2BB_lambda_B(Npts, lambdamax);

    // We now attempt to find a solution...

    int niter = 0;
    do {
      Iterate_XXZ_b2BB_lambda_B (zeta, B, b2BB_lambda_B);
      //cout << eps_GS.diff << endl;
      niter++;
    } while (b2BB_lambda_B.diff > req_prec);

    //if (lambda_p + lambda_h + 2.0*B > 0.0) cout << "Nonzero backflow possible." << endl;
    //cout << "Kbackflow: niter = " << niter << endl;
    return(b2BB_lambda_B);
  }

  void Iterate_XXZ_b2BB_lambda_lambdap (DP zeta, DP B, DP lambdap, Root_Density& b2BB)
  {
    // Calculates the vector corresponding to b2BB (lambda,lambdap)

    // First, copy the actual values into the previous ones:
    for (int i = 0; i < b2BB.Npts; ++i) b2BB.prev_value[i] = b2BB.value[i];

    // Calculate the new values:
    DP conv = 0.0;
    DP sin2zeta = sin(2.0*zeta);
    DP cos2zeta = cos(2.0*zeta);
    for (int i = 0; i < b2BB.Npts; ++i) {
      if (fabs(b2BB.lambda[i]) > B) b2BB.value[i] = 0.0;
      else {
	// First, calculate the convolution
	conv = 0.0;
	for (int j = 0; j < b2BB.Npts; ++j)
	  conv += fabs(b2BB.lambda[j]) > B ? 0.0 : 
	    XXZ_a2_kernel (sin2zeta, cos2zeta, b2BB.lambda[i] - b2BB.lambda[j]) 
	    * b2BB.prev_value[j] * b2BB.dlambda[j];
	b2BB.value[i] = -XXZ_a2_kernel(sin2zeta, cos2zeta, b2BB.lambda[i] - lambdap) 
	  - conv;
      }      
    }
    // Calculate the sum of differences:
    b2BB.diff = 0.0;
    for (int i = 0; i < b2BB.Npts; ++i) 
      b2BB.diff += fabs(b2BB.value[i] - b2BB.prev_value[i]);

    return;
  }

  Root_Density XXZ_b2BB_lambda_lambdap (DP Delta, DP B, DP lambdap, DP lambdamax, int Npts, DP req_prec) 
  {

    // This function returns a Root_Density object corresponding to the inverse
    // kernel b2BB (lambda, lambdap) used in Kbar function.

    if (Delta >= 1.0 || Delta <= 0.0) JSCerror("Delta out of bounds in XXZ_Kbackflow_GS.");

    DP zeta = acos(Delta);

    Root_Density b2BB_lambda_lambdap(Npts, lambdamax);

    // We now attempt to find a solution...

    int niter = 0;
    do {
      Iterate_XXZ_b2BB_lambda_lambdap (zeta, B, lambdap, b2BB_lambda_lambdap);
      //cout << eps_GS.diff << endl;
      niter++;
    } while (b2BB_lambda_lambdap.diff > req_prec);

    //if (lambda_p + lambda_h + 2.0*B > 0.0) cout << "Nonzero backflow possible." << endl;
    //cout << "Kbackflow: niter = " << niter << endl;
    return(b2BB_lambda_lambdap);
  }

  void Iterate_XXZ_Kbackflow_GS (DP zeta, DP B, DP lambda_p, DP lambda_h, Root_Density& Kbackflow_GS)
  {
    // Performs an iteration of the TBA equation for Kbackflow_GS:

    // First, copy the actual values into the previous ones:
    for (int i = 0; i < Kbackflow_GS.Npts; ++i) Kbackflow_GS.prev_value[i] = Kbackflow_GS.value[i];

    // Calculate the new values:
    DP conv = 0.0;
    DP sin2zeta = sin(2.0*zeta);
    DP cos2zeta = cos(2.0*zeta);
    for (int i = 0; i < Kbackflow_GS.Npts; ++i) {
      //if (Kbackflow_GS.lambda[i] < -B || Kbackflow_GS.lambda[i] > lambda_p + lambda_h + B) Kbackflow_GS.value[i] = 0.0;
      if (false && fabs(Kbackflow_GS.lambda[i]) > B) Kbackflow_GS.value[i] = 0.0;
      else {
	// First, calculate the convolution
	conv = 0.0;
	for (int j = 0; j < Kbackflow_GS.Npts; ++j)
	  conv += fabs(Kbackflow_GS.lambda[j]) > B ? 0.0 : 
	    //(Kbackflow_GS.lambda[i] < -B || Kbackflow_GS.lambda[i] > lambda_p + lambda_h + B) ? 0.0 : 
	    XXZ_a2_kernel (sin2zeta, cos2zeta, Kbackflow_GS.lambda[i] - Kbackflow_GS.lambda[j]) 
	    * Kbackflow_GS.prev_value[j] * Kbackflow_GS.dlambda[j];
	Kbackflow_GS.value[i] = -XXZ_a2_kernel(sin2zeta, cos2zeta, Kbackflow_GS.lambda[i] - lambda_p) 
	  + XXZ_a2_kernel(sin2zeta, cos2zeta, Kbackflow_GS.lambda[i] - lambda_h) - conv;
      }      
    }
    // Calculate the sum of differences:
    Kbackflow_GS.diff = 0.0;
    for (int i = 0; i < Kbackflow_GS.Npts; ++i) 
      Kbackflow_GS.diff += fabs(Kbackflow_GS.value[i] - Kbackflow_GS.prev_value[i]);

    return;
  }

  Root_Density XXZ_Kbackflow_GS (DP Delta, DP B, DP lambdamax, DP lambda_p, DP lambda_h, int Npts, DP req_prec) 
  {

    // This function returns a Root_Density object corresponding to the ground state
    // backflow function of the XXZ model.

    if (Delta >= 1.0 || Delta <= 0.0) JSCerror("Delta out of bounds in XXZ_Kbackflow_GS.");

    DP zeta = acos(Delta);

    Root_Density Kbackflow_GS(Npts, lambdamax);

    // We now attempt to find a solution...

    int niter = 0;
    do {
      Iterate_XXZ_Kbackflow_GS (zeta, B, lambda_p, lambda_h, Kbackflow_GS);
      //cout << eps_GS.diff << endl;
      niter++;
    } while (Kbackflow_GS.diff > req_prec);

    //if (lambda_p + lambda_h + 2.0*B > 0.0) cout << "Nonzero backflow possible." << endl;
    //cout << "Kbackflow: niter = " << niter << endl;
    return(Kbackflow_GS);
  }

  void Iterate_XXZ_Fbackflow_GS (DP zeta, DP B, DP lambda_p, DP lambda_h, Root_Density& Fbackflow_GS)
  {
    // Performs an iteration of the TBA equation for Fbackflow_GS:

    // First, copy the actual values into the previous ones:
    for (int i = 0; i < Fbackflow_GS.Npts; ++i) Fbackflow_GS.prev_value[i] = Fbackflow_GS.value[i];

    // Calculate the new values:
    DP conv = 0.0;
    DP sin2zeta = sin(2.0*zeta);
    DP cos2zeta = cos(2.0*zeta);
    for (int i = 0; i < Fbackflow_GS.Npts; ++i) {
      if (fabs(Fbackflow_GS.lambda[i]) > B) Fbackflow_GS.value[i] = 0.0;
      else {
	// First, calculate the convolution
	conv = 0.0;
	for (int j = 0; j < Fbackflow_GS.Npts; ++j)
	  conv += fabs(Fbackflow_GS.lambda[j]) > B ? 0.0 : 
	    XXZ_a2_kernel (sin2zeta, cos2zeta, Fbackflow_GS.lambda[i] - Fbackflow_GS.lambda[j]) 
	    * Fbackflow_GS.prev_value[j] * Fbackflow_GS.dlambda[j];
	Fbackflow_GS.value[i] = (-XXZ_phi2_kernel(zeta, Fbackflow_GS.lambda[i] - lambda_p) 
				 + XXZ_phi2_kernel(zeta, Fbackflow_GS.lambda[i] - lambda_h))/twoPI - conv;
      }      
    }
    // Calculate the sum of differences:
    Fbackflow_GS.diff = 0.0;
    for (int i = 0; i < Fbackflow_GS.Npts; ++i) 
      Fbackflow_GS.diff += fabs(Fbackflow_GS.value[i] - Fbackflow_GS.prev_value[i]);

    return;
  }

  Root_Density XXZ_Fbackflow_GS (DP Delta, DP B, DP lambdamax, DP lambda_p, DP lambda_h, int Npts, DP req_prec) 
  {

    // This function returns a Root_Density object corresponding to the ground state
    // F backflow function of the XXZ model.

    if (Delta >= 1.0 || Delta <= 0.0) JSCerror("Delta out of bounds in XXZ_Fbackflow_GS.");

    DP zeta = acos(Delta);

    Root_Density Fbackflow_GS(Npts, lambdamax);

    // We now attempt to find a solution...

    int niter = 0;
    do {
      Iterate_XXZ_Fbackflow_GS (zeta, B, lambda_p, lambda_h, Fbackflow_GS);
      //cout << eps_GS.diff << endl;
      niter++;
    } while (Fbackflow_GS.diff > req_prec);

    //if (lambda_p + lambda_h + 2.0*B > 0.0) cout << "Nonzero backflow possible." << endl;
    //cout << "Fbackflow: niter = " << niter << endl;
    return(Fbackflow_GS);
  }

  void Iterate_XXZ_Z_GS (DP zeta, DP B, Root_Density& Z_GS)
  {
    // Performs an iteration of the TBA equation for the dressed charge:

    // First, copy the actual values into the previous ones:
    for (int i = 0; i < Z_GS.Npts; ++i) Z_GS.prev_value[i] = Z_GS.value[i];

    // Calculate the new values:
    DP conv = 0.0;
    DP sin2zeta = sin(2.0*zeta);
    DP cos2zeta = cos(2.0*zeta);
    for (int i = 0; i < Z_GS.Npts; ++i) {
      // First, calculate the convolution
      conv = 0.0;
      for (int j = 0; j < Z_GS.Npts; ++j)
	conv += fabs(Z_GS.lambda[j]) > B ? 0.0 : 
	  XXZ_a2_kernel (sin2zeta, cos2zeta, Z_GS.lambda[i] - Z_GS.lambda[j]) 
	  * Z_GS.prev_value[j] * Z_GS.dlambda[j];
      Z_GS.value[i] = 1.0 - conv;
    }      

    // Calculate the sum of differences:
    Z_GS.diff = 0.0;
    for (int i = 0; i < Z_GS.Npts; ++i) 
      Z_GS.diff += fabs(Z_GS.value[i] - Z_GS.prev_value[i]);

    return;
  }

  Root_Density XXZ_Z_GS (DP Delta, DP B, DP lambdamax, int Npts, DP req_prec) 
  {

    // This function returns a Root_Density object corresponding to the ground state
    // dressed charge function of the XXZ model.

    if (Delta >= 1.0 || Delta <= 0.0) JSCerror("Delta out of bounds in XXZ_Z_GS.");

    DP zeta = acos(Delta);

    Root_Density Z_GS(Npts, lambdamax);

    // We now attempt to find a solution...

    do {
      Iterate_XXZ_Z_GS (zeta, B, Z_GS);
      //cout << depsdlambda_GS.diff << endl;
    } while (Z_GS.diff > req_prec);

    return(Z_GS);
  }

  /*
  void XXZ_Compare_Lattice_and_Continuum_Backflows_base_1010 (DP Delta, int N, int M, long long int id)
  {
    // Define the chain:  J, Delta, h, Nsites
    Heis_Chain chain(1.0, Delta, 0.0, N);

    // Define the base:  chain, Mdown
    Heis_Base gbase(chain, M);
    
    // Define the ground state
    XXZ_Bethe_State gstate(chain, gbase);
    
    // Compute everything about the ground state
    gstate.Compute_All(true);
    
    // Define the number of rapidities for an excited state
    Vect_INT Nrapidities(0, chain.Nstrings);
    Nrapidities[0] = M;
    
    Vect_INT Nexcitations(0, 2* chain.Nstrings + 2);
    Nexcitations[0] = 0;
    Nexcitations[1] = 1;
    Nexcitations[2] = 0;
    Nexcitations[3] = 1;

    // Define a base configuration for this set of rapidity numbers
    Heis_Base ebase (chain, Nrapidities);
      
    // Define the excited state 
    XXZ_Bethe_State estate(chain, ebase);
    
    // Define an offset from a base and a number of holes
    Ix2_Offsets offsets(ebase, Nexcitations);
    
    // Set the offset to the desired id
    offsets.Set_to_id (id);
	
    // Set the offset data into the quantum numbers
    estate.Set_Ix2_Offsets(offsets);
	
    // Compute everything about this eigenstate
    estate.Compute_All(true);

    DP k_ext_N = estate.K - gstate.K;
    DP omega_N = estate.E - gstate.E;
    DP lambdap = estate.lambda[0][0];
    DP lambdah = 0.0;
    for (int alpha = 1; alpha < estate.base[0] - 1; ++alpha)
      if (estate.Ix2[0][alpha + 1] > estate.Ix2[0][alpha] + 2) {
	lambdah = 0.5 * (estate.lambda[0][alpha] + estate.lambda[0][alpha + 1]);
      }


    // Now solve the thermodynamic equations:
    DP req_prec = 1.0e-12;
    DP Hz_N = H_vs_M (Delta, N, M);
    cout << "Hz_N = " << Hz_N << endl;
    DP lambdamax = 4.0 * fabs(gstate.lambda[0][0]);
    int Npts = 2000;
    Root_Density eps_GS = XXZ_eps_GS (Delta, Hz_N, lambdamax, Npts, req_prec);

    // We can read off the value of B from the obtained eps_GS function:
    DP B_eps = 0.0;
    int iB_eps = Npts/2;
    for (int i = Npts/2; i < Npts - 1; ++i)
      if (eps_GS.value[i] * eps_GS.value[i+1] < 0.0) {
	B_eps = 0.5 * (eps_GS.lambda[i] + eps_GS.lambda[i+1]);
	iB_eps = i;
      }
    if (B_eps == 0.0) {
      cout << "Delta = " << Delta << "\tN = " << N << "\tM = " << M << "\tHz_N = " << Hz_N << "\tid = " << id << endl;
      JSCerror("B not found.");
    }
    if (B_eps > 0.5 * lambdamax) {
      cout << "Delta = " << Delta << "\tN = " << N << "\tM = " << M << "\tHz_N = " << Hz_N << "\tid = " << id << endl;
      JSCerror("Use a higher value of lambdamax.");
    }
    //DP lambdamax = lambdamax_eps;
    //DP lambdamax = 2.0 * fabs(gstate.lambda[0][0]); // window of definition of Kbackflow is covered
    // Start by finding the appropriate B:
    DP B = DP(M)/N; // initial guess;
    DP B_old = B;
    DP dB = 0.5 * B; // variation we allow;
    DP m_TBA = 0.0;
    //Vect<Root_Density> rhotot_GS_iter(10);// = XXZ_rhotot_GS (Delta, B, lambdamax, Npts, req_prec);
    Root_Density rhotot_GS = XXZ_rhotot_GS (Delta, B, lambdamax, Npts, req_prec);
    int iter = 0;
    do {
      B_old = B;
      rhotot_GS = XXZ_rhotot_GS (Delta, B, lambdamax, Npts, req_prec);
      m_TBA = 0.0;
      for (int i1 = 0; i1 < rhotot_GS.Npts; ++i1) if (fabs(rhotot_GS.lambda[i1]) < B) m_TBA += rhotot_GS.value[i1] * rhotot_GS.dlambda[i1];
      if (m_TBA < DP(M)/N) B += dB;
      else B -= dB;
      cout << "B_old = " << B_old << "\tB = " << B << "\tdB = " << dB << "\tm = " << m_TBA << "\tM/N = " << DP(M)/N << endl;
      dB *= 1.0/1.9;
    } while (dB > 0.001 && iter < 9);

    cout << "Check of consistency:  B_eps = " << B_eps << "\tB = " << B << "\tm_TBA = " << m_TBA << "\tM/N = " << DP(M)/N << endl;

    //Root_Density rhotot_GS = XXZ_rhotot_GS (Delta, B, lambdamax, Npts, req_prec);
    Root_Density b2BB_lambda_B = XXZ_b2BB_lambda_B (Delta, B, lambdamax, Npts, req_prec);
    Root_Density Kbackflow_GS = XXZ_Kbackflow_GS (Delta, B, lambdamax, lambdap, lambdah, Npts, req_prec);
    Root_Density Fbackflow_GS = XXZ_Fbackflow_GS (Delta, B, lambdamax, lambdap, lambdah, Npts, req_prec);

    // Calculate the momentum and energy of the TBA state:
    // Momentum:
    DP conv = 0.0;
    DP zeta = acos(Delta);
    for (int l = 0; l < rhotot_GS.Npts; ++l) {
      conv += fabs(rhotot_GS.lambda[l]) > B ? 0.0 : 
	(XXZ_phi2_kernel(zeta, lambdap - rhotot_GS.lambda[l]) 
	 - XXZ_phi2_kernel(zeta, lambdah - rhotot_GS.lambda[l]))
	* rhotot_GS.value[l] * rhotot_GS.dlambda[l];
    }
    DP k_ext_TBA = -XXZ_phi1_kernel (zeta, lambdap) + XXZ_phi1_kernel (zeta, lambdah) + conv; 
    // Energy:
    conv = 0.0;
    DP sinzeta = sin(zeta);
    DP coszeta = cos(zeta);
    for (int l = 0; l < Kbackflow_GS.Npts; ++l)
      conv += fabs(Kbackflow_GS.lambda[l]) > B ? 0.0 :
	(Hz_N - PI * sinzeta * XXZ_a1_kernel(sinzeta, coszeta, Kbackflow_GS.lambda[l]))
	* Kbackflow_GS.value[l] * Kbackflow_GS.dlambda[l];
    DP omega_TBA = -PI * sin(zeta) * (XXZ_a1_kernel(sinzeta, coszeta, lambdap)
				      - XXZ_a1_kernel(sinzeta, coszeta, lambdah)) + conv;

    cout << "k_ext_N = " << k_ext_N << "\tk_ext_TBA = " << k_ext_TBA << "\tomega_N = " << omega_N << "\tomega_TBA = " << omega_TBA << endl;

    // Get back to the finite lattice case:
    // Define densities as 1/N \sum_j delta (lambda - lambda_j);  use Gaussian smoothing of functions.
    DP gwidth = 2.0 * B * log(DP(N))/M; // mean rapidity spacing times some factor, for smoothing of Gaussians
    DP gwidthsq = gwidth * gwidth;
    Vect_DP rhoGS_N (0.0, Npts);
    Vect_DP rhoexc_N (0.0, Npts);
    Vect_DP Kflow_N (0.0, Npts);
    for (int iout = 0; iout < Npts; ++iout) {
      DP lambdaout = rhotot_GS.lambda[iout];
      // We compute the densities at each point...
      for (int alpha = 0; alpha < M; ++alpha)
	rhoGS_N[iout] += exp(-(gstate.lambda[0][alpha] - lambdaout) * (gstate.lambda[0][alpha] - lambdaout)/gwidthsq);
      rhoGS_N[iout] /= N * sqrt(PI) * gwidth;
      for (int alpha = 1; alpha < estate.base[0]; ++alpha) 
	rhoexc_N[iout] += exp(-(estate.lambda[0][alpha] - lambdaout) * (estate.lambda[0][alpha] - lambdaout)/gwidthsq);
      rhoexc_N[iout] /= N * sqrt(PI) * gwidth;
      
      Kflow_N[iout] = N * (rhoexc_N[iout] - rhoGS_N[iout] + (1.0/N) * exp(-(lambdah - lambdaout) * (lambdah - lambdaout)/gwidthsq)/(sqrt(PI) * gwidth));
    } // for iout

    cout << "Here 1" << endl;
    // Now produce a file with the density flow:
    stringstream flo_stringstream;
    flo_stringstream << "Flow_D_" << Delta << "_N_" << N << "_M_" << M << "_base_id_" << estate.base_id << "_type_id_" << estate.type_id 
		     << "_id_" << estate.id << ".flo";
    string flo_string = flo_stringstream.str();
    const char* flo_Cstr = flo_string.c_str();
    
    ofstream outfile_flo;
    outfile_flo.open(flo_Cstr);
    outfile_flo.precision(16);

    for (int iout = 0; iout < Npts; ++iout) 
      outfile_flo << rhotot_GS.lambda[iout] << "\t" << rhotot_GS.value[iout] << "\t" << rhoGS_N[iout] << "\t" 
		  << rhoexc_N[iout] << "\t" << Kbackflow_GS.value[iout] << "\t" 
		  << (fabs(Kbackflow_GS.lambda[iout]) > B ? 0.0 : Kbackflow_GS.value[iout]) - b2BB_lambda_B.value[iout] * Fbackflow_GS.value[iB_eps] 
	+ b2BB_lambda_B.value[b2BB_lambda_B.Npts - 1 - iout] * Fbackflow_GS.value[Fbackflow_GS.Npts - 1 - iB_eps] << "\t" << Kflow_N[iout] << endl;

    outfile_flo.close();

    // ... and a file with the remainder of the data:
    stringstream dat_stringstream;
    dat_stringstream << "Flow_D_" << Delta << "_N_" << N << "_M_" << M << "_base_id_" << estate.base_id << "_type_id_" << estate.type_id 
		     << "_id_" << estate.id << ".dat";
    string dat_string = dat_stringstream.str();
    const char* dat_Cstr = dat_string.c_str();
    
    ofstream outfile_dat;
    outfile_dat.open(dat_Cstr);
    outfile_dat.precision(16);

    outfile_dat << "lambdap\tlambdah\tk_ext_N\tk_ext_TBA\tomega_N\tomega_TBA" 
		<< lambdap << "\t" << lambdah << "\t" << k_ext_N << "\t" << k_ext_TBA << "\t" << omega_N << "\t" << omega_TBA << endl;

    outfile_dat.close();

    return;
  }
  */

}  // namespace JSC



