/**********************************************************

This software is part of J.-S. Caux's C++ library.

Copyright (c) 2006.

-----------------------------------------------------------

File:  SF_4p_server.cc

Purpose:  4 spinon contrib to S(k, omega) for XXX_h0, parallel code using MPICH.
          Server part.

Last modified:  13/10/06

***********************************************************/

#include "mpi.h"
#include "JSC.h"

namespace JSC {


  void SF_4p_opt_server (MPI_Comm comm, int N, DP omegamax, int Nomega, DP req_prec, int Npts_K, int Npts_W, I_table Itable, 
			 int Max_Secs, bool refine)
  {
    double Start_Time_MPI = MPI_Wtime();

    Heis_Write_Kfull_File (N);
    Heis_Write_w_File (Nomega, omegamax);

    stringstream SF_stringstream;
    string SF_string;
    SF_stringstream << "SF_4p_N_" << N << "_No_" << Nomega << "_omax_" << omegamax 
		    << "_prec_" << req_prec << "_Npts_K_" << Npts_K << "_Npts_W_" << Npts_W << ".dat";
    SF_string = SF_stringstream.str();
    const char* SF_Cstr = SF_string.c_str();
    
    fstream SF_outfile;
    if (!refine) SF_outfile.open(SF_Cstr, fstream::out | fstream::trunc);
    else SF_outfile.open(SF_Cstr, fstream::in | fstream::out); 
    if (SF_outfile.fail()) JSCerror("Could not open SF_outfile... "); 

    SF_outfile.precision(10);

    int dim_k = N/2 + 1;

    DP* SF_4p = new DP[dim_k * Nomega];

    // Initialize to zero to be sure:
    for (int i = 0; i < Nomega * dim_k; ++i) SF_4p[i] = 0.0;

    // If refining, load SF from existing file

    int total_nr_req = 0;
    DP buff;
    if (refine) {
      for (int iomega = 0; iomega < Nomega; ++iomega) {
	for (int ik = 0; ik < dim_k; ++ik) {
	  SF_outfile >> SF_4p[dim_k * iomega + ik];
	  if (SF_4p[dim_k * iomega + ik] == 0.0) total_nr_req++;
	  //cout << ik << "\t" << iomega << "\t" << SF_4p[dim_k * iomega + ik] << "\t" << (SF_4p[dim_k * iomega + ik] == 0.0) << "\t" << total_nr_req << endl;
	  
	  // We only load the LHS of the BZ, so we load N/2 empty values...
	}
	for (int ibuff = 0; ibuff < N/2; ++ibuff) SF_outfile >> buff;
      }
    }
    else if (!refine) total_nr_req = dim_k * Nomega;

    // List iomega and ik which need to be computed:

    int* needed_ik = new int[total_nr_req];
    int* needed_iomega = new int[total_nr_req];

    int index = 0;
    for (int iomega = 0; iomega < Nomega; ++iomega) 
      for (int ik = 0; ik < dim_k; ++ik) {
	if (SF_4p[dim_k * iomega + ik] == 0) {
	  needed_ik[index] = ik;
	  needed_iomega[index] = iomega;
	  index++;
	}
      }

    cout << total_nr_req << " points required." << endl;

    if (index != total_nr_req) {
      cout << total_nr_req << "\t" << index << endl;
      JSCerror("Not counting total_nr_req correctly in SF_4p_opt_server");
    }
    //ofstream SFsrc_outfile;
    //SFsrc_outfile.open(SFsrc_Cstr);
    
    //DP omegamin_used = 0.5 * wmin_4p (k); // Correct for factor of 2 in E between me & Bougourzi
    //DP omegamax_used = 0.5 * wmax_4p (k);

    int nr_machines;
    MPI_Comm_size (comm, &nr_machines);
    int nr_clients = nr_machines - 1;  // one for the server

    int client_request_size = 7;
    DP client_request[client_request_size];
    // this is:
    // ik
    // iomega
    // k
    // omega
    // req_prec
    // Npts_K
    // Npts_W
    int client_result_size = 8;
    DP client_result[client_result_size]; // same, plus SF_4p

    MPI_Status status;

    int scanning = 1;

    int nr_sent_out = 0;
    int nr_returned = 0;

    for (int i = 1; i <= nr_clients && i <= total_nr_req; ++i) {

      // Send request to client i, in the form of the req_id_array vector
      client_request[0] = DP(needed_ik[nr_sent_out]);
      client_request[1] = DP(needed_iomega[nr_sent_out]);
      client_request[2] = (twoPI * needed_ik[nr_sent_out])/N;
      client_request[3] = (needed_iomega[nr_sent_out] + 0.5) * omegamax/Nomega;
      client_request[4] = req_prec;
      client_request[5] = DP(Npts_K);
      client_request[6] = DP(Npts_W);

      MPI_Send(client_request, client_request_size, MPI_DOUBLE, i, scanning, comm);
      //cout << "Server:  sent request ";
      //for (int ii = 0; ii < client_request_size; ++ii) cout << client_request[ii] << " ";
      //cout << "with tag " << scanning << " to client " << i << endl;
      nr_sent_out++;
      //cout << "nr_sent_out = " << nr_sent_out << endl;
    }

    DP Actual_Time_MPI = MPI_Wtime();

    //while (nr_returned < total_nr_req) { 
    while (nr_returned < nr_sent_out) {

      //cout << "Server:  waiting for answers... " << endl;
      MPI_Recv (client_result, client_result_size, MPI_DOUBLE, MPI_ANY_SOURCE, MPI_ANY_TAG, comm, &status);
      nr_returned++;
      //cout << "Server:  received answer from client " << status.MPI_SOURCE << endl;
      // unbuffer result
      SF_4p[dim_k * int(client_result[1]) + int(client_result[0])] = client_result[7];

      Actual_Time_MPI = MPI_Wtime();

      // Send out new request if needed and time available
      if (nr_sent_out < total_nr_req && Actual_Time_MPI - Start_Time_MPI < Max_Secs) {
	client_request[0] = needed_ik[nr_sent_out];
	client_request[1] = needed_iomega[nr_sent_out];
	client_request[2] = (twoPI * needed_ik[nr_sent_out])/N;
	client_request[3] = (needed_iomega[nr_sent_out] + 0.5) * omegamax/Nomega;
	MPI_Send (client_request, client_request_size, MPI_DOUBLE, status.MPI_SOURCE, scanning, comm);
	nr_sent_out++;
	//cout << "nr_sent_out = " << nr_sent_out << endl;
      }

    } // while (nr_returned < total_nr_req)

    cout << endl << "Computed " << nr_returned << " points ouf of " << total_nr_req << " required. " << endl;

    cout << endl << "Server saving file..." << endl;

    // Output all data:  double up to full [0, 2pi] interval in k with symmetry 
    SF_outfile.seekp(0);
    for (int iomega = 0; iomega < Nomega; ++iomega) {
      for (int ik = 0; ik < dim_k; ++ik) SF_outfile << SF_4p[dim_k * iomega + ik] << "\t";
      for (int ik = 2; ik <= dim_k; ++ik) SF_outfile << SF_4p[dim_k * iomega + dim_k - ik] << "\t";
      SF_outfile << endl;
    }

    SF_outfile.close();
    
    cout << endl << "Done !" << endl;

    // Send term signal to clients
    int scanning_completed = 0;
    for (int i = 1; i <= nr_clients; ++i)
      MPI_Send (client_request, client_request_size, MPI_DOUBLE, i, scanning_completed, comm);

    return;
  }

  /****************************************************

  // TO COMPLETE FOR 5.5 (commented out in 5.7)

  *********************************************************/
  /*
  void SF_4p_opt_server (MPI_Comm comm, DP k_needed, DP omegamax, int Nomega, DP req_prec, int Npts_K, int Npts_W, I_table Itable, 
			 int Max_Secs, bool refine)
  {
    double Start_Time_MPI = MPI_Wtime();

    Heis_Write_w_File (Nomega, omegamax);

    stringstream SF_stringstream;
    string SF_string;
    SF_stringstream << "SF_4p_k_" << k_needed << "_No_" << Nomega << "_omax_" << omegamax 
		    << "_prec_" << req_prec << "_Npts_K_" << Npts_K << "_Npts_W_" << Npts_W << ".dat";
    SF_string = SF_stringstream.str();
    const char* SF_Cstr = SF_string.c_str();
    
    fstream SF_outfile;
    if (!refine) SF_outfile.open(SF_Cstr, fstream::out | fstream::trunc);
    else SF_outfile.open(SF_Cstr, fstream::in | fstream::out); 
    if (SF_outfile.fail()) JSCerror("Could not open SF_outfile... "); 

    SF_outfile.precision(12);

    int dim_k = 1;

    DP* SF_4p = new DP[dim_k * Nomega];

    // Initialize to zero to be sure:
    for (int i = 0; i < Nomega * dim_k; ++i) SF_4p[i] = 0.0;

    // If refining, load SF from existing file

    int total_nr_req = 0;
    DP buff;
    if (refine) {
      for (int iomega = 0; iomega < Nomega; ++iomega) {
	for (int ik = 0; ik < dim_k; ++ik) {
	  SF_outfile >> SF_4p[dim_k * iomega + ik];
	  if (SF_4p[dim_k * iomega + ik] == 0.0) total_nr_req++;
	  //cout << ik << "\t" << iomega << "\t" << SF_4p[dim_k * iomega + ik] << "\t" << (SF_4p[dim_k * iomega + ik] == 0.0) << "\t" << total_nr_req << endl;
	  
	  // We only load the LHS of the BZ, so we load N/2 empty values...
	}
	for (int ibuff = 0; ibuff < N/2; ++ibuff) SF_outfile >> buff;
      }
    }
    else if (!refine) total_nr_req = dim_k * Nomega;

    // List iomega and ik which need to be computed:

    int* needed_ik = new int[total_nr_req];
    int* needed_iomega = new int[total_nr_req];

    int index = 0;
    for (int iomega = 0; iomega < Nomega; ++iomega) 
      for (int ik = 0; ik < dim_k; ++ik) {
	if (SF_4p[dim_k * iomega + ik] == 0) {
	  needed_ik[index] = ik;
	  needed_iomega[index] = iomega;
	  index++;
	}
      }

    cout << total_nr_req << " points required." << endl;

    if (index != total_nr_req) {
      cout << total_nr_req << "\t" << index << endl;
      JSCerror("Not counting total_nr_req correctly in SF_4p_opt_server");
    }
    //ofstream SFsrc_outfile;
    //SFsrc_outfile.open(SFsrc_Cstr);
    
    //DP omegamin_used = 0.5 * wmin_4p (k); // Correct for factor of 2 in E between me & Bougourzi
    //DP omegamax_used = 0.5 * wmax_4p (k);

    int nr_machines;
    MPI_Comm_size (comm, &nr_machines);
    int nr_clients = nr_machines - 1;  // one for the server

    int client_request_size = 7;
    DP client_request[client_request_size];
    // this is:
    // ik
    // iomega
    // k
    // omega
    // req_prec
    // Npts_K
    // Npts_W
    int client_result_size = 8;
    DP client_result[client_result_size]; // same, plus SF_4p

    MPI_Status status;

    int scanning = 1;

    int nr_sent_out = 0;
    int nr_returned = 0;

    for (int i = 1; i <= nr_clients && i <= total_nr_req; ++i) {

      // Send request to client i, in the form of the req_id_array vector
      client_request[0] = DP(needed_ik[nr_sent_out]);
      client_request[1] = DP(needed_iomega[nr_sent_out]);
      client_request[2] = k_needed;//(twoPI * needed_ik[nr_sent_out])/N;
      client_request[3] = (needed_iomega[nr_sent_out] + 0.5) * omegamax/Nomega;
      client_request[4] = req_prec;
      client_request[5] = DP(Npts_K);
      client_request[6] = DP(Npts_W);

      MPI_Send(client_request, client_request_size, MPI_DOUBLE, i, scanning, comm);
      //cout << "Server:  sent request ";
      //for (int ii = 0; ii < client_request_size; ++ii) cout << client_request[ii] << " ";
      //cout << "with tag " << scanning << " to client " << i << endl;
      nr_sent_out++;
      //cout << "nr_sent_out = " << nr_sent_out << endl;
    }

    DP Actual_Time_MPI = MPI_Wtime();

    //while (nr_returned < total_nr_req) { 
    while (nr_returned < nr_sent_out) {

      //cout << "Server:  waiting for answers... " << endl;
      MPI_Recv (client_result, client_result_size, MPI_DOUBLE, MPI_ANY_SOURCE, MPI_ANY_TAG, comm, &status);
      nr_returned++;
      //cout << "Server:  received answer from client " << status.MPI_SOURCE << endl;
      // unbuffer result
      SF_4p[int(client_result[1])] = client_result[7];

      Actual_Time_MPI = MPI_Wtime();

      // Send out new request if needed and time available
      if (nr_sent_out < total_nr_req && Actual_Time_MPI - Start_Time_MPI < Max_Secs) {
	client_request[0] = needed_ik[nr_sent_out];
	client_request[1] = needed_iomega[nr_sent_out];
	client_request[2] = k_needed;//(twoPI * needed_ik[nr_sent_out])/N;
	client_request[3] = (needed_iomega[nr_sent_out] + 0.5) * omegamax/Nomega;
	MPI_Send (client_request, client_request_size, MPI_DOUBLE, status.MPI_SOURCE, scanning, comm);
	nr_sent_out++;
	//cout << "nr_sent_out = " << nr_sent_out << endl;
      }

    } // while (nr_returned < total_nr_req)

    cout << endl << "Computed " << nr_returned << " points ouf of " << total_nr_req << " required. " << endl;

    cout << endl << "Server saving file..." << endl;

    // Output all data:  
    SF_outfile.seekp(0);
    for (int iomega = 0; iomega < Nomega; ++iomega) {
      SF_outfile << omega[iomega] << "\t" << SF_4p[iomega] << endl;
    }

    SF_outfile.close();
    
    cout << endl << "Done !" << endl;

    // Send term signal to clients
    int scanning_completed = 0;
    for (int i = 1; i <= nr_clients; ++i)
      MPI_Send (client_request, client_request_size, MPI_DOUBLE, i, scanning_completed, comm);

    return;
  }
  */
  /*
  // Function producing a fixed k scan, with data file
  void SF_4p_rec_par (MPI_Comm comm, DP k, DP req_prec, int max_rec_w, int max_rec, I_table Itable)
  {
    int Npts_w = int(pow(3.0, max_rec_w + 2));

    stringstream SFraw_stringstream;
    string SFraw_string;
    SFraw_stringstream << "SF_4p_k_" << k << "_prec_" << req_prec << "_max_rec_w_" << max_rec_w << "_max_rec_" << max_rec << ".raw";
    SFraw_string = SFraw_stringstream.str();
    const char* SFraw_Cstr = SFraw_string.c_str();
    
    stringstream SF_stringstream;
    string SF_string;
    SF_stringstream << "SF_4p_k_" << k << "_prec_" << req_prec << "_max_rec_w_" << max_rec_w << "_max_rec_" << max_rec << ".dat";
    SF_string = SF_stringstream.str();
    const char* SF_Cstr = SF_string.c_str();
    
    stringstream SFsrc_stringstream;
    string SFsrc_string;
    SFsrc_stringstream << "SF_4p_k_" << k << "_prec_" << req_prec << "_max_rec_w_" << max_rec_w << "_max_rec_" << max_rec << ".src";
    SFsrc_string = SFsrc_stringstream.str();
    const char* SFsrc_Cstr = SFsrc_string.c_str();
    
    ofstream SFraw_outfile;
    SFraw_outfile.open(SFraw_Cstr);
    //ofstream SFsrc_outfile;
    //SFsrc_outfile.open(SFsrc_Cstr);
    
    //DP omegamin_used = 0.5 * wmin_4p (k); // Correct for factor of 2 in E between me & Bougourzi
    //DP omegamax_used = 0.5 * wmax_4p (k);

    int nr_machines;
    MPI_Comm_size (comm, &nr_machines);
    int nr_clients = nr_machines - 1;  // one for the server

    int client_request_size = 2;
    DP client_request[client_request_size]; // this is k, alpha
    int client_result_size = 3;
    DP client_result[client_result_size]; // this is k, alpha, SF_4p_kwKW_alpha

    MPI_Status status;

    int scanning = 1;

    int total_nr_req = Npts_w;
    int nr_sent_out = 0;
    int nr_returned = 0;

    Vect_DP alpha_req(Npts_w);
    for (int iw = 0; iw < Npts_w; ++iw) alpha_req[iw] = 0.5 * PI * (iw + 0.5)/Npts_w;

    //cout << "alpha_req = " << alpha_req << endl;

    for (int i = 1; i <= nr_clients && i <= total_nr_req; ++i) {

      // Send request to client i, in the form of the req_id_array vector
      client_request[0] = k;
      client_request[1] = alpha_req[nr_sent_out];

      MPI_Send(client_request, client_request_size, MPI_DOUBLE, i, scanning, comm);
      //cout << "Server:  sent request ";
      //for (int ii = 0; ii < client_request_size; ++ii) cout << client_request[ii] << " ";
      //cout << "with tag " << scanning << " to client " << i << endl;
      nr_sent_out++;
      //cout << "nr_sent_out = " << nr_sent_out << endl;
    }

    while (nr_returned < total_nr_req) {
      
      //cout << "Server:  waiting for answers... " << endl;
      MPI_Recv (client_result, client_result_size, MPI_DOUBLE, MPI_ANY_SOURCE, MPI_ANY_TAG, comm, &status);
      nr_returned++;
      //cout << "Server:  received answer from client " << status.MPI_SOURCE << endl;

      // unbuffer result
      SFraw_outfile << client_result[1] << "\t" << client_result[2] << endl;

      // Send out new request if needed
      if (nr_sent_out < total_nr_req) {
	client_request[1] = alpha_req[nr_sent_out];
	MPI_Send (client_request, client_request_size, MPI_DOUBLE, status.MPI_SOURCE, scanning, comm);
	nr_sent_out++;
	//cout << "nr_sent_out = " << nr_sent_out << endl;
      }

    } // while (nr_returned < total_nr_req)

    // Send term signal to clients
    int scanning_completed = 0;
    for (int i = 1; i <= nr_clients; ++i)
      MPI_Send (client_request, client_request_size, MPI_DOUBLE, i, scanning_completed, comm);

    SFraw_outfile.close();
    
    //SFsrc_outfile << answer << endl;
    //SFsrc_outfile.close();
    
    // Translate raw data into SF_4p (k,omega) data
    
    //Translate_raw_4p_data (k, max_rec_w, SFraw_Cstr, SF_Cstr, SFsrc_Cstr, Itable);
    Translate_raw_4p_data (k, int(pow(3.0, max_rec_w + 2)), SFraw_Cstr, SF_Cstr, SFsrc_Cstr, Itable);
    
    return;
  }

  void SF_4p_opt_par (MPI_Comm comm, DP k, DP req_prec, int Npts_w, int Npts_K, int Npts_W, I_table Itable)
  {
    stringstream SFraw_stringstream;
    string SFraw_string;
    SFraw_stringstream << "SF_4p_k_" << k << "_prec_" << req_prec << "_Npts_" << Npts_w << "_" << Npts_K << "_" << Npts_W << ".raw";
    SFraw_string = SFraw_stringstream.str();
    const char* SFraw_Cstr = SFraw_string.c_str();
    
    stringstream SF_stringstream;
    string SF_string;
    SF_stringstream << "SF_4p_k_" << k << "_prec_" << req_prec << "_Npts_" << Npts_w << "_" << Npts_K << "_" << Npts_W << ".dat";
    SF_string = SF_stringstream.str();
    const char* SF_Cstr = SF_string.c_str();
    
    stringstream SFsrc_stringstream;
    string SFsrc_string;
    SFsrc_stringstream << "SF_4p_k_" << k << "_prec_" << req_prec << "_Npts_" << Npts_w << "_" << Npts_K << "_" << Npts_W << ".src";
    SFsrc_string = SFsrc_stringstream.str();
    const char* SFsrc_Cstr = SFsrc_string.c_str();
    
    ofstream SFraw_outfile;
    SFraw_outfile.open(SFraw_Cstr);
    //ofstream SFsrc_outfile;
    //SFsrc_outfile.open(SFsrc_Cstr);
    
    //DP omegamin_used = 0.5 * wmin_4p (k); // Correct for factor of 2 in E between me & Bougourzi
    //DP omegamax_used = 0.5 * wmax_4p (k);

    int nr_machines;
    MPI_Comm_size (comm, &nr_machines);
    int nr_clients = nr_machines - 1;  // one for the server

    int client_request_size = 2;
    DP client_request[client_request_size]; // this is k, alpha
    int client_result_size = 3;
    DP client_result[client_result_size]; // this is k, alpha, SF_4p_kwKW_alpha_opt

    MPI_Status status;

    int scanning = 1;

    int total_nr_req = Npts_w;
    int nr_sent_out = 0;
    int nr_returned = 0;

    Vect_DP alpha_req(Npts_w);
    for (int iw = 0; iw < Npts_w; ++iw) alpha_req[iw] = 0.5 * PI * (iw + 0.5)/Npts_w;

    //cout << "alpha_req = " << alpha_req << endl;

    for (int i = 1; i <= nr_clients && i <= total_nr_req; ++i) {

      // Send request to client i, in the form of the req_id_array vector
      client_request[0] = k;
      client_request[1] = alpha_req[nr_sent_out];

      MPI_Send(client_request, client_request_size, MPI_DOUBLE, i, scanning, comm);
      //cout << "Server:  sent request ";
      //for (int ii = 0; ii < client_request_size; ++ii) cout << client_request[ii] << " ";
      //cout << "with tag " << scanning << " to client " << i << endl;
      nr_sent_out++;
      //cout << "nr_sent_out = " << nr_sent_out << endl;
    }

    while (nr_returned < total_nr_req) {
      
      //cout << "Server:  waiting for answers... " << endl;
      MPI_Recv (client_result, client_result_size, MPI_DOUBLE, MPI_ANY_SOURCE, MPI_ANY_TAG, comm, &status);
      nr_returned++;
      //cout << "Server:  received answer from client " << status.MPI_SOURCE << endl;

      // unbuffer result
      SFraw_outfile << client_result[1] << "\t" << client_result[2] << endl;

      // Send out new request if needed
      if (nr_sent_out < total_nr_req) {
	client_request[1] = alpha_req[nr_sent_out];
	MPI_Send (client_request, client_request_size, MPI_DOUBLE, status.MPI_SOURCE, scanning, comm);
	nr_sent_out++;
	//cout << "nr_sent_out = " << nr_sent_out << endl;
      }

    } // while (nr_returned < total_nr_req)

    // Send term signal to clients
    int scanning_completed = 0;
    for (int i = 1; i <= nr_clients; ++i)
      MPI_Send (client_request, client_request_size, MPI_DOUBLE, i, scanning_completed, comm);

    SFraw_outfile.close();
    
    //SFsrc_outfile << answer << endl;
    //SFsrc_outfile.close();
    
    // Translate raw data into SF_4p (k,omega) data
    
    //Translate_raw_4p_data (k, max_rec_w, SFraw_Cstr, SF_Cstr, SFsrc_Cstr, Itable);
    Translate_raw_4p_data (k, Npts_w, SFraw_Cstr, SF_Cstr, SFsrc_Cstr, Itable);
    
    return;
  }
  */

} // namespace JSC
